package org.protux.core;

public class Fft
	{
	public static final int MAX_FFT_SIZE = 32768;
	public static final float PI = 3.14159265358979323846f;


	public static final boolean IS_POWER_OF_TWO(int x)
		{
		return false;//MIG ((x >= 2) && !(x & (x - 1)));
		}

	public final static float NORMALIZATION_FACTOR = 50.0f;
	public final static int HAMMING_WINDOW = 1;
	public final static int VON_HANN_WINDOW = 2;
	public final static int BLACKMANN_WINDOW = 3;
	public final static int BARTLETT_WINDOW = 4;
	public final static int BLACKMAN_HARRIS_WINDOW = 5;

	float buf_re[] = new float[MAX_FFT_SIZE];
	float buf_im[] = new float[MAX_FFT_SIZE];
	int size;
	int sampleRate;
	float windowCoef[] = new float[MAX_FFT_SIZE];
	int winType;
	float dcComponent;
	float noiseLine;


	int numberOfBitsNeeded(int param)
		{
		int i;
		if (param < 2)
			{
			//PERROR("%d is too small! ",param );
			return 0;
			}
		for (i = 0;; i++)
			{
			if (false) //MIG param & (1 << i))
				return i;
			}
		}


	int reverseBits(int index, int numBits)
		{
		int i, rev;
		for (i = rev = 0; i < numBits; i++)
			{
			rev = (rev << 1) | (index & 1);
			index >>= 1;
			}
		return rev;
		}


	public Fft(int pSize, int pWinType, int pSampleRate)
		{
		size = pSize;
		winType = pWinType;
		sampleRate = pSampleRate;

		switch (winType)
			{
			case BLACKMANN_WINDOW: // Blackman
				{
				for (int n = 0; n < size; n++)
					{
					windowCoef[n] = (float) (0.42f - 0.5f * Math.cos(2.0f * PI * n / (size - 1)) + 0.08 * Math.cos(2.0f * 2.0f * PI * n / (size - 1)));
					}
				break;
				}
			case BARTLETT_WINDOW: // Bartlett
				{
				for (int n = 0; n <= (size - 1) / 2; n++)
					{
					windowCoef[n] = 2.0f * n / (size - 1);
					}
				for (int n = (size - 1) / 2; n < size; n++)
					{
					windowCoef[n] = 2.0f - 2.0f * n / (size - 1);
					}
				break;
				}
			case HAMMING_WINDOW: // Hamming
				{
				for (int n = 0; n < size; n++)
					{
					windowCoef[n] = (float) (0.54 - 0.46 * Math.cos(2.0f * PI * n / (size - 1)));
					}

				break;
				}
			case VON_HANN_WINDOW: // von Hann (sometimes improperly called Hanning)
				{
				for (int n = 0; n < size; n++)
					{
					windowCoef[n] = (float) (0.5f * (1.0f - Math.cos(2.0f * PI * n / (size - 1))));
					}
				break;
				}
			case BLACKMAN_HARRIS_WINDOW: // 4 term Blackman-Harris
				{
				for (int n = 0; n < size; n++)
					{
					windowCoef[n] = (float) (0.35875f - 0.48829f * Math.cos(2.0f * PI * (float) (n + 0.5) / size) + 0.14128 * Math.cos(2.0f * PI * 2.0f * (float) (n + 0.5) / size) - 0.01168 * Math.cos(2.0f * PI * 3. * (float) (n + 0.5) / size));
					}
				break;
				}
			}

		noiseLine = NORMALIZATION_FACTOR / 10; // also empiric. Must be parametrized
		}


	void perform(float[] data_re, float[] outputLevels)
		{

		// apply the smooth window on input data
		for (int i = 0; i < size; i++)
			data_re[i] = data_re[i] * windowCoef[i];

		// Part of this code was adapted from FFT code by Don Cross <dcross@intersrv.com>

		int numBits; // Number of bits needed to store indices
		int k;

		float angleNumerator = 2.0f * PI;
		float tr, ti;

		// Uncomment the following line if you want reverse fft
		// angleNumerator = - angleNumerator;

		numBits = numberOfBitsNeeded(size);

		// Do simultaneous data copy and bit-reversal ordering into outputs...

		for (int i = 0; i < size; i++)
			{
			int j = reverseBits(i, numBits);
			buf_re[j] = data_re[i];
			buf_im[j] = 0;
			}

		int blockEnd = 1;
		for (int blockSize = 2; blockSize <= size; blockSize <<= 1)
			{
			float delta_angle = angleNumerator / blockSize;
			float sm2 = (float) Math.sin(-2 * delta_angle);
			float sm1 = (float) Math.sin(-delta_angle);
			float cm2 = (float) Math.cos(-2 * delta_angle);
			float cm1 = (float) Math.cos(-delta_angle);
			float w = 2 * cm1;
			float ar[] = new float[3], ai[] = new float[3];

			for (int i = 0; i < size; i += blockSize)
				{
				ar[2] = cm2;
				ar[1] = cm1;

				ai[2] = sm2;
				ai[1] = sm1;

				for (int j = i, n = 0; n < blockEnd; j++, n++)
					{
					ar[0] = w * ar[1] - ar[2];
					ar[2] = ar[1];
					ar[1] = ar[0];

					ai[0] = w * ai[1] - ai[2];
					ai[2] = ai[1];
					ai[1] = ai[0];

					k = j + blockEnd;
					tr = ar[0] * buf_re[k] - ai[0] * buf_im[k];
					ti = ar[0] * buf_im[k] + ai[0] * buf_re[k];

					buf_re[k] = buf_re[j] - tr;
					buf_im[k] = buf_im[j] - ti;

					buf_re[j] += tr;
					buf_im[j] += ti;
					}
				}

			blockEnd = blockSize;
			}

		for (int i = 0; i < size; i++)
			{
			buf_re[i] /= size;
			buf_im[i] /= size;
			}

		dcComponent = (buf_re[0] / size) * (buf_re[0] / size);

		for (int kk = 1; kk < size / 2; kk++)
			{
			outputLevels[kk] = (float) Math.sqrt((buf_re[kk] * buf_re[kk]) + (buf_im[kk] * buf_im[kk])) / NORMALIZATION_FACTOR; // - dcComponent;
			}

		}


	float index_to_frequency(int index)
		{
		float f;
		if ((index < 0) || (index > size / 2))
			return -1.0f;
		f = (float) index / (float) size;
		return f * sampleRate;
		}


	int frequency_to_index(float freq)
		{
		if ((freq < 0) || (freq > sampleRate / 2))
			return -1;
		float dindex = size * (freq / sampleRate);
		return (int) dindex;
		}


	public float get_level_noise_line()
		{
		return noiseLine;
		}


	public void set_level_noise_line(float val)
		{
		val = val > 100.0f ? 100.0f : val;
		val = val < 0.0f ? 0.0f : val;
		noiseLine = val;
		}

	//eof

	}
