package org.protux.core;


import static org.protux.core.Debugger.PMESG;


import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;

import org.protux.core.exceptions.GlobalPropertiesException;
import org.protux.core.gui.Popup;

import com.sun.jmx.snmp.Enumerated;

public class GlobalProperties
	{
		private static Properties props = new Properties();
		private static String userPath;
		private static String projectsBaseDir;

		private static String KEY_LAST_PROJECT = "project.last";
		public static String KEY_LOAD_LAST_ON_BOOT = "project.flag.load.last.onboot";
		public static String KEY_WORKSPACE_PATH = "global.workspace.path";
		public static String KEY_SLICE_WIDTH_FACTOR = "audio.slice.width.factor";
		public static String KEY_JMB_CLEAR_TIME = "jmb.clear.time";
		public static String KEY_JMB_HOLD_TIMEOUT = "jmb.hold.time";
		public static String KEY_JMB_DOUBLE_FACT_TIMEOUT = "jmb.double.fact.timeout";
		public static String KEY_DEFAULT_NUM_TRACKS = "editor.default.number.of.tracks";
		public static String KEY_HZOOM_LEVEL = "editor.default.hzoom";
		public static String KEY_DEFAULT_SAMPLE_RATE = "audio.default.sample.rate";
		public static String KEY_DEFAULT_BIT_DEPTH = "audio.default.bit.depth";
		public static String KEY_AUDIO_MAPPER_FPS = "audio.default.fps";
		public static String KEY_USE_EXTERNAL_OGG_DECODER = "factory.ogg.external.decoder";
		public static String KEY_USE_EXTERNAL_OGG_ENCODER = "factory.ogg.external.enconder";
		public static String KEY_MP3_DECODE_COMMAND = "factory.mp3.external.decode.command";
		public static String KEY_OGG_DECODE_COMMAND = "factory.ogg.external.decode.command";
		public static String KEY_OGG_ENCODE_COMMAND = "factory.ogg.external.encode.command";


		public static void init() throws GlobalPropertiesException
			{
			userPath = System.getProperty("user.home");
			PMESG(2, "User Home Directory : " + userPath.toString());
			File rcFile = getRcFile();
			props = new Properties();
			try
				{
				props.load(new FileInputStream(rcFile));
				projectsBaseDir = props.getProperty(KEY_WORKSPACE_PATH);
				while (projectsBaseDir == null)
					{
					projectsBaseDir = chooseProjectdir();
					if (props != null)
						{
						props.setProperty(KEY_WORKSPACE_PATH, projectsBaseDir);
						saveProps();
						}
					}
				}
			catch (FileNotFoundException ex)
				{
				warn("init",rcFile + " was not found. Creating it...");
				projectsBaseDir = chooseProjectdir();
				props.setProperty(KEY_SLICE_WIDTH_FACTOR, "1");
				props.setProperty(KEY_JMB_CLEAR_TIME, "2000");
				props.setProperty(KEY_JMB_HOLD_TIMEOUT, "200");
				props.setProperty(KEY_JMB_DOUBLE_FACT_TIMEOUT, "200");
				props.setProperty(KEY_DEFAULT_NUM_TRACKS, "6");
				props.setProperty(KEY_HZOOM_LEVEL, "10");
				props.setProperty(KEY_DEFAULT_SAMPLE_RATE, "44100");
				props.setProperty(KEY_DEFAULT_BIT_DEPTH, "16");
				props.setProperty(KEY_AUDIO_MAPPER_FPS, "20");
				props.setProperty(KEY_USE_EXTERNAL_OGG_DECODER, "1");
				props.setProperty(KEY_USE_EXTERNAL_OGG_ENCODER, "1");
				props.setProperty(KEY_MP3_DECODE_COMMAND, "mpg123 -w %of %if");
				props.setProperty(KEY_OGG_DECODE_COMMAND, "ogg123 -d wav -f %of %if");
				props.setProperty(KEY_OGG_ENCODE_COMMAND, "oggenc -b 320 -o %of %if");
				props.setProperty(KEY_LOAD_LAST_ON_BOOT, "YES");
				props.setProperty(KEY_WORKSPACE_PATH, projectsBaseDir);
				saveProps();
				}
			catch (IOException ex)
				{
				throw new GlobalPropertiesException("" + ex);
				}
			}


		private static void saveProps()
			{
			try
				{
				File rcFile = getRcFile();
				FileOutputStream fs = new FileOutputStream(rcFile);
				props.store(fs, null);
				fs.close();
				}
			catch (IOException ex1)
				{
				error("savePros","Cannot save " + getRcFile() + " : " + ex1);
				}
			}


		private static String chooseProjectdir()
			{
			File f = Popup.selectDir("Please choose a place as Projects directory");
			String dir = f.toString();
			PMESG(4, "Projects Repository : " + dir);
			return dir;
			}


		public static void set(String key, String value)
			{
			props.put(key, value);
			saveProps();
			}


		public static int getInt(String key)
			{
			int res = 0;
			try
				{
				res = Integer.parseInt(get(key));
				}
			catch (Exception ex)
				{
				ex.printStackTrace();
				}
			return res;
			}


		public static String get(String key)
			{
			return (String) props.get(key);
			}


		public static final File getRcFile()
			{
			return new File(getHome() + "/" + Const.RCFILE);
			}


		public static String getHome()
			{
			return userPath;
			}


		public static void setUserPath(String path)
			{
			userPath = path;
			}


		public static String getProjectsBaseDir()
			{
			return projectsBaseDir;
			}


		public static void setProjectPath(String projectPath)
			{
			projectsBaseDir = projectPath;
			}


		/*
		 * File fProjectPath = new File(projectsBaseDir); if (!fProjectPath.exists()) { PMESG(2, "Project Path does not exist. Creating " + projectsBaseDir.toString()); if (!fProjectPath.mkdirs()) { PERROR("Project Path could not be created !"); } else { Popup.warn("No protuxprojects directory under your home dir.\nI created one for you in " +
		 * projectsBaseDir.toString()); } }
		 */

		public static boolean getBoolean(String key)
			{
			String value = props.getProperty(key);
			return value.equalsIgnoreCase("YES") || value.equalsIgnoreCase("TRUE") || value.equalsIgnoreCase("ON") || value.equalsIgnoreCase("1");
			}


		public static void setLastProjectName(String newProjectName)
			{
			set(KEY_LAST_PROJECT, newProjectName);
			}


		public static String getLastProjectName()
			{
			return get(GlobalProperties.KEY_LAST_PROJECT);
			}

		private void log(String string)
		{
		PMESG(2, "[ GlobalProperties ] " + string);
		}

	private static void warn(String where, String what)
		{
		Debugger.warn("GlobalProperties",where,what);
		}


	private static void error(String where, String what)
		{
		Debugger.error("GlobalProperties",where,what);
		}


	
	}
