package org.protux.core;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import org.protux.core.editor.Editor;
import org.protux.core.editor.Song;
import org.protux.core.engines.audio.SoftBusMapper;

public class Project
	{
		public static final int MAX_SONGS_PER_PROJECT = 30;
		private int numSongs = 0;
		private String title;
		private String rootDir;
		private String engineer;
		private Song currentSong = null;
		public Song song[] = new Song[MAX_SONGS_PER_PROJECT];

		public Project(String title)
			{
			this.title = title;
			rootDir = GlobalProperties.getProjectsBaseDir() + "/" + this.title;
			}


		/** try to load the project by its title */
		public void load() throws IOException
			{
			this.numSongs = 1; // FIXME (must read how many songs are inside the project dir. if none, must create at least one)
			String propFile = rootDir + "/" + Const.PROJECT_PROPERTIES_FILE;
			Properties props = new Properties(propFile);
			engineer = props.getProperty("engineer");
			int currentSongId = props.getInt("currentSongId");
			Protux.info("PROJECT " + title + " LOADED");
			PMESG(2, "Loading project " + title + "  NumSongs=" + numSongs);
			Protux.info("Loading Project...");
			for (int i = 0; i < numSongs; i++)
				{
				int songNumber = i;
				song[i] = Factory.createOrOpenSong(this, songNumber);
				}
			setCurrentSong(currentSongId);
			}


		public boolean anyOtherSongGoing()
			{
			for (int i = 0; i < numSongs; i++)
				{
				Song xso = song[i];
				if (xso.softMapper.status == SoftBusMapper.GOING)
					{
					if (xso != currentSong)
						{
						return true;
						}
					}
				}
			return false;
			}


		public boolean has_changed()
			{
			for (int i = 0; i < numSongs; i++)
				{
				if (song[i].isChanged)
					return true;
				}
			return false;
			}


		public void addSong(Song x)
			{
			song[numSongs] = x;
			setCurrentSong(numSongs - 1);
			numSongs++;
			}


		public Song createSong(int rate, int bitDepth)
			{
			if (numSongs == MAX_SONGS_PER_PROJECT)
				{
				return null;
				}
			int songNumber = numSongs + 1;
			Song newSong = null;
			try
				{
				newSong = Factory.createOrOpenSong(this, songNumber);
				}
			catch (IOException ex)
				{
				System.out.println("Error : " + ex);
				}
			return newSong;
			}


		public void set_current_song(Song pSong)
			{
			currentSong = pSong;
			getEditor().setCurrentSong(currentSong);
			currentSong.update_properties();
			// MIG pvp.cursorManager.set_current_song(song[currentSongId]);
			// MIG pSong.update_properties();
			// MIG pSong.mta.recreate();
			updateInfo();
			}


		public void setCurrentSong(int id)
			{
			if ((id < 0) || (id >= numSongs))
				{
				PERROR("Internal Error : Trying to set current song : invalid index " + id);
				}
			else
				{
				set_current_song(song[id]);
				}
			}


		public Song get_current_song()
			{
			return currentSong;
			}


		public final Editor getEditor()
			{
			return Protux.editor;
			}


		public Song get_song(int id)
			{
			if ((id < 0) || (id > MAX_SONGS_PER_PROJECT))
				{
				return null;
				}
			return song[id];
			}


		public void save() throws IOException
			{
			saveProperties();
			for (int i = 0; i < numSongs; i++)
				{
				song[i].save();
				}
			Protux.info("Project Sucessfully Saved !");
			}


		public void saveProperties() throws IOException
			{
			PMESG(2, "Saving Properties for Project " + title);
			java.util.Properties props = new java.util.Properties();
			String propsFile = rootDir + "/" + Const.PROJECT_PROPERTIES_FILE;
			props.setProperty("numSongs", "" + numSongs);
			props.setProperty("title", title);
			props.setProperty("currentSongId", "" + currentSong.id);
			props.store(new FileOutputStream(propsFile), null);
			}


		public void resetProperties() throws IOException
			{
			PMESG(2, "Reseting Properties for Project " + title);
			java.util.Properties props = new java.util.Properties();
			String propsFile = getRootDir() + "/" + Const.PROJECT_PROPERTIES_FILE;
			props.setProperty("currentSongId", "0");
			FileOutputStream fs = new FileOutputStream(propsFile);
			props.store(fs, null);
			fs.close();
			}


		public String get_capture_dir()
			{
			return currentSong.captureDir;
			}


		public int remove_song(int id)
			{
			if (id >= 0 && id <= numSongs)
				{
				// MIG parentInterface.statusBarTimer.stop();
				Song s = song[id];
				for (int i = id; i < numSongs; i++)
					{
					song[i] = song[i + 1];
					}
				numSongs--;
				s = null;
				if (numSongs == 0)
					{
					// ??? createSong(); why ? cant I have an empty project ???
					}
				else if (currentSong.id == numSongs)
					{
					setCurrentSong(numSongs - 1);
					}
				else if (currentSong.id <= numSongs)
					{
					setCurrentSong(currentSong.id);
					}
				// MIG parentInterface.statusBarTimer.start(1000);
				return 1;
				}
			return -1;
			}


		public int rename()
			{
			return 0;
			}


		public int render()
			{
			// TODO
			return 0;
			}


		public void updateInfo()
			{
			Protux.infoBox.recreate();
			}


		public void import_audio(File file) throws IOException
			{
			getCurrentSong().importAudio(file);
			}


		public boolean changed()
			{
			for (int i = 0; i < numSongs; i++)
				{
				if (song[i].changed())
					return true;
				}
			return false;
			}


		// MIG void set_output_spectrogram(Spectrogram* pSpectrogram)
		// MIG {
		// MIG spectrogram=pSpectrogram;
		// MIG }

		public void recreate()
			{
			log("Recreating...");
			if (currentSong != null)
				{
				currentSong.recreateEditor();
				}
			Protux.recreateBuses();
			updateInfo();
			}


		private void log(String string)
			{
			PMESG(2, "[ Project ] " + string);
			}


		public void get_peak_load_lock()
			{
			// MIG mutex.lock();
			}


		public void release_peak_load_lock()
			{
			// MIG mutex.unlock();
			}


		public void setTitle(String pTitle)
			{
			title = pTitle;
			}


		public void setEngineer(String pEngineer)
			{
			engineer = pEngineer;
			}


		public String getTitle()
			{
			return this.title;
			}


		public String getEngineer()
			{
			return this.engineer;
			}


		public String getRootDir()
			{
			return this.rootDir;
			}


		public void setRootDir(String rootDir)
			{
			this.rootDir = rootDir;
			}


		public int getNumSongs()
			{
			return this.numSongs;
			}


		public void setNumSongs(int numSongs)
			{
			this.numSongs = numSongs;
			}


		public Song getCurrentSong()
			{
			return currentSong;
			}


		public int getCurrentSongId()
			{
			return currentSong.id;
			}


		public void setCurrentSongId(int currentSongId)
			{
			this.currentSong = song[currentSongId];
			}

	}
