package org.protux.core;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.SwingUtilities;
import javax.swing.ToolTipManager;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import org.protux.core.editor.Editor;
import org.protux.core.editor.Rack;
import org.protux.core.engines.audio.Bus;
import org.protux.core.engines.audio.HardBus;
import org.protux.core.engines.audio.HardBusContainer;
import org.protux.core.engines.audio.HardBusMapper;
import org.protux.core.engines.audio.SoftBus;
import org.protux.core.engines.audio.SoftBusContainer;
import org.protux.core.exceptions.GlobalPropertiesException;
import org.protux.core.gui.BusRack;
import org.protux.core.gui.ButtonBar;
import org.protux.core.gui.ColorManager;
import org.protux.core.gui.Help;
import org.protux.core.gui.InfoBox;
import org.protux.core.gui.Lcd;
import org.protux.core.gui.Popup;
import org.protux.core.gui.ProjectManager;
import org.protux.core.gui.ProtuxContainer;
import org.protux.core.gui.SplashScreen;
import org.protux.core.gui.SystemStatusBox;
import org.protux.core.gui.dialogs.GlobalPropertiesDialog;
import org.protux.input.jmb.Jmb;
import org.protux.input.jmb.JmbDispatcher;
import org.protux.input.jmb.JmbHandler;

public class Protux extends JPanel implements JmbHandler, JmbDispatcher, ActionListener
	{
		public static Protux appSingleton = null;
		private ProtuxContainer myContainer = null;
		private JFrame myFrame = null;
		private static boolean isLive = false;

		public static GlobalPropertiesDialog globalPropertiesDialog = null;
		public static SoftBusContainer softInsContainer = null;
		public static SoftBusContainer softOutsContainer = null;
		public static HardBusContainer hardInsContainer = null;
		public static HardBusContainer hardOutsContainer = null;
		public static SplashScreen splash = null;
		public static InfoBox infoBox = null;
		public static Lcd defaultLcd = null;
		public static Editor editor = null;
		public static Rack rack = null;

		private Hashtable<Integer, ArrayList<JmbHandler>> jmbHandlerHash = new Hashtable<Integer, ArrayList<JmbHandler>>();

		private SystemStatusBox statusBox = null;
		protected ButtonBar bottonButtonBar = null;
		private Help helpWindow = null;
		private ProjectManager projectManager = null;
		private Jmb jmb = null;
		private static String resourceDir = null;


		public Protux(ProtuxContainer parentContainer, String resourceDir, boolean isLive)
			{
			super();
			setNativeLookAndFeel();
			Protux.resourceDir = resourceDir;
			Protux.isLive = isLive;
			appSingleton = this;

			defaultLcd = new Lcd();
			softInsContainer = new SoftBusContainer("softInsContainer");
			softOutsContainer = new SoftBusContainer("softsOutContainer");
			hardInsContainer = new HardBusContainer("hardInsContainer");
			hardOutsContainer = new HardBusContainer("hardOutsContainer");
			infoBox = new InfoBox();
			editor = new Editor();
			rack = new Rack();
			globalPropertiesDialog = new GlobalPropertiesDialog();

			this.myContainer = parentContainer;
			this.myFrame = myContainer.getWindow();
			this.projectManager = ProjectManager.createManager();

			if (resourceDir != null)
				{
				helpWindow = new Help();
				splash = new SplashScreen();
				}

			// Protux Window settings
			if (myFrame != null)
				{
				this.jmb = new Jmb(myFrame, defaultLcd); // FIXME
				this.myFrame.setIconImage(java.awt.Toolkit.getDefaultToolkit().getImage("resources/images/protux.xpm"));
				ToolTipManager.sharedInstance().setEnabled(true);
				}

			// TOP Elements-----------------------
			JPanel topPane = new JPanel();
			topPane.setMinimumSize(Const.TOP_PANEL_SIZE);
			topPane.setPreferredSize(Const.TOP_PANEL_SIZE);
			topPane.setLayout(new BorderLayout());
			JPanel panelLeft = new JPanel();
			panelLeft.setLayout(new BorderLayout());
			panelLeft.setMinimumSize(new Dimension(300, Const.TOP_PANEL_HEIGHT));
			panelLeft.setPreferredSize(new Dimension(300, Const.TOP_PANEL_HEIGHT));
			panelLeft.add(defaultLcd, BorderLayout.NORTH);
			panelLeft.add(infoBox, BorderLayout.CENTER);
			BusRack busTabbedPane = new BusRack();
			busTabbedPane.add("Soft IN", softInsContainer);
			busTabbedPane.add("Soft OUT", softOutsContainer);
			busTabbedPane.add("Hard IN", hardInsContainer);
			busTabbedPane.add("Hard OUT", hardOutsContainer);
			topPane.add(panelLeft, BorderLayout.WEST);
			topPane.add(busTabbedPane, BorderLayout.CENTER);

			// MIDDLE ELEMENTS
			JSplitPane middleSplitter = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
			middleSplitter.setDividerLocation(0.5);
			middleSplitter.setDividerSize(3);
			middleSplitter.setLeftComponent(rack);
			middleSplitter.setRightComponent(editor);

			// BOTTON ELEMENTS 
			JPanel bottonPanel = new JPanel();
			bottonPanel.setLayout(new GridLayout(1, 2));
			bottonButtonBar = new ButtonBar();
			statusBox = new SystemStatusBox();
			bottonPanel.add(bottonButtonBar);
			bottonPanel.add(statusBox);

			// SHOW GUI
			JSplitPane topSplitter = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
			topSplitter.setDividerSize(3);
			topSplitter.setTopComponent(topPane);
			topSplitter.setBottomComponent(middleSplitter);
			this.setLayout(new BorderLayout());
			this.add(topSplitter, BorderLayout.CENTER);
			this.add(bottonPanel, BorderLayout.SOUTH);

			// register JMB handlers
			addActionHandler(Const.ACTION_QUIT, this);
			addActionHandler(Const.ACTION_SHOW_PROJECT_MANAGER, this);
			addActionHandler(Const.ACTION_IMPORT_AUDIO, this);
			addActionHandler(Const.ACTION_SAVE_PROJECT, this);
			addActionHandler(Const.ACTION_ADD_SOFT_IN, this);
			addActionHandler(Const.ACTION_ADD_SOFT_OUT, this);
			addActionHandler(Const.ACTION_TOGGLE_PLAY_STOP, this);
			addActionHandler(Const.ACTION_ZOOM_IN, editor.getJmbHandler());
			addActionHandler(Const.ACTION_ZOOM_OUT, editor.getJmbHandler());
			addActionHandler(Const.ACTION_TOGGLE_PLAY_STOP, editor.getJmbHandler());
			addActionHandler(Const.ACTION_ADD_TRACK, editor.getJmbHandler());
			addActionHandler(Const.ACTION_REMOVE_ACTIVE_TRACK, editor.getJmbHandler());
			}


		public void init()
			{
			try
				{
				GlobalProperties.init();
				initIcons();
				ColorManager.init();
				HardBusMapper.init();
				hardInsContainer.addBuses(HardBusMapper.getHardIns());
				hardOutsContainer.addBuses(HardBusMapper.getHardOuts());
				HardBus defaultHardIn = (HardBus) hardInsContainer.getDefaultBus();
				HardBus defaultHardOut = (HardBus) hardOutsContainer.getDefaultBus();
				SoftBus defaultSoftIn = softInsContainer.createBus("Default Soft In", Bus.Direction.IN);
				SoftBus defaultSoftOut = softOutsContainer.createBus("Default Soft Out", Bus.Direction.OUT);
				defaultSoftIn.setHardBus(defaultHardIn);
				defaultSoftOut.setHardBus(defaultHardOut);
				PluginManager.init();
				if (myFrame != null)
					{
					myFrame.pack();
					myFrame.setSize(Const.MIN_PROTUX_SIZE);
					myFrame.setMinimumSize(Const.MIN_PROTUX_SIZE);
					//myFrame.setExtendedState(Frame.MAXIMIZED_BOTH);
					myFrame.setVisible(true);
					}

				// UNCO splash.show();
				defaultLcd.rollMessage("Welcome to Protux 0.50 - Press H for Quick Help  ", 20, 1);
				statusBox.activate();
				if (ProjectManager.numProjects() == 0)
					{
					ProjectManager.createFirstProject();
					}
				else
					loadLastProject();
				editor.startPeakChecker();
				}
			catch (GlobalPropertiesException ex)
				{
				error("init", "Global Properties could not be initilized : " + ex);
				}
			catch (IOException ex)
				{
				warn("init", "Last project could not be loaded.");
				}
			catch (Exception e)
				{
				error("init", "Unexpected Exception : " + e);
				e.printStackTrace();
				}
			}


		@Override
		public void finalize()
			{
			}


		public void globalPropertiesChanged()
			{
			if (jmb != null)
				{
				jmb.set_clear_time(GlobalProperties.getInt(GlobalProperties.KEY_JMB_CLEAR_TIME));
				jmb.set_hold_sensitiveness(GlobalProperties.getInt(GlobalProperties.KEY_JMB_HOLD_TIMEOUT));
				jmb.set_double_fact_interval(GlobalProperties.getInt(GlobalProperties.KEY_JMB_DOUBLE_FACT_TIMEOUT));
				}
			HardBusMapper.probeValidModes();
			Project cp = getCurrentProject();
			if (cp != null)
				{
				cp.get_current_song().recreateEditor();
				}
			}


		public static Project getCurrentProject()
			{
			return ProjectManager.getCurrentProject();
			}


		public void setCurrentProject(Project project)
			{
			projectManager.setCurrentProject(project);
			}


		public ProjectManager getProjectManager()
			{
			return projectManager;
			}


		private void loadLastProject() throws IOException
			{
			boolean loadProjectAtStartUp = GlobalProperties.getBoolean(GlobalProperties.KEY_LOAD_LAST_ON_BOOT);
			if (loadProjectAtStartUp)
				{
				String lastProjectName = GlobalProperties.getLastProjectName();
				if ((lastProjectName != null) && (!lastProjectName.equals("")))
					{
					try
						{
						Project proj = projectManager.loadProject(lastProjectName);
						proj.recreate();
						}
					catch (IOException ex)
						{
						info("Cannot load project. Check logs.");
						error("loadLastProject", "Project indicated to be autoload in .protuxrc (" + lastProjectName + ") could not be loaded : " + ex);
						throw ex;
						}
					}
				}
			}


		public static void info(String message)
			{
			if (appSingleton != null)
				{
				if (defaultLcd != null)
					{
					defaultLcd.displayMainMessage(message);
					}
				}
			}


		public void set_default_lcd(Lcd lcd)
			{
			defaultLcd = lcd;
			}


		public Lcd get_default_lcd()
			{
			return defaultLcd;
			}


		public boolean closed()
			{
			if (Popup.confirm("Are you sure you want to quit Protux ?"))
				{
				PMESG(1, "Thank you for using Protux ! Visit our web site for more information http://www.nongnu.org/protux\n");
				System.exit(0);
				}
			return false;
			}


		// ---------------------------------------------------------------------------------------
		// METHODS WHICH ARE BEING REVIEWED ------------------------------------------------------
		// ---------------------------------------------------------------------------------------
		public void show_tuner()
			{
			/*
			 * MIG if (tuner != null) { tuner = null; } else { tuner = new Tuner(1, true); tuner.init_frontend(); tuner.show(); }
			 */
			}


		public void lock_everything()
			{
			if (jmb != null)
				{
				defaultLcd.displayMainMessage("LOCK/UNLOCK EVERYTHING");
				if (jmb.is_locked())
					{
					jmb.unlock();
					}
				else
					{
					jmb.lock();
					}
				}
			}


		public void exit()
			{
			if (ProjectManager.getCurrentProject() != null)
				{
				try
					{
					ProjectManager.getCurrentProject().save();
					}
				catch (IOException ex)
					{
					System.out.println("Error : " + ex);
					}
				}
			System.exit(0);
			}


		public void full_screen()
			{
			/*
			 * MIG if (isFullScreen()) { showNormal(); } else { showFullScreen(); }
			 */
			}


		public void quick_help()
			{
			}


		public void show_jmb_map()
			{
			}


		public void about_protux()
			{
			defaultLcd.displayMainMessage("ABOUT PROTUX");
			String authors = "The Protux Team : Luciano Giordana, Reinhard Amersberger, Tapio Laxstrom, David Bain, Nathan Langford, Fabio dos Santos, Martin Herren, Eduardo Pimenta, Valery Beluntsov, Marcio Enrico, Remon Sijrier, Gustavo Lemos, Alexandre beraldo, and all the people from Free Software world, who made important technologies on which Protux is based ( Gcc, QT, KDE, XFree, Linux, and so on...)                                              ";
			defaultLcd.rollMessage(authors, 80, 1);
			show_splash_screen();
			}


		@Deprecated
		private void resizeEvent()
			{
			if (ProjectManager.getCurrentProject() != null)
				{
				ProjectManager.getCurrentProject().recreate();
				}
			}


		void show_splash_screen()
			{
			// MIG splashScreen = new SplashScreen(RESOURCES_DIR + "/protuxSplash.png");
			}


		public void importAudio()
			{
			File file = Popup.selectFile("Import Audio");
			try
				{
				ProjectManager.getCurrentProject().import_audio(file);
				}
			catch (IOException ex)
				{
				PERROR(this, "Error : " + ex);
				}
			}


		public void saveProject()
			{
			try
				{
				ProjectManager.getCurrentProject().save();
				}
			catch (IOException ex)
				{
				Debugger.error("Protux","saveProject","Error : " + ex);
				}
			}


		// -------------------------------------------------------
		// ACTION HANDLING (JMB AND CLASSIC)
		// -------------------------------------------------------

		public void dispatchJmbAction(int commandIndex)
			{
			ArrayList<JmbHandler> jhlist = getHandlerList(commandIndex);
			if (jhlist == null)
				error("dispatchJmbAction", "No Jmb Handlers for Action " + commandIndex);
			else
				{
				for (JmbHandler jh : jhlist)
					jh.execJmbAction(commandIndex);
				}
			}


		public void dispatchClassicAction(int commandIndex)
			{
			dispatchJmbAction(commandIndex);
			}


		public void addActionHandler(int commandIndex, JmbHandler jmbHandler)
			{
			ArrayList<JmbHandler> handlerList = getHandlerList(commandIndex);
			if (handlerList == null)
				{
				handlerList = new ArrayList<JmbHandler>();
				jmbHandlerHash.put(new Integer(commandIndex), handlerList);
				}
			handlerList.add(jmbHandler);
			}


		public ArrayList<JmbHandler> getHandlerList(int commandIndex)
			{
			return jmbHandlerHash.get(commandIndex);
			}


		public void execJmbAction(int commandIndex)
			{
			switch (commandIndex)
				{
				case Const.ACTION_SHOW_PROJECT_MANAGER:
					ProjectManager.popup();
					break;
				case Const.ACTION_QUIT:
					if (closed())
						{
						// ?
						}
					break;
				case Const.ACTION_IMPORT_AUDIO:
					importAudio();
					break;
				case Const.ACTION_SAVE_PROJECT:
					saveProject();
					break;
				case Const.ACTION_ADD_SOFT_IN:
					softInsContainer.createBus("New In", Bus.Direction.IN);
					break;
				case Const.ACTION_ADD_SOFT_OUT:
					softOutsContainer.createBus("New Out", Bus.Direction.OUT);
					break;
				case Const.ACTION_TOGGLE_PLAY_STOP:
					if (editor.isPlaying())
						bottonButtonBar.buttonTogglePlay.setText(Const.LABEL_PLAY);
					else
						bottonButtonBar.buttonTogglePlay.setText(Const.LABEL_STOP);
					break;
				}
			}


		public static SoftBus getDefaultSoftBusIn()
			{
			return (SoftBus) softInsContainer.getDefaultBus();
			}


		public static SoftBus getDefaultSoftBusOut()
			{
			return (SoftBus) softOutsContainer.getDefaultBus();
			}


		public void actionPerformed(ActionEvent e)
			{
			if (e.getSource() == bottonButtonBar.buttonTogglePlay)
				{
				dispatchClassicAction(Const.ACTION_TOGGLE_PLAY_STOP);
				}
			}


		private void log(String string)
			{
			PMESG(2, "[ Protux ] " + string);
			}


		private void warn(String where, String what)
			{
			Debugger.warn("Protux", where, what);
			}


		private void error(String where, String what)
			{
			Debugger.error("Protux", where, what);
			}


		public static void recreateBuses()
			{
			softOutsContainer.recreateBuses();
			hardOutsContainer.recreateBuses();
			softInsContainer.recreateBuses();
			hardInsContainer.recreateBuses();
			}


		public static String getResourcesDir()
			{
			return resourceDir;
			}


		private static void initIcons()
			{
			if (resourceDir != null)
				{
				Toolkit tk = java.awt.Toolkit.getDefaultToolkit();
				iconHash.put("ledOn", new ImageIcon(tk.getImage(resourceDir + "/images/ledOn.jpg")));
				iconHash.put("ledOff", new ImageIcon(tk.getImage(resourceDir + "/images/ledOff.jpg")));
				}
			}

		public static Hashtable<String, Icon> iconHash = new Hashtable<String, Icon>();


		public static Icon getIcon(String key)
			{
			return iconHash.get(key);
			}


		public static final boolean isLive()
			{
			return Protux.isLive;
			}


		private static void setNativeLookAndFeel()
			{
			try
				{
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
				}
			catch (Exception e)
				{
				Debugger.error("Protux","setNativeLookAndFeel()","Error setting Look&Feel : " + e);
				}
			}

	}
