package org.protux.core;

import static org.protux.core.Debugger.PWARN;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;

public class SchemaParser
	{
	/** return top level sections of a "kind-of-xml" file */
	public static ArrayList<String> parse(String filename) throws IOException
		{
		BufferedReader input = null;
		ArrayList<String> topLevelSectionList = new ArrayList<String>();
		try
			{
			File aFile = new File(filename);
			input = new BufferedReader(new FileReader(aFile));
			String line = null; //not declared within while loop
			while ((line = input.readLine()) != null)
				{
				int x = line.indexOf("<");
				if (x >= 0)
					{
					StringBuffer topLevelSection = new StringBuffer(line + "\n");
					int y1 = line.indexOf(">", x + 1);
					int y2 = line.indexOf(" ", x + 1);
					int y = (y1 < 0) ? y2 : ((y2 < 0) ? y1 : ((y1 < y2) ? y1 : y2));
					String key = line.substring(x + 1, y);
					while (((line = input.readLine()) != null) && (line.indexOf("</" + key + ">") < 0))
						{
						topLevelSection.append(line + "\n");
						}
					topLevelSection.append(line + "\n");
					topLevelSectionList.add(topLevelSection.toString());
					}
				}
			}
		catch (FileNotFoundException ex)
			{
			PWARN("Help: File " + filename + " not found!");
			}
		catch (IOException ex)
			{
			PWARN("Help: File " + filename + " cannot be read!");
			}
		catch (Exception ex)
			{
			PWARN("Help: File " + filename + " cannot be read!");
			}
		finally
			{
			try
				{
				if (input != null)
					{
					input.close();
					}
				}
			catch (Exception ex)
				{
				//PERROR("File Not Found "+ex);
				}
			}
		return topLevelSectionList;
		}



	/** return the schema string for a given section name from a list of many sections */
	public static String getSection(ArrayList<String> sectionList, String sectionName)
		{
		for (int i = 0; i < sectionList.size(); i++)
			{
			String src = sectionList.get(i);
			int x = src.indexOf("<" + sectionName);
			if (x >= 0)
				{
				return src;
				}
			}
		return null;
		}



	public static ArrayList<String> getSections(String src, String sectionName)
		{
		ArrayList<String> list = new ArrayList<String>();
		try
			{
			BufferedReader input = new BufferedReader(new StringReader(src));
			String line = null; //not declared within while loop
			while ((line = input.readLine()) != null)
				{
				int x1 = line.indexOf("<" + sectionName + " ");
				int x2 = line.indexOf("<" + sectionName + ">");
				int x = (x1 < 0) ? ((x2 < 0) ? -1 : x2) : x1;
				if (x >= 0)
					{
					StringBuffer sb = new StringBuffer(line + "\n");
					while (((line = input.readLine()) != null) && (line.indexOf("</" + sectionName + ">") < 0))
						{
						sb.append(line + "\n");
						}
					sb.append(line + "\n");
					list.add(sb.toString());
					}
				}
			}
		catch (IOException ex)
			{
			System.out.println("Error : " + ex);
			}
		return list;
		}



	public static boolean getXmlBooleanProperty(String src, String key)
		{
		String r = getXmlProperty(src, key);
		return (r.equalsIgnoreCase("1") || r.equalsIgnoreCase("true") || r.equalsIgnoreCase("on"));
		}



	public static float getXmlFloatProperty(String src, String key)
		{
		String r = getXmlProperty(src, key);
		float f = 0;
		try
			{
			f = Float.parseFloat(r);
			}
		catch (Exception ex)
			{
			}
		return f;
		}



	public static int getXmlIntProperty(String src, String key)
		{
		String r = getXmlProperty(src, key);
		int i = 0;
		try
			{
			i = Integer.parseInt(r);
			}
		catch (Exception ex)
			{
			}
		return i;
		}



	public static long getXmlLongProperty(String src, String key)
		{
		String r = getXmlProperty(src, key);
		long l = 0;
		try
			{
			l = Long.parseLong(r);
			}
		catch (Exception ex)
			{
			}
		return l;
		}



	public static String getXmlProperty(String src, String key)
		{
		int a1 = src.indexOf(key + "=");
		int a2 = src.indexOf(key + "=\"");
		int a = (a2 < 0) ? a1 + key.length() + 1 : a2 + key.length() + 2;
		int b1 = src.indexOf("\"", a);
		int b2 = src.indexOf(" ", a);
		int b3 = src.indexOf(">", a);
		int b = (b1 < 0) ? (( b2 <0) ? b3 : b2) : b1;
		String r = src.substring(a, b);
		return r;
		}



	public static void saveXml(File filename, String src) throws IOException
		{
		Util.saveContents(filename, src);
		}

	}
