package org.protux.core;

import static org.protux.core.Debugger.PWARN;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

public class Util
	{
	/**
	 * Fetch the entire contents of a text file, and return it in a String.
	 * This style of implementation does not throw Exceptions to the caller.
	 *
	 * @param aFile is a file which already exists and can be read.
	 */
	public static String getContents(String fn)
		{
		BufferedReader input = null;
		StringBuffer contents = null;
		try
			{
			File aFile = new File(fn);
			input = new BufferedReader(new FileReader(aFile));
			String line = null; //not declared within while loop
			contents = new StringBuffer();
			while ((line = input.readLine()) != null)
				{
				contents.append(line);
				contents.append("\n");
				}
			}
		catch (FileNotFoundException ex)
			{
			PWARN("Help: File " + fn + " not found!");
			}
		catch (IOException ex)
			{
			PWARN("Help: File " + fn + " cannot be read!");
			}
		catch (Exception ex)
			{
			PWARN("Help: File " + fn + " cannot be read!");
			}
		finally
			{
			try
				{
				if (input != null)
					{
					input.close();
					}
				}
			catch (Exception ex)
				{
				return "File Not Found";
				}
			}
		if (contents != null)
			{
			return contents.toString();
			}
		return null;
		}


	public static String saveContents(File fn, String data)
		{
		BufferedWriter output = null;
		StringBuffer contents = null;
		try
			{
			output = new BufferedWriter(new FileWriter(fn));
			output.write(data);
			}
		catch (FileNotFoundException ex)
			{
			PWARN("Help: File " + fn + " not found!");
			}
		catch (IOException ex)
			{
			PWARN("Help: File " + fn + " cannot be read!");
			}
		catch (Exception ex)
			{
			PWARN("Help: File " + fn + " cannot be read!");
			}
		finally
			{
			try
				{
				if (output != null)
					{
					output.close();
					}
				}
			catch (Exception ex)
				{
				return "File Not Found";
				}
			}
		if (contents != null)
			{
			return contents.toString();
			}
		return null;
		}




	/** takes a string , replace all /t by " " (spaces) ,trim it and truncate sequentail spaces */
	public static String cleanSpaces(String s)
		{
		String x = s.replaceAll("\t", " ");
		x = x.trim();
		return x;
		}



	public static int toInt(String s)
		{
		int x;
		try
			{
			x = Integer.parseInt(s);
			}
		catch (Exception e)
			{
			x = 0;
			}
		return x;
		}



	public static long toLong(String s)
		{
		long x;
		try
			{
			x = Long.parseLong(s);
			}
		catch (Exception e)
			{
			x = 0;
			}
		return x;
		}



	public static String baseName(String name)
		{
		return name.substring(0, name.indexOf("."));
		}


	// This is for VU : db = 20 * log ( sample / MaxSample )
	public static float dB_to_scale_factor(float dB)
		{
		// examples :
		// dB = 0    will return 1.0
		// db = -6.0 will return 0.5
		// db = -inf will return 0.0
		return dB > -120.0f ? (float) Math.pow(10.0f, dB * 0.05f) : 0.0f;
		}


	/** converts 0.0-1.0 norm peak value to the dB factor (-120.0 ... 0)
	 * examples :
	 * coeff = 1.0 will return 0 dB
	 * coeff = 0.5 will return -6 dB
	 * coeff = 0.0 will return -120 dB (which represents -infinite)
	 * @param coeff
	 * @return
	 */
	public static float coefficient_to_dB(float coeff)
		{
		if (coeff < 0.000001f) //Should be (coeff == 0), but this will do...
			{
			return (-120.0f); //Should be minus infinity, but it will do for busMonitor purposes
			}
		return 20.0f * (float) Math.log10(coeff);
		}

	public static String blockToSmpte(long bpos, int sampleRate)
		{
		float f = sampleRate / 30;
		float flo = bpos / f;
		long framepos = (long) flo;
		int pfr = (int) framepos % 30;
		int pho = (int) (framepos / 108000) % 60;
		int pmi = (int) (framepos / 1800) % 60;
		int pse = (int) (framepos / 30) % 60;
		if (pho < 0)
			pho = 0;
		if (pmi < 0)
			pmi = 0;
		if (pse < 0)
			pse = 0;
		if (pfr < 0)
			pfr = 0;
		String spho = "" + pho;
		String spmi = "" + pmi;
		if (pmi < 10)
			spmi = "0" + spmi;
		String spse = "" + pse;
		if (pse < 10)
			spse = "0" + spse;
		String spfr = "" + pfr;
		if (pfr < 10)
			spfr = "0" + spfr;
		String spos = spho + ":" + spmi + ":" + spse + "," + spfr;
		return spos;
		}



	}
