package org.protux.core.editor;

import static org.protux.core.Debugger.PMESG;
import static org.protux.core.gui.ColorManager.CM_CLIP_BG;
import static org.protux.core.gui.ColorManager.CM_CLIP_BG_ACTIVE;
import static org.protux.core.gui.ColorManager.CM_CLIP_BG_MUTED;
import static org.protux.core.gui.ColorManager.CM_CLIP_BG_SELECTED;
import static org.protux.core.gui.ColorManager.CM_CLIP_INFO_AREA_BG;
import static org.protux.core.gui.ColorManager.CM_CLIP_INFO_AREA_BG_ACTIVE;
import static org.protux.core.gui.ColorManager.CM_CLIP_PEAK_MACROVIEW;
import static org.protux.core.gui.ColorManager.CM_CLIP_PEAK_MICROVIEW;
import static org.protux.core.gui.ColorManager.CM_CLIP_PEAK_OVERLOADED_SAMPLE;
import static org.protux.core.gui.ColorManager.CM_DARK_TEXT;
import static org.protux.core.gui.ColorManager.CM_MAGENTA;
import static org.protux.core.gui.ColorManager.CM_TRACK_BG;
import static org.protux.core.gui.ColorManager.CM_TRACK_BG_ACTIVE;
import static org.protux.core.gui.ColorManager.getCmColor;
import static org.protux.core.gui.ColorManager.setCmBlur;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.font.FontRenderContext;
import java.awt.font.LineMetrics;
import java.awt.geom.Rectangle2D;

import org.protux.core.Const;
import org.protux.core.gui.Rectangle;
import org.protux.core.media.audio.AudioMedia;

public class AudioClip
	{
		// ---------- move these static coloes to CM
		public static final Color PEAK_MACRO_COLOR_UNDER_RIGHT_CROSS = new Color(0, 80, 100);
		public static final Color BG_COLOR_UNDER_RIGHT_CROSS_REGION = new Color(160, 180, 170);
		public static final Color CM_CLIP_AXIS = new Color(150, 204, 208);

		// and these are proxies..
		public static final Color COLOR_OVERLOAD = getCmColor(CM_CLIP_PEAK_OVERLOADED_SAMPLE);
		public static final Color PEAK_MICRO_COLOR = getCmColor(CM_CLIP_PEAK_MICROVIEW);
		public static final Color PEAK_MACRO_COLOR = getCmColor(CM_CLIP_PEAK_MACROVIEW);
		public static final Color CLIP_INFO_TEXT_COLOR = Color.BLACK;

		public static int lastId = 0;

		public String name;
		public int id;
		public boolean isMuted = false;
		public boolean isTake = false;
		public boolean isSelected = false;
		public long fadeOuBlocks = 0;
		public long fadeInBlocks = 0;
		public long sourceFirstBlock;
		public long sourceLastBlock;
		public long trackFirstBlock;
		public long trackLastBlock;
		public int channels;
		public boolean peaksDrawn = false;
		public Track parentTrack;
		public Song proxySong;
		public Editor proxyEditor;
		public AudioMedia audioSource;
		public AudioClip prev;
		public AudioClip next;
		private short[] peakBuffer;
		public float gain = 1.0f;
		private int peaksFilled;


		public AudioClip(int id, Track track, AudioMedia pAudioSource, long pSourceFirstBlock, long lenght, long pTrackInsertBlock)
			{
			this.parentTrack = track;
			this.proxySong = parentTrack.parentSong;
			this.proxyEditor = proxySong.proxyEditor;
			this.audioSource = pAudioSource;
			this.name = "NoName";
			this.id = id;
			this.channels = audioSource.channels;
			this.sourceFirstBlock = pSourceFirstBlock;
			this.sourceLastBlock = sourceFirstBlock + lenght;
			this.trackFirstBlock = pTrackInsertBlock;
			updateTrackLastBlockProxy();
			}


		public final void draw()
			{
			Graphics2D painter = proxyEditor.getGBuffer();
			clear(painter);
			drawAxis(painter);
			drawPeaks(painter);
			drawFades(painter);
			drawInfo(painter);
			drawContours(painter);
			}


		private final void clear(Graphics2D painter)
			{
			final int visibleFirstX = this.getVisibleFirstX();
			final int by = this.getBaseY();
			final int h = this.getHeight();
			final int w = getVisibleWidthX();
			final int EW = Const.CLIP_INFO_AREA_HEIGHT;
			int lCrossX = doesCrossLeft() ? calcCrossLeftX(getRealFirstX()) : 0;
			int rCrossX = doesCrossRight() ? calcCrossRightX(getRealLastX()) : 0;
			// Crear clip info area bg
			painter.setColor(parentTrack.isActive() ? getCmColor(CM_CLIP_INFO_AREA_BG_ACTIVE) : getCmColor(CM_CLIP_INFO_AREA_BG));
			painter.fillRect(visibleFirstX, by, w, EW);
			// main bg (except under cross regions)
			painter.setColor(isMuted ? getCmColor(CM_CLIP_BG_MUTED) : (isSelected ? getCmColor(CM_CLIP_BG_SELECTED) : getCmColor(CM_CLIP_BG)));
			painter.fillRect(visibleFirstX + lCrossX, by + Const.CLIP_INFO_AREA_HEIGHT, w - rCrossX - lCrossX, h - EW);
			// bg under RIGHT crossfade region ( only RIGHT !)
			painter.setColor(BG_COLOR_UNDER_RIGHT_CROSS_REGION);
			painter.fillRect(visibleFirstX + w - rCrossX, by + 16, rCrossX, h - 16);
			}


		private final void drawAxis(Graphics2D painter)
			{
			final int fvx = getVisibleFirstX();
			final int lvx = getVisibleLastX();
			final int by = getBaseY();
			final int h = getHeight();
			final int channelFullHeight = (h - Const.CLIP_INFO_AREA_HEIGHT) / channels;
			final int channelHalfHeight = channelFullHeight / 2;
			painter.setColor(CM_CLIP_AXIS);
			for (int k = 0; k < channels; k++)
				{
				int yc1 = channelHalfHeight + Const.CLIP_INFO_AREA_HEIGHT + channelFullHeight * k;
				final int axisY = by + yc1;
				painter.drawLine(fvx, axisY, lvx, axisY);
				}
			}


		private void drawPeaks(Graphics2D painter)
			{
			final int hzoom = proxySong.hzoom;
			final int visibleFirstX = this.getVisibleFirstX();
			final int currentBaseY = this.getBaseY();
			final int currentHeight = this.getHeight();
			final int firstRealX = getRealFirstX();
			long firstBlockOffset = firstRealX < 0 ? Math.abs(firstRealX) * Const.ZOOM_STEP[hzoom] : 0;
			int numBlocksToRead = getVisibleWidthX();
			Peak pk = audioSource.getPeak();
			if (pk != null)
				{
				int status = pk.getBuildStatus(hzoom);
				if (status == Peak.BLOCKED)
					{
					log("Peak is blocked.");
					peaksDrawn = true;
					painter.setColor(Color.BLACK);
					painter.drawString("PEAKS BLOCKED!", visibleFirstX + 10, currentBaseY + Const.CLIP_INFO_AREA_HEIGHT + currentHeight / 3);
					}
				else
					{
					if ((status == Peak.BUILDING) || (status == Peak.NOT_AVAILABLE))
						{
						int progress = pk.getBuildingProgress(hzoom);
						// TODO draw a small clock with an little peak drawing inside the clip...
						painter.setColor(Color.BLACK);
						painter.drawString("Building peaks (" + progress + "%)", visibleFirstX + 10, currentBaseY + Const.CLIP_INFO_AREA_HEIGHT + currentHeight / 3);
						}
					else
						{
						int totalSamples = numBlocksToRead * channels;
						if ((peakBuffer == null) || (peakBuffer.length != totalSamples)) // just instantiate in case it changes its size
							{
							peakBuffer = new short[totalSamples];
							}
						peaksFilled = pk.fillBuffer(peakBuffer, sourceFirstBlock + firstBlockOffset, numBlocksToRead, hzoom);
						if (peaksFilled > 0)
							{
							peaksDrawn = false;
							if (hzoom > Const.MAX_ZOOM_USING_SOURCEFILE)
								drawMacro(painter, pk);
							else
								drawMicro(painter, pk);
							peaksDrawn = true;
							}
						}
					}
				}
			}


		private final void drawMicro(Graphics2D painter, Peak pk)
			{
			final int visibleFirstX = this.getVisibleFirstX();
			final int channelFullHeight = (getHeight() - Const.CLIP_INFO_AREA_HEIGHT) / channels;
			final int channelHalfHeight = channelFullHeight / 2;
			final int by = this.getBaseY();
			final float gainFactor = gain / Const.MAX_SAMPLE;
			float fadeInFactor;
			float fadeOutFactor;
			float normPeak;
			int fadeInLastX = calcFadeInLastX();
			int fadeOutFirstX = calcFadeOutFirstX();
			int rCrossX = doesCrossRight() ? calcCrossRightX(getRealLastX()) : 0;
			int maxX = visibleFirstX + peaksFilled - rCrossX;
			for (int k = 0; k < channels; k++)
				{
				int ppos = 0;
				int basePeakY = by + Const.CLIP_INFO_AREA_HEIGHT + (channelFullHeight * k) + channelHalfHeight;
				int lasty = basePeakY;
				int y = basePeakY;
				painter.setColor(PEAK_MICRO_COLOR);
				for (int x = visibleFirstX; x < maxX; x++, ppos += channels)
					{
					fadeInFactor = ((fadeInBlocks == 0) || (x >= fadeInLastX)) ? 1.0f : (float) ((x - visibleFirstX) / (fadeInLastX - visibleFirstX));
					fadeOutFactor = ((fadeOuBlocks == 0) || (x <= fadeOutFirstX)) ? 1.0f : 1.0f - (x - fadeOutFirstX) / (visibleFirstX + peaksFilled - fadeOutFirstX);
					normPeak = gainFactor * peakBuffer[ppos] * fadeOutFactor * fadeInFactor;
					int dy = (int) (channelFullHeight * normPeak);
					lasty = y;
					y = basePeakY - dy;
					if (dy == channelFullHeight)
						{
						painter.setColor(COLOR_OVERLOAD);
						painter.drawLine(x - 1, lasty, x, y);
						painter.setColor(PEAK_MICRO_COLOR);
						}
					else
						painter.drawLine(x - 1, lasty, x, y);
					}
				}
			}


		private final int calcFadeOutFirstX()
			{
			int fadeOutFirstX = getRealLastX();
			if (fadeOuBlocks > 0)
				fadeOutFirstX = proxySong.blockToEditorX(trackLastBlock - fadeOuBlocks);
			return fadeOutFirstX;
			}


		private final int calcFadeInLastX()
			{
			int fadeInLastX = getRealFirstX();
			if (fadeInBlocks > 0)
				fadeInLastX = proxySong.blockToEditorX(trackFirstBlock + fadeInBlocks);
			return fadeInLastX;
			}


		private final void drawMacro(Graphics2D painter, Peak pk)
			{
			int fadeInLastX = getRealFirstX();
			int fadeOutFirstX = getRealLastX();
			if (fadeInBlocks > 0)
				fadeInLastX = proxySong.blockToEditorX(trackFirstBlock + fadeInBlocks);
			if (fadeOuBlocks > 0)
				fadeOutFirstX = proxySong.blockToEditorX(trackLastBlock - fadeOuBlocks);
			final int bx = this.getVisibleFirstX();
			final int by = this.getBaseY();
			final int height = this.getHeight();
			final int channelFullHeight = (height - Const.CLIP_INFO_AREA_HEIGHT) / channels;
			final float gainFactor = gain / Const.MAX_SAMPLE;
			float fadeInFactor;
			float fadeOutFactor;
			float normPeak;
			int lastX = bx + peaksFilled;
			int rCrossX = doesCrossRight() ? calcCrossRightX(getRealLastX()) : 0;
			final int lastXBeforeRightCross = lastX - rCrossX;
			for (int k = 0; k < channels; k++)
				{
				int basePeakY = by + Const.CLIP_INFO_AREA_HEIGHT + (channelFullHeight * (k + 1));
				painter.setColor(PEAK_MACRO_COLOR);
				for (int x = bx, ppos = 0; x < lastX; x++, ppos += channels)
					{
					if (x == lastXBeforeRightCross)
						painter.setColor(PEAK_MACRO_COLOR_UNDER_RIGHT_CROSS);
					fadeInFactor = ((fadeInBlocks == 0) || (x >= fadeInLastX)) ? 1.0f : (float) ((x - bx) / (fadeInLastX - bx));
					fadeOutFactor = ((fadeOuBlocks == 0) || (x <= fadeOutFirstX)) ? 1.0f : 1.0f - (x - fadeOutFirstX) / (bx + peaksFilled - fadeOutFirstX);
					normPeak = gainFactor * Math.abs(peakBuffer[ppos]) * fadeOutFactor * fadeInFactor;
					int dy = (int) (channelFullHeight * normPeak);
					if (dy == channelFullHeight)
						{
						painter.setColor(COLOR_OVERLOAD);
						painter.drawLine(x, basePeakY, x, basePeakY - dy);
						if (x > lastXBeforeRightCross)
							painter.setColor(PEAK_MACRO_COLOR_UNDER_RIGHT_CROSS);
						else
							painter.setColor(PEAK_MACRO_COLOR);
						}
					else
						painter.drawLine(x, basePeakY, x, basePeakY - dy);
					}
				}

			}


		private void drawInfo(Graphics2D painter)
			{
			final int currentBaseX = this.getVisibleFirstX();
			final int currentBaseY = this.getBaseY();
			String clipInfo = "" + id + ":" + name + " " + audioSource.rate + " " + audioSource.bitDepth + " G:" + gain + " " + (isMuted ? "M" : "");
			String fullClipInfo = clipInfo + " [" + trackFirstBlock + "-" + (trackFirstBlock + trackLastBlock) + "] (" + audioSource.getPureName()+")";
			int x = currentBaseX + 4;
			final int RADIUS = 8;
			for (int i = 0; i < channels; i++)
				{
				painter.setColor(isTake ? Color.RED : Color.GREEN);
				painter.fillOval(x + 4 * i, currentBaseY + Const.CLIP_INFO_AREA_HEIGHT / 2 - RADIUS / 2, RADIUS, RADIUS);
				}
			for (int i = 0; i < channels; i++)
				{
				painter.setColor(Color.BLACK);
				painter.drawOval(x + 4 * i, currentBaseY + Const.CLIP_INFO_AREA_HEIGHT / 2 - RADIUS / 2, RADIUS, RADIUS);
				}
			painter.setColor(Color.BLACK);
			painter.setFont(Const.FONT_CLIP_INFO);
			painter.setClip(currentBaseX - 2, currentBaseY - 1, getVisibleWidthX() + 2, 80);
			painter.drawString(fullClipInfo, x + 20, currentBaseY + 12);
			}


		void drawFades(Graphics2D painter)
			{
			final int firstRealX = getRealFirstX();
			final int lastRealX = getRealLastX();
			final int baseX = this.getVisibleFirstX();
			final int baseY = this.getBaseY();
			final int height = this.getHeight();
			final int clipXWidth = getVisibleWidthX();
			int fadeInLastX = firstRealX;
			int fadeOutFirstX = lastRealX;
			if (fadeInBlocks > 0)
				fadeInLastX = proxySong.blockToEditorX(trackFirstBlock + fadeInBlocks);
			if (fadeOuBlocks > 0)
				fadeOutFirstX = proxySong.blockToEditorX(trackLastBlock - fadeOuBlocks);
			int gy = (int) ((float) baseY + height - gain * (height - Const.CLIP_INFO_AREA_HEIGHT));
			if (proxySong.editingMode == Const.EDITING_MODE_NORMAL)
				{
				painter.setColor(getCmColor(CM_MAGENTA));
				if (fadeInBlocks > 0)
					{
					painter.drawLine(baseX, baseY + height, fadeInLastX, gy);
					}
				if (fadeOuBlocks > 0)
					{
					painter.drawLine(baseX + clipXWidth - 1, baseY + height, fadeOutFirstX, gy);
					}
				if (gain < 1.0)
					{
					painter.drawLine(fadeInLastX, gy, fadeOutFirstX, gy);
					}
				}

			if (doesCrossLeft())
				{
				final int lCrossX = calcCrossLeftX(firstRealX);
				painter.setColor(Color.GREEN);
				painter.drawArc(baseX - lCrossX, baseY + Const.CLIP_INFO_AREA_HEIGHT, lCrossX * 2, (height - Const.CLIP_INFO_AREA_HEIGHT) * 2, 90, -90);
				}
			if (doesCrossRight())
				{
				final int rCrossX = calcCrossRightX(lastRealX);
				painter.setColor(Color.YELLOW);
				painter.drawArc(baseX + clipXWidth - rCrossX, baseY + Const.CLIP_INFO_AREA_HEIGHT, rCrossX * 2, (height - Const.CLIP_INFO_AREA_HEIGHT) * 2, 180, -90);
				}
			}


		private boolean doesCrossRight()
			{
			boolean crossAtRight = false;
			if (next != null)
				crossAtRight = (next.trackFirstBlock < trackLastBlock);
			return crossAtRight;
			}


		private boolean doesCrossLeft()
			{
			boolean crossAtLeft = false;
			if (prev != null)
				crossAtLeft = (prev.trackLastBlock > trackFirstBlock);
			return crossAtLeft;
			}


		private final void drawContours(Graphics2D painter)
			{
			final int bx = this.getVisibleFirstX();
			final int by = this.getBaseY();
			final int w = getVisibleWidthX();
			final int h = this.getHeight();
			final int EW = Const.CLIP_INFO_AREA_HEIGHT;
			painter.setColor(Color.BLACK);
			painter.drawRect(bx, by, w, EW);
			painter.drawRect(bx, by + EW, w, h - EW);
			// draw the blue line tha indicates whole extension of the clip
			painter.setColor(Color.BLUE);
			painter.drawLine(getRealFirstX(), by + Const.CLIP_INFO_AREA_HEIGHT, getRealLastX(), by + h);
			}


		void setTrackFirstBlock(long newTrackFirstBlock)
			{
			trackFirstBlock = newTrackFirstBlock;
			updateTrackLastBlockProxy();
			}


		void setMuted(boolean b)
			{
			isMuted = b;
			parentTrack.recreateClipArea(proxyEditor.getGBuffer());
			}


		void set_left_edge(long block)
			{
			long olen = sourceLastBlock - sourceFirstBlock;
			long otfb = trackFirstBlock;
			trackFirstBlock = block;
			long lenght = olen + otfb - block;
			sourceFirstBlock = sourceLastBlock - lenght;
			}


		private void set_right_edge(long block)
			{
			long lenght = sourceLastBlock - sourceFirstBlock;
			long olen = lenght;
			long otlb = trackLastBlock;
			trackLastBlock = block;
			lenght = olen + block - otlb;
			sourceLastBlock = sourceFirstBlock + lenght;
			}


		private void set_first_source_block(long block)
			{
			// TODO
			}


		void set_last_source_block(long block)
			{
			sourceLastBlock = block;
			long lenght = sourceLastBlock - sourceFirstBlock;
			trackLastBlock = trackFirstBlock + lenght;
			}


		public final int getLenght()
			{
			return (int) (sourceLastBlock - sourceFirstBlock);
			}


		public void setBlur(boolean stat)
			{
			setCmBlur(CM_CLIP_BG, stat);
			setCmBlur(CM_CLIP_BG_ACTIVE, stat);
			setCmBlur(CM_CLIP_BG_SELECTED, stat);
			setCmBlur(CM_CLIP_BG_MUTED, stat);
			setCmBlur(CM_CLIP_INFO_AREA_BG, stat);
			setCmBlur(CM_CLIP_INFO_AREA_BG_ACTIVE, stat);
			setCmBlur(CM_CLIP_PEAK_MICROVIEW, stat);
			setCmBlur(CM_CLIP_PEAK_MACROVIEW, stat);
			setCmBlur(CM_CLIP_PEAK_OVERLOADED_SAMPLE, stat);
			setCmBlur(CM_DARK_TEXT, stat);
			}


		public int getBaseY()
			{
			return parentTrack.getRealBaseY() + Const.CLIP_VERTICAL_UPPER_MARGIN;
			}


		public int getHeight()
			{
			return parentTrack.getHeight() - Const.CLIP_VERTICAL_UPPER_MARGIN - Const.CLIP_VERTICAL_LOWER_MARGIN - Track.SEPARATOR_HEIGHT;
			}


		public void setFadeIn(long b)
			{
			fadeInBlocks = b;
			}


		public void setFadeOut(long b)
			{
			fadeOuBlocks = b;
			}


		public float getFadeFactorFor(long pos) // pos : position in the track
			{
			float fi, fo;
			if (fadeInBlocks == 0)
				fi = 1.0f;
			else
				{
				if (pos < trackFirstBlock) // that happens in the edge, the current playing pos , when rounded to locate the clip, may lead to a position "considered" inside the clip, but it is not.
					fi = 0.0f;
				else if (pos > trackFirstBlock + fadeInBlocks)
					fi = 1.0f;
				else
					fi = (float) ((double) (pos - trackFirstBlock) / fadeInBlocks);
				}
			if (fadeOuBlocks == 0)
				{
				fo = 1.0f;
				}
			else
				{
				trackLastBlock = trackFirstBlock + sourceLastBlock - sourceFirstBlock;
				if (pos > trackLastBlock)
					fo = 0.0f;
				else if (pos < trackLastBlock - fadeOuBlocks)
					fo = 1.0f;
				else
					fo = (float) ((double) (trackLastBlock - pos) / fadeOuBlocks);
				}
			// TODO calculate cross-fades-gains
			return fi * fo;
			}


		public void setGain(float g)
			{
			g = g < 0.0f ? 0.0f : g;
			g = g > 1.0f ? 1.0f : g;
			gain = g;
			}


		public void update()
			{
			final int baseX = getVisibleFirstX();
			final int clipXWidth = getVisibleWidthX();
			proxyEditor.repaint(baseX, parentTrack.getRealBaseY(), clipXWidth, parentTrack.getHeight());
			}


		private final int getRealFirstX()
			{
			return proxySong.blockToEditorX(trackFirstBlock);
			}


		private final int getRealLastX()
			{
			return proxySong.blockToEditorX(trackLastBlock);
			}


		private final int getRealWidthX()
			{
			return getLenght() / Const.ZOOM_STEP[proxySong.hzoom];
			}


		private final int getVisibleFirstX()
			{
			int firstRealX = getRealFirstX();
			return (firstRealX < 0) ? 0 : firstRealX;
			}


		private final int getVisibleLastX()
			{
			int lx = getRealLastX();
			int w = proxyEditor.getWidth();
			if (lx > w)
				lx = w;
			return lx;
			}


		private final int getVisibleWidthX()
			{
			return getVisibleLastX() - getVisibleFirstX();
			}


		private final int calcCrossRightX(int lastRealX)
			{
			return lastRealX - proxySong.blockToEditorX(next.trackFirstBlock);
			}


		private final int calcCrossLeftX(int firstRealX)
			{
			return proxySong.blockToEditorX(prev.trackLastBlock) - firstRealX;
			}


		public void slide(int deltaX)
			{
			trackFirstBlock += deltaX * Const.ZOOM_STEP[proxySong.hzoom];
			updateTrackLastBlockProxy();
			}


		public boolean isVisible()
			{
			int firstX = getVisibleFirstX();
			int lastX = proxySong.blockToEditorX(trackLastBlock);
			int maxX = proxyEditor.getWidth();
			return !((lastX < 0) || (firstX > maxX));
			}


		public boolean isWaitingPeaks(int hzoom)
			{
			Peak peak = audioSource.getPeak();
			if (peak == null)
				return false;
			int status = peak.getBuildStatus(hzoom);
			return ((status != Peak.BLOCKED) && ((status == Peak.BUILDING) || (status == Peak.NOT_AVAILABLE)));
			}


		public boolean peaksBlocked(int hzoom)
			{
			Peak peak = audioSource.getPeak();
			if (peak != null)
				return peak.getBuildStatus(hzoom) == Peak.BLOCKED;
			return false;
			}


		private final void updateTrackLastBlockProxy()
			{
			trackLastBlock = trackFirstBlock + (sourceLastBlock - sourceFirstBlock);
			}


		public AudioMedia getAudioSource()
			{
			return audioSource;
			}


		protected final void log(String message)
			{
			PMESG(3, "[ AudioClip : " + name + " ] " + message);
			}


		public synchronized void reparent(Track track)
			{
			parentTrack.removeClip(this);
			track.addClip(this);
			setParent(track);
			proxySong = parentTrack.parentSong;
			proxyEditor = proxySong.proxyEditor;
			}


		public void setParent(Track track)
			{
			this.parentTrack = track;
			}


		/** split the clip at the given point, creating a new one STILL NOT INSERTED ON THE TRACK. The responsability to inserted the new clip is clients */
		public AudioClip split(long splitPoint)
			{
			long leftLenght = splitPoint - (trackFirstBlock);
			long clipLenght = sourceLastBlock - sourceFirstBlock;
			long rightLenght = clipLenght - leftLenght;
			long rightSourceFirstBlock = sourceFirstBlock + leftLenght;
			AudioClip leftClip = this;
			leftClip.sourceLastBlock = leftClip.sourceFirstBlock + leftLenght;
			leftClip.trackLastBlock = splitPoint;
			AudioClip rightClip = new AudioClip(AudioClip.lastId++, parentTrack, leftClip.audioSource, rightSourceFirstBlock, rightLenght, splitPoint);
			rightClip.setName(leftClip.getName() + "-split");
			rightClip.isTake = leftClip.isTake;
			return rightClip;
			}


		public int getId()
			{
			return id;
			}


		public final String getName()
			{
			return name;
			}


		public final void setName(String clipName)
			{
			this.name = clipName;
			}

	}
