package org.protux.core.editor;

import static org.protux.core.Debugger.PERROR;

import java.awt.Frame;
import java.util.Timer;

import org.protux.plug.AudioPlugin;


public class Controller
	{
	public Chain parentChain;
	static final int MAX_CURVES = 10;
	Frame setupFrontend;
	int currentCurveIndex;
	Timer blinkCurveTimer;
	Timer blinkCurveNodeTimer;
	Curve curve[] = new Curve[MAX_CURVES];
	AudioPlugin assocPlugin;


	public Controller(Chain chain)
		{
		this.parentChain = chain;
		currentCurveIndex = -1;
		for (int i = 0; i < MAX_CURVES; i++)
			curve[i] = null;
		//MIG blinkCurveTimer = new QTimer(this);
		//MIG blinkCurveNodeTimer = new QTimer(this);
		//MIG connect( blinkCurveTimer , SIGNAL(timeout()), this, SLOT(blink_curve()));
		//MIG connect( blinkCurveNodeTimer , SIGNAL(timeout()), this, SLOT(blink_curve_node()));
		setupFrontend = new Frame();
		setupFrontend.resize(300, 200);
		}


	void setup()
		{
		setupFrontend.show();
		}


	void node_setup()
		{
		PERROR("Ooops ! I should have not been called ... I am a virtual method!");
		}


	String get_type()
		{
		PERROR("Ooops ! I should have not been called ... I am a virtual method!");
		return "";
		}


	int process(char[] data, int size)
		{
		PERROR("Ooops ! I should have not been called ... I am a virtual method!");
		return -1;
		}


	void draw()
		{
		curve[currentCurveIndex].setVisible(true);
		}


	void update_current_node(int x)
		{
		long blockPos = parentChain.parentTrack.parentSong.editorXToBlock(x);
		CurveNode lastnode = curve[currentCurveIndex].currentNode;
		if (blockPos >= 0)
			{
			CurveNode cn = curve[currentCurveIndex].getNearestNode(blockPos);
			if (cn != null)
				curve[currentCurveIndex].currentNode = cn;
			if ((lastnode != curve[currentCurveIndex].currentNode) && (lastnode.isHighLighted))
				{
				lastnode.toggle_highlight();
				}
			}
		}


	void activate()
		{
		start_blinking();
		}


	void deactivate()
		{
		stop_blinking();
		}


	void blink_curve()
		{
		//curve[currentCurveIndex]....
		}


	void blink_curve_node()
		{
		curve[currentCurveIndex].currentNode.toggle_highlight();
		}


	void start_blinking()
		{
		//MIG blinkCurveTimer.start(250);
		//MIG blinkCurveNodeTimer.start(250);
		}


	void stop_blinking()
		{
		//MIG blinkCurveTimer.stop();
		//MIG blinkCurveNodeTimer.stop();
		if ((curve[currentCurveIndex].currentNode != null) && (curve[currentCurveIndex].currentNode.isHighLighted))
			{
			curve[currentCurveIndex].currentNode.toggle_highlight();
			}
		}


	Curve get_current_curve()
		{
		return curve[currentCurveIndex];
		}


	void set_current_curve(String type)
		{
		for (int i = 0; i < MAX_CURVES; i++)
			{
			if ((curve[i] != null) && (curve[i].get_type() == type))
				{
				set_current_curve(i);
				break;
				}
			}
		}


	void set_current_curve(int index)
		{
		currentCurveIndex = index;
		}


	Curve get_curve(String curveType)
		{
		for (int i = 0; i < MAX_CURVES; i++)
			{
			if ((curve[i] != null) && (curveType == curve[i].get_type()))
				{
				return curve[i];
				}
			}
		return null;
		}


	int get_curve_index(Curve c)
		{
		for (int i = 0; i < MAX_CURVES; i++)
			{
			if ((curve[i] != null) && (c == curve[i]))
				{
				return i;
				}
			}
		return -1;
		}


	Curve add_curve(String type)
		{
		int ind = -1;
		for (int i = 0; i < MAX_CURVES; i++)
			{
			// cannot add a same audio plugin twice
			if ((curve[i] != null) && (curve[i].get_type() == type))
				{
				return null;
				}
			}
		for (int i = 0; i < MAX_CURVES; i++)
			{
			if (curve[i] == null)
				{
				ind = i;
				break;
				}
			}
		if (ind >= 0)
			{
			curve[ind] = new Curve(this, type);
			set_current_curve(ind);
			return curve[ind];
			}
		return null;
		}


	AudioPlugin get_audio_plugin()
		{
		return assocPlugin;
		}


	int cleanup()
		{
		//MIG assocPlugin.cleanup();
		return 0;
		}


	String get_schema()
		{
		String schema = "";
		for (int i = 0; i < MAX_CURVES; i++)
			{
			if (curve[i] != null)
				{
				schema = schema + "            <curve type=\"" + curve[i].get_type() + "\">\n";
				schema = schema + curve[i].get_schema();
				schema = schema + "            </curve>\n";
				}
			}
		return schema;
		}
	}
