package org.protux.core.editor;

import static org.protux.core.Debugger.PMESG;
import static org.protux.core.gui.ColorManager.CM_CURVE_ACTIVE;
import static org.protux.core.gui.ColorManager.CM_CURVE_NONACTIVE;
import static org.protux.core.gui.ColorManager.getCmColor;

import java.awt.Graphics2D;
import java.util.LinkedList;
import java.util.ListIterator;


 public class Curve extends LinkedList<CurveNode>
	{

	public CurveNode currentNode;
	public Controller parentController;
	public String type;
	public Track proxyParentTrack; 
	private boolean active;

	public Curve(Controller assocController, String pType)
		{
		this.parentController = assocController;
		proxyParentTrack = parentController.parentChain.parentTrack;
		type = pType;
		PMESG(3, "Creating Curve : " + pType);
		active = false;
		PMESG(3,"Adding root node [0:0.00]");
		CurveNode cn = new CurveNode(this, 0, 0.0f);// Initial node, always added.
		add(cn);
		currentNode = cn;
		}


	public CurveNode getNearestNode(long pos)
		{
		if (pos < 0)
			{
			return this.getFirst();
			}
		CurveNode nn = null;
		ListIterator<CurveNode> li = listIterator();
		while (li.hasNext()) //while (no != null)
			{
			CurveNode no = li.next();
			if (Math.abs(pos - no.pos) < Math.abs(pos - no.next.pos))
				{
				nn = no;
				break;
				}
			}
		return nn;
		}


	/** @deprecated Use setActive(boolean flag) instead */
	public void activate()
		{
		active = true;
		}


	/** @deprecated Use setActive(boolean flag) instead */
	public void deactivate()
		{
		active = false;
		}


	public void setActive(boolean b)
		{
		active = b;
		}


	public CurveNode add_node(long p, float v)
		{
		CurveNode no;
		if (p == 0)
			{
			no = getFirst();
			no.set_value(v);
			}
		else
			{
			no = new CurveNode(this, p, v);
			add_node(no);
			}
		return no;
		}


	public void add_node(CurveNode node) // ARRUMAR , MUDAR PARA RECURSOS DO LINKEDLIST
		{
		CurveNode n1 = getFirst();
		CurveNode n2 = getFirst();
		while (n1 != null && (n1.pos < node.pos))
			{
			n2 = n1;
			n1 = n1.next;
			}
		if (n1 == null)
			{
			n2.next = node;
			node.prev = n2;
			node.next = null;
			}
		else
			{
			node.prev = n2;
			node.next = n1;
			n2.next = node;
			n1.prev = node;
			}
		currentNode = node;
		}


	/** @deprecated Use remove(CurveNode node instead */
	public void delete_node(CurveNode node)
		{
		this.remove(node);
		}


	public float get_value_at(long pos) // THIS FUNCTION NEEDS HUGE OPTIMIZATION !! It finds the nearest node everytime it needs to get the value. for playback this is a waste of CPU. it could "follow" the playback and keep  updating the nearest node for playback cursor during the playback.
		{
		float value = 0.0f;
		CurveNode cn = getNearestNode(pos);
		if ((cn.next == null) && (pos >= cn.pos)) // cn.next indicates that we found the last node as the nearest; after that, all points use the last node value
			{
			return cn.value;
			}
		CurveNode cl;
		CurveNode cr;
		if (pos >= cn.pos)
			{
			cl = cn;
			cr = cn.next;
			}
		else
			{
			cl = cn.prev;
			cr = cn;
			}
		long dx = pos - cl.pos;
		long dd = cr.pos - cl.pos;
		float dv = cr.value - cl.value;
		float qd = dx / dd;
		float dy = dv * qd;
		value = cl.value + dy;
		return value;
		}


	public CurveNode get_current_curve_node()
		{
		return currentNode;
		}


	public String get_type()
		{
		return type;
		}


	public String get_schema()
		{
		String schema = "";
		ListIterator<CurveNode> li = listIterator();
		while (li.hasNext()) //while (no != null)
			{
			CurveNode n = li.next();
			String sp = "" + n.pos;
			String sv = "" + n.value;
			schema = schema + "                 <node pos=" + sp + " value=" + sv + " />\n";
			}
		return schema;
		}


	public void setVisible(boolean b)
		{
		int baseY = proxyParentTrack.getBaseY();
		int half = proxyParentTrack.getHeight() / 2;
		Graphics2D p = null; //MIG parentTrack.get_painter();

		CurveNode first = getFirst();
		int yc = baseY + half;
		int lastx = 0;
		int lasty = yc;

		//if ( n.pos < firstBlock ) continue; FOR LATER
		//if ( n.pos > lastBlock ) break; FOR LATER
		Song parentSong = proxyParentTrack.parentSong;
		if (active)
			{
			p.setColor(getCmColor(CM_CURVE_ACTIVE));
			}
		else
			{
			p.setColor(getCmColor(CM_CURVE_NONACTIVE));
			}
		ListIterator<CurveNode> li = listIterator();
		while (li.hasNext()) 
			{
			CurveNode n = li.next();
			int x = parentSong.blockToEditorX(n.pos);
			int y = yc + (int) (-1 * n.value * half / 100);
			p.drawLine(lastx, lasty, x, y);
			lastx = x;
			lasty = y;
			}
		p.drawLine(lastx, lasty, proxyParentTrack.get_width(), lasty);
		ListIterator<CurveNode> li2 = listIterator();
		while (li2.hasNext()) 
			{
			CurveNode n = li2.next();
			n.draw();
			}
		}

	}
