package org.protux.core.editor;

import static org.protux.core.Debugger.PMESG;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;

import org.protux.core.Const;
import org.protux.core.Protux;
import org.protux.core.SimpleTimer;
import org.protux.core.Util;
import org.protux.core.gui.FastPanel;
import org.protux.core.gui.InfoBox;
import org.protux.input.classic.EditorClassicActionListener;
import org.protux.input.jmb.JmbHandler;

/**
 * EVERY editorX refers to the real X in the editor, where "0" is the most left pixel in the editor, in the track PANEL area! editorX refers to the x value in the CLIP AREA (not including the track panel, so editorX=200 might something like editorX=340, if track panel is 140 width, for example
 */
public class Editor extends FastPanel implements MouseListener, MouseMotionListener
	{
		private static final int MODE_NOTDEFINED = 0;
		private static final int MODE_NONE = 1;
		private static final int MODE_FLOAT = 2;
		private static final int MODE_DRAG = 3;
		private static final int MODE_PLAY = 4;
		private static final int MAX_CURSORS = 20;
		private static final int CURSOR_FLOAT = 0;
		private static final int CURSOR_FLOAT_OVER_CLIP = 1;
		private static final int CURSOR_FLOAT_OVER_TRACK = 2;
		private static final int CURSOR_FLOAT_OVER_SPLITTER = 3;
		private static final int CURSOR_HOLD_UD = 4;
		private static final int CURSOR_HOLD_LR = 5;
		private static final int CURSOR_HOLD_LRUD = 6;
		private static final int CURSOR_DRAG = 7;
		private static final int CURSOR_SELECT = 8;
		private static final int CURSOR_MAGIC_ZOOM = 9;
		private static final Dimension MINIMUM_SIZE = new Dimension(800, Const.EDITOR_LOCATOR_HEIGHT + 200);
		private static final Color workCursorColor = new Color(150, 50, 50); // TODO move to CM
		private static final int FONT_LOCATOR_HEIGHT = 10;
		private static final Font FONT_LOCATOR = new Font("Helvetica", 0, FONT_LOCATOR_HEIGHT);
		private static final Color LOCATOR_FGCOLOR = new Color(100, 80, 40);
		private static final int FOLLOWER_TIMER_FPS = 25;

		public int mode = MODE_NOTDEFINED;
		public int floatX = 0;
		public int floatY;
		public int xOrig;
		public int yOrig;
		public int xOffset;
		public int yOffset;
		private int currentCursorMapIndex;
		private int cursorW;
		private int cursorH;
		private int workCusorEditorX = 0;
		private int editorXToStartPlaying = 0;
		private int mouseEditorX;
		private int mouseEditorY;
		private boolean isEnabled = true;

		protected JmbHandler jmbHandler = new EditorJmbHandler();
		protected EditorClassicActionListener classicHandler = new EditorClassicActionListener(this);

		public InfoBox infoBox;
		public Cursor cursorMap[] = new Cursor[MAX_CURSORS];
		private AudioClip selectedClip = null;
		private int lastMouseEditorX = -999;
		private int lastMouseEditorY = -999;
		private long lastMouseEventMoment;
		private boolean wasMouseHold;
		private SimpleTimer peakAvailabilityCheckTimer;
		private SimpleTimer playFollowerTimer;

		Song currentSong = null;
		boolean showPositionInSmpte = true;
		private Track targetTrackAfterMoveClip;
		private boolean draggingClip = false;


		// MIG cursorMap[CURSOR_FLOAT] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorFloat.xpm") );
		// MIG cursorMap[CURSOR_FLOAT_OVER_CLIP] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorFloatOverClip.xpm") );
		// MIG cursorMap[CURSOR_FLOAT_OVER_TRACK] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorFloatOverTrack.xpm") );
		// MIG cursorMap[CURSOR_FLOAT_OVER_SPLITTER] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorFloatOverSplitter.xpm") );
		// MIG cursorMap[CURSOR_HOLD_UD] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorHoldUd.xpm") );
		// MIG cursorMap[CURSOR_HOLD_LR] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorHoldLr.xpm") );
		// MIG cursorMap[CURSOR_HOLD_LRUD] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorHoldLrud.xpm") );
		// MIG cursorMap[CURSOR_DRAG] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorDrag.xpm") );
		// MIG cursorMap[CURSOR_SELECT] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorSelect.xpm") );
		// MIG cursorMap[CURSOR_MAGIC_ZOOM] = new Cursor( QPixmap(RESOURCES_DIR + "/cursorMagicZoom.xpm") );

		public Editor()
			{
			super("Editor");
			setBackground(Color.DARK_GRAY);
			addMouseMotionListener(this);
			addMouseListener(this);
			this.setMinimumSize(MINIMUM_SIZE);
			set_cursor(CURSOR_FLOAT);
			infoBox = Protux.infoBox;
			peakAvailabilityCheckTimer = new PeakChecker(this);
			}


		@Override
		public void recreate(Graphics2D painter)
			{
				log("Recreating...");
				if (currentSong != null)
					{
					drawLocator(painter);
					drawClips(painter);
					drawWorkCursor(painter);
					}
				else
					drawDumbX(painter);
			}


		public void setCurrentSong(Song song)
			{
			this.currentSong = song;
			}


		public void startPeakChecker()
			{
			// UNCO peakAvailabilityCheckTimer.start(2000);
			}


		public void recreateClipArea() // really used ?
			{
			drawClips(getGBuffer());
			}


		public void drawClips(Graphics2D painter)
			{
			clearRootSpace(painter);
			int t = currentSong.getNumTracks();
			for (int i = 0; i < t; i++)
				{
				Track tr = currentSong.trackList.get(i);
				if (tr.visible())
					{
					tr.recreateClipArea();
					}
				}
			}


		private void clearRootSpace(Graphics2D painter)
			{
			int lastEditorY = getLastY();
			painter.setColor(Color.BLACK);
			int w = getWidth();
			int h = getHeight();
			painter.fillRect(0, lastEditorY, w, h - lastEditorY);
			// repaint(0, lastEditorY, w, h - lastEditorY);
			}


		int getLastY()
			{
			int nt = currentSong.getNumTracks();
			if (nt > 0)
				{
				Track tr = currentSong.trackList.get(nt - 1);
				int r = tr.getRealBaseY() + tr.getHeight();
				return r;
				}

			return 0;
			}


		private void drawLocator(Graphics2D painter)
			{
			painter.setColor(Color.LIGHT_GRAY);
			int caw = getClipAreaWidth();
			painter.fillRect(0, 0, caw, Const.EDITOR_LOCATOR_HEIGHT);
			painter.setColor(Color.DARK_GRAY);
			painter.drawLine(0, 0, getWidth(), 0);
			int zp = Const.ZOOM_STEP[currentSong.hzoom];
			long firstb = currentSong.firstBlock;
			long lastb = firstb + caw * zp + 1;

			/*
			 * MIG if (currentSong.get_current_jogmode() == Const.JOG_MODE_CREATE_REGION) { painter.setColor(Color.BLUE); painter.setFont(FONT_LOCATOR); int xs = currentSong.clipBlockToClipX(currentSong.origBlockL) + baseClipAreaX; int xe = getMouseX(); if (xs < 0) xs = 0; if (xe > caw) xe = caw; painter.fillRect(xs, 0, xe - xs,
			 * Const.EDITOR_LOCATOR_HEIGHT); }
			 */

			painter.setFont(FONT_LOCATOR);
			int d = FONT_LOCATOR_HEIGHT / 2 + 2;
			for (int j = 0; j < 2; j++)
				{
				int k = 0;
				int fp = (zp + 1) * 10;
				long sfb = firstb + (firstb % 5000);
				if (j == 0)
					painter.setColor(Color.DARK_GRAY);
				else
					painter.setColor(Color.WHITE);
				for (long b = sfb; b < lastb; b += fp)
					{
					int x = currentSong.blockToEditorX(b) + 0;
					painter.drawLine(x + j, Const.EDITOR_LOCATOR_HEIGHT - 5, x + j, Const.EDITOR_LOCATOR_HEIGHT);
					if (++k > 10)
						{
						String s = showPositionInSmpte ? Util.blockToSmpte(b, currentSong.rate) : "" + b;
						painter.drawLine(x + j, Const.EDITOR_LOCATOR_HEIGHT - 10, x + j, Const.EDITOR_LOCATOR_HEIGHT);
						if (j == 0)
							{
							painter.drawString(s, x + j + 3, Const.EDITOR_LOCATOR_HEIGHT - d);
							}
						k = 0;
						}
					}
				}

			int xl = currentSong.blockToEditorX(currentSong.lastBlock) + 0;
			if (xl < caw)
				{
				painter.setColor(Color.GRAY);
				painter.fillRect(xl, 0, caw, Const.EDITOR_LOCATOR_HEIGHT);
				}
			}


		public void setCursorEnabled(boolean b)
			{
			isEnabled = b;
			}


		public void start()
			{
			// MIG set_mode(MODE_FLOAT);
			// MIG setMouseTracking(true);
			isEnabled = false;
			setEnabled(true);
			}


		public void set_mode(int pMode, int pOrigX, int pOrigY, int pCursorW, int pCursorH, int pOffsetX, int pOffsetY)
			{
			mode = pMode;
			xOrig = pOrigX;
			yOrig = pOrigY;
			xOffset = pOffsetX;
			yOffset = pOffsetY;
			cursorW = pCursorW;
			cursorH = pCursorH;
			}


		public void set_cursor(int cMode)
			{
			if ((cMode >= 0) && (cMode <= CURSOR_MAGIC_ZOOM))
				{
				// MIG setCursor( *(cursorMap[cMode]) );
				currentCursorMapIndex = cMode;
				}
			}


		public int get_current_cursor_map_index()
			{
			return currentCursorMapIndex;
			}


		private void update(int x, int y)
			{
			switch (mode)
				{
				case MODE_FLOAT:
					update_float_cursor(x);
					break;
				case MODE_DRAG:
					update_drag_cursor(x, y);
					break;
				case MODE_PLAY:
					updatePlayCursor();
					break;
				case MODE_NOTDEFINED:
					break;
				}
			}


		private int get_orig_x()
			{
			return xOrig;
			}


		private int get_orig_y()
			{
			return yOrig;
			}


		private int get_x_offset()
			{
			return xOffset;
			}


		private int get_y_offset()
			{
			return yOffset;
			}


		private void set_x_offset(int x)
			{
			xOffset = x;
			}


		private void set_y_offset(int y)
			{
			yOffset = y;
			}


		private void set_work_at(int x)
			{
			updateWorkCursor(x);
			}


		protected void updatePlayCursor()
			{
			Graphics2D painter = this.getGBuffer();
			int fullW = this.getClipAreaWidth();
			int fullH = this.getHeight();
			int x = currentSong.getPlayingXpos();
			if (x != editorXToStartPlaying)
				{
				painter.setXORMode(Color.white);
				int playXOld = editorXToStartPlaying;
				// painter.setColor(Color.GREEN);
				painter.drawLine(editorXToStartPlaying, 0, editorXToStartPlaying, fullH);
				editorXToStartPlaying = x;
				// painter.setColor(Color.BLACK);
				painter.drawLine(editorXToStartPlaying, 0, editorXToStartPlaying, fullH);
				// repaint(playXOld, 0, editorXToStartPlaying - playXOld, fullH);
				}
			if (x > fullW)
				currentSong.firstBlock = currentSong.getPlayingPos();
			}


		private void update_float_cursor(int x)
			{
			toggle_float_cursor(floatX, getHeight());
			floatX = x;
			toggle_float_cursor(floatX, getHeight());
			}


		private void update_drag_cursor(int x, int y)
			{
			toggle_drag_cursor();
			floatX = x;
			floatY = y;
			toggle_drag_cursor();
			}


		private void updateWorkCursor(long workPos)
			{
			int editorX = currentSong.blockToEditorX(workPos);
			workCusorEditorX = editorX;
			editorXToStartPlaying = editorX;
			}


		private void toggle_float_cursor(int x, int h)
			{
			Graphics2D painter = this.getGBuffer();
			if (x > 0)
				{
				// MIG painter.setRasterOp(Qt.XorROP);
				painter.setColor(new Color(250, 255, 0));
				painter.drawLine(x, 0, x, h);
				// MIG painter.setRasterOp(Qt.CopyROP);
				// MIG update(x, 0, 1, -1);
				}
			}


		private void toggle_drag_cursor()
			{
			// MIG painter.setRasterOp(Qt.XorROP);
			Graphics2D painter = this.getGBuffer();
			painter.setColor(new Color(155, 105, 155));
			int x0 = floatX - xOffset;
			int x1 = x0 + cursorW;
			int y0 = floatY - yOffset;
			int xc = floatX;
			int fullH = this.getHeight();
			if (x0 < 0)
				x0 = 0;
			if (x1 < 0 + cursorW)
				x1 = 0 + cursorW;
			if (xc < 0)
				xc = 0;
			painter.drawLine(x0, y0, x0 + cursorW, y0);
			painter.drawLine(x0, y0 + cursorH, x0 + cursorW, y0 + cursorH);
			painter.drawLine(x0, 0, x0, fullH);
			painter.drawLine(xc, 0, xc, fullH);
			painter.drawLine(x1, 0, x1, fullH);
			// MIG painter.setRasterOp(Qt.CopyROP);
			// MIG update(x0, floatY - yOffset, cursorW + 1, cursorH + 1);
			// MIG update(x0 - 1, 0, 1, fullH);
			// MIG update(x1 - 1, 0, 1, fullH);
			}


		private void drawWorkCursor(Graphics2D painter)
			{
			if (isEnabled)
				{
				if (workCusorEditorX >= 0)
					{
					int lx = workCusorEditorX;
					int ly = getLastY();
					painter.setColor(Color.RED);
					painter.drawLine(lx, Const.EDITOR_AND_RACK_TRACKS_BASE_Y, lx, ly);
					painter.drawLine(lx - 4, Const.EDITOR_AND_RACK_TRACKS_BASE_Y + 4, lx, Const.EDITOR_AND_RACK_TRACKS_BASE_Y);
					painter.drawLine(lx, Const.EDITOR_AND_RACK_TRACKS_BASE_Y, lx + 4, Const.EDITOR_AND_RACK_TRACKS_BASE_Y + 4);
					}
				}
			}


		private void drawCursorForMode()
			{
			if (isEnabled)
				{
				switch (mode)
					{
					case MODE_FLOAT:
						{
						toggle_float_cursor(floatX, this.getHeight());
						break;
						}
					case MODE_DRAG:
						{
						toggle_drag_cursor();
						break;
						}
					case MODE_PLAY:
						{
						// TODO
						break;
						}
					case MODE_NOTDEFINED:
						{
						break;
						}
					}
				}
			}


		public void set_work_cursor_color(int r, int g, int b)
			{
			// MIG workCursorColor.setRgb(r, g, b);
			}


		public int getClipAreaWidth()
			{
			return getWidth() - 0;
			}


		public int getBaseYForTracks()
			{
			return Const.EDITOR_LOCATOR_HEIGHT;
			}


		public AudioClip getClipUnderMouse()
			{
			return getClipUnderEditorXY(mouseEditorX, mouseEditorY);
			}


		public AudioClip getClipUnderEditorXY(int x, int y)
			{
			Track t = getTrackUnderEditorY(y);
			if (t != null)
				{
				long blockPos = currentSong.editorXToBlock(x);
				AudioClip c = t.getClipUnderBlock(blockPos);
				return c;
				}
			return null;
			}


		public void splitClipUnderMouse()
			{
			Track t = getTrackUnderEditorY(mouseEditorY);
			if (t != null)
				{
				long splitpoint = currentSong.editorXToBlock(workCusorEditorX);
				AudioClip newRightClip = t.splitClip(splitpoint);
				t.addClip(newRightClip);
				t.recreateClipArea(getGBuffer());
				}
			}


		public Track getTrackUnderEditorY(int y)
			{
			ArrayList<Track> tracks = currentSong.getTracks();
			int tt = tracks.size();
			for (int t = 0; t < tt; t++)
				{
				Track tr = tracks.get(t);
				int rby = tr.getRealBaseY();
				if ((y >= rby) && (y <= rby + tr.getHeight()))
					return tr;
				}
			return null;
			}


		public void touch()
			{
			touch(mouseEditorX, mouseEditorY);
			}


		public void touch(int editorX, int editorY)
			{
			long workingBlock = currentSong.editorXToBlock(editorX);
			updateWorkCursor(workingBlock);
			drawWorkCursor(getGBuffer());
			int w = getWidth();
			if (editorX > w * 0.8)
				currentSong.scrollLeft();
			if (editorX < w * 0.2)
				currentSong.scrollRight();

			Track t = getTrackUnderEditorY(editorY);
			if (t != null)
				currentSong.touch(workingBlock, t);
			else
				currentSong.touch(workingBlock);
			System.out.println("Touch at block " + workingBlock);
			}


		public void mouseDragged(MouseEvent me)
			{
			Graphics2D painter = getGBuffer();
			updateMousePositionProxy(me);
			if (selectedClip != null)
				{
				draggingClip = true;
				Track t = getTrackUnderEditorY(me.getY());
				Track originalTrack = selectedClip.parentTrack;
				originalTrack.clearClipArea(painter);
				int delta = mouseEditorX - lastMouseEditorX;
				selectedClip.slide(delta);
				selectedClip.draw();
				originalTrack.recreateClipArea(painter);
				originalTrack.repaintClipArea();
				// originalTrack.updateClipArea();
				if (t != originalTrack)
					{
					targetTrackAfterMoveClip = t;
					targetTrackAfterMoveClip.drawGhostClip(painter, selectedClip);
					targetTrackAfterMoveClip.repaintClipArea();
					}
				else
					targetTrackAfterMoveClip = null;
				}
			}


		public void mouseMoved(MouseEvent me)
			{
			updateMousePositionProxy(me);
			}


		private synchronized void updateMousePositionProxy(MouseEvent me)
			{
			lastMouseEditorX = mouseEditorX;
			lastMouseEditorY = mouseEditorY;
			mouseEditorX = me.getX();
			mouseEditorY = me.getY();
			}


		public void mouseClicked(MouseEvent arg0)
			{
			updateMousePositionProxy(arg0);
			if ((!wasMouseHold) && (arg0.getButton() == MouseEvent.BUTTON1))
				{
				if (selectedClip == null)
					selectedClip = getClipUnderMouse();
				if (selectedClip != null)
					{
					selectedClip.isSelected = !selectedClip.isSelected;
					selectedClip.draw();
					selectedClip.update();
					if (!selectedClip.isSelected)
						selectedClip = null;
					}
				}
			touch();
			}


		public void mousePressed(MouseEvent e)
			{
			if (e.getButton() == MouseEvent.BUTTON1)
				{
				selectedClip = getClipUnderEditorXY(e.getX(), e.getY());
				lastMouseEventMoment = e.getWhen();
				}
			}


		public void mouseReleased(MouseEvent e)
			{
			if (e.isPopupTrigger())
				{
				classicHandler.showContextMenu(e);
				}
			else if (e.getButton() == MouseEvent.BUTTON1)
				{
				wasMouseHold = (e.getWhen() - lastMouseEventMoment) > 150;
				if (draggingClip)
					checkIfClipWasReparented();
				}
			draggingClip = false;
			}


		private void checkIfClipWasReparented()
			{
			if ((selectedClip != null) && (targetTrackAfterMoveClip != null) && (targetTrackAfterMoveClip != selectedClip.parentTrack))
				{
				selectedClip.reparent(targetTrackAfterMoveClip);
				recreate();
				targetTrackAfterMoveClip = null;
				selectedClip = null;
				}
			}


		public void mouseEntered(MouseEvent arg0)
			{
			}


		public void mouseExited(MouseEvent arg0)
			{
			}


		public final synchronized int getMouseX()
			{
			return mouseEditorX;
			}


		public final synchronized int getMouseY()
			{
			return mouseEditorY;
			}


		public void recheckPeakAvailability()
			{
			String sd = new SimpleDateFormat("H:m:s,S").format(new Date(System.currentTimeMillis()));
			PMESG(2, "Rechecking peak avaiability ( at " + sd + " )");
			if (currentSong != null)
				currentSong.recheckPeakAvailability();
			}


		public JmbHandler getJmbHandler()
			{
			return jmbHandler;
			}


		public ActionListener getClassicHandler()
			{
			return classicHandler;
			}


		public void zoomOut()
			{
			currentSong.zoomOut();
			}


		public void zoomIn()
			{
			currentSong.zoomIn();
			}


		private void log(String string)
			{
			PMESG(2, "[ Editor ] " + string);
			}


		public boolean empty()
			{
			return false; // TODO
			}


		public boolean isPlaying()
			{
			return currentSong.softMapper.isGoing();
			}


		public void startFollow()
			{
			playFollowerTimer = new FollowTimer();
			playFollowerTimer.start(1000 / FOLLOWER_TIMER_FPS);
			}


		public void stopFollow()
			{
			playFollowerTimer.cancel();
			playFollowerTimer = null;
			}

		class FollowTimer extends SimpleTimer
			{

				@Override
				public void execute()
					{
					Protux.rack.updatePlayPos();
					updatePlayCursor();
					}

			}

	}

// methods apparently not being used
/*
 * private void recreateClipArea() { setCursorEnabled(false); set_work_at(currentSong.workingBlock); final int t = currentSong.getNumTracks(); for (int i = 0; i < t; i++) { if (currentSong.track[i].visible()) { currentSong.track[i].recreate_clip_area(); } } //drawLocator(); recreateSplitter(); //MIG enable(); } private void recreatePanelArea() {
 * int t = currentSong.getNumTracks(); for (int i = 0; i < t; i++) { if (currentSong.track[i].visible()) currentSong.track[i].recreate_panel(); } } private void recreateSplitter() { Graphics2D painter = getGBuffer(); //MIG clear(trackPanelAreaWidth, 0, splitterWidth, getHeight()); painter.setColor(new Color(190, 190, 190));
 * painter.drawLine(trackPanelAreaWidth, 0, trackPanelAreaWidth, getHeight()); painter.setColor(new Color(220, 220, 220)); painter.drawLine(trackPanelAreaWidth + 1, 0, trackPanelAreaWidth + 1, getHeight()); painter.setColor(new Color(250, 250, 250)); painter.drawLine(trackPanelAreaWidth + 2, 0, trackPanelAreaWidth + 2, getHeight());
 * painter.setColor(new Color(220, 220, 220)); painter.drawLine(trackPanelAreaWidth + 3, 0, trackPanelAreaWidth + 3, getHeight()); painter.setColor(new Color(190, 190, 190)); painter.drawLine(trackPanelAreaWidth + 4, 0, trackPanelAreaWidth + 4, getHeight()); repaint(trackPanelAreaWidth, 0, trackPanelAreaWidth + SPLITTER_WIDTH, getHeight()); }
 */
