package org.protux.core.editor;

import static org.protux.core.Debugger.PMESG;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import org.protux.core.Const;
import org.protux.core.Protux;
import org.protux.core.Util;
import org.protux.core.gui.FastPanel;

public class Rack extends FastPanel implements MouseListener
	{
		protected static final int FONT_NAVIGATOR_HEIGHT = 14;
		protected static final Font FONT_NAVIGATOR = new Font("Helvetica", Font.BOLD, FONT_NAVIGATOR_HEIGHT);
		protected static final int INFO_BASE_X = 5;
		protected static final int PLAYING_POS_BASE_X = 82;
		protected static final Color NAVIGATOR_BG_COLOR = Color.BLACK;
		static final Font FONT_TRACK_NAME = new Font("Arial", 0, 10);
		static final float SLIDER_WIDTH_FACTOR = 0.4f;
		static final int FONT_PANEL_Y_OFFSET = 8;
		static final int PANEL_FIRST_ROW_OFFSET = 5;
		static final int PANEL_ROW_HEIGHT = 16;
		static final int BUS_LED_HEIGHT = 9;
		static final int BUS_LED_WIDTH = 9;
		static final int BUSES_LEDS_Y = 24;
		static final int PANEL_HEAD_HEIGHT = 20;

		private static final Dimension MIN_SIZE = new Dimension(170, 50);
		private Editor proxyEditor;


		public Rack()
			{
			super("Rack");
			proxyEditor = Protux.editor;
			setMinimumSize(MIN_SIZE);
			setPreferredSize(MIN_SIZE);
			}


		@Override
		public void recreate(Graphics2D painter)
			{
			log("Recreating...");
			if (getCurrentSong() != null)
				{
				clearNavigator(painter);
				clearRootSpace(painter);
				drawNavigator(painter);
				drawPanels(painter);
				}
			else
				drawDumbX(painter);
			}


		private void clearNavigator(Graphics2D painter)
			{
			painter.setColor(NAVIGATOR_BG_COLOR);
			painter.fillRect(0, 0, getWidth(), Const.EDITOR_LOCATOR_HEIGHT);
			}


		private static void log(String m)
			{
			PMESG(2, "[ Rack ] " + m);
			}


		public Song getCurrentSong()
			{
			return proxyEditor.currentSong;
			}


		private void clearRootSpace(Graphics2D painter)
			{
			int lastEditorY = proxyEditor.getLastY();
			int w = getWidth();
			int h = getHeight();
			// painter.setColor(Color.BLACK);
			painter.clearRect(0, lastEditorY, w, h - lastEditorY);
			//repaint(0, lastEditorY, w, h - lastEditorY);
			// painter.fillRect(0, lastEditorY, w, w - lastEditorY);
			// repaint(0, lastEditorY, w, h - lastEditorY);
			}


		private void drawPanels(Graphics2D painter)
			{
			clearRootSpace(painter);
			Song song = proxyEditor.currentSong;
			int t = song.getNumTracks();
			for (int i = 0; i < t; i++)
				{
				Track tr = song.trackList.get(i);
				if (tr.visible())
					{
					tr.panel.recreate();
					}
				}
			}


		private void drawNavigator(Graphics2D painter)
			{
			drawWorkPosition(painter);
			drawPlayPosition(painter);
			}


		private void drawWorkPosition(Graphics2D painter)
			{
			Song currentSong = proxyEditor.currentSong;
			long bpos = currentSong.workingBlock;
			String s = (proxyEditor.showPositionInSmpte) ? Util.blockToSmpte(bpos, currentSong.rate) : "" + bpos;
			painter.setColor(Color.WHITE);
			painter.setFont(FONT_NAVIGATOR);
			painter.drawString(s, INFO_BASE_X, Const.EDITOR_LOCATOR_HEIGHT - (FONT_NAVIGATOR_HEIGHT / 2) + 2);
			}


		private void drawPlayPosition(Graphics2D painter)
			{
			Song currentSong = proxyEditor.currentSong;
			long playingPos = currentSong.getPlayingPos();
			String s = (proxyEditor.showPositionInSmpte) ? Util.blockToSmpte(playingPos, currentSong.rate) : "" + currentSong.getPlayingPos();
			painter.setColor(Color.WHITE);
			painter.setFont(FONT_NAVIGATOR);
			painter.drawString(s, INFO_BASE_X + PLAYING_POS_BASE_X, Const.EDITOR_LOCATOR_HEIGHT - (FONT_NAVIGATOR_HEIGHT / 2) + 2);
			}


		public boolean anySolo()
			{
			Song song = proxyEditor.currentSong;
			int nt = song.getNumTracks();
			for (int t = 0; t < nt; t++)
				{
				if (song.getTrack(t).isSolo())
					{
					return true;
					}
				}
			return false;
			}


		public void updatePlayPos()
			{
			Graphics2D painter = getGBuffer();
			int w = getWidth();
			painter.setColor(Color.BLACK);
			painter.fillRect(INFO_BASE_X + PLAYING_POS_BASE_X, 20, w - PLAYING_POS_BASE_X, Const.EDITOR_LOCATOR_HEIGHT - 20);
			drawPlayPosition(painter);
			//repaint(0, 0, w, Const.EDITOR_LOCATOR_HEIGHT);
			}


		public void mouseClicked(MouseEvent e)
			{
			int x = e.getX();
			int y = e.getY();
			Track t = proxyEditor.getTrackUnderEditorY(y);
			if (t != null)
				{
				t.panel.clicked(x, y);
				}
			}


		public void mousePressed(MouseEvent e)
			{
			}


		public void mouseReleased(MouseEvent e)
			{
			}


		public void mouseEntered(MouseEvent e)
			{
			}


		public void mouseExited(MouseEvent e)
			{
			}

	}
