package org.protux.core.editor;

import static org.protux.core.Debugger.PMESG;
import static org.protux.core.Debugger.PWARN;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;

import org.protux.core.Const;
import org.protux.core.Factory;
import org.protux.core.Project;
import org.protux.core.Properties;
import org.protux.core.Protux;
import org.protux.core.SchemaParser;
import org.protux.core.SimpleTimer;
import org.protux.core.engines.audio.SoftBusMapper;
import org.protux.core.gui.AudioPluginSelector;
import org.protux.core.gui.BusSelector;
import org.protux.core.gui.ProjectManager;
import org.protux.core.gui.dialogs.SongPropertiesDialog;
import org.protux.core.media.audio.AudioMedia;
import org.protux.core.media.audio.Praf;

public class Song
	{
		public Editor proxyEditor;
		public AudioClip workingClip;
		public AudioSourceList audioSourcesList = new AudioSourceList();
		public BusSelector busSelector;
		public AudioPluginSelector audioPluginSelector;
		public SoftBusMapper softMapper = null;
		public Project parentProject = null;
		public SongPropertiesDialog propertiesDialog = new SongPropertiesDialog(this);
		public ArrayList<Track> trackList = new ArrayList<Track>();
		public LinkedList<Region> regionList = new LinkedList<Region>();

		public String artists;
		public String title = Const.DEFAULT_SONG_TITLE;
		public String rootDir;
		public String audioSourcesDir;
		public String captureDir;
		public String schemaFilename;

		public long firstBlock;
		public long lastBlock;
		public long origBlockL;
		public long origBlockR;
		public long workingBlock;

		public int editingMode;
		private Track currentTrack;

		private SimpleTimer shuttleTimer;
		private SimpleTimer updateProgressTimer;
		private SimpleTimer BusMonitorRecreateTimer;
		private SimpleTimer splitter;

		public int baseJogZoomXFactor;
		public int rate;
		public int bitDepth;
		public int hzoom;

		private int draggingEdge;
		private int shuttleFactor;
		private int jogZoomTotalX;
		private int jogZoomTotalY;
		private int lastJogZoomXFactor;
		private int gainingTrack = -1;
		private int panningTrack;

		public int id;

		private int origX;
		private int origY;
		private int scrollAmount = 50;
		private int takeNumber = 1;
		private int verticalScrollAmount = 20;
		private int verticalJogZoomLastY;
		private int verticalScrollTotalHeight;

		public boolean isSnapOn = false;
		public boolean isChanged = false;

		public float masterGain;

		public Rack proxyRack;

		class ShuttleTimer extends SimpleTimer
			{

				@Override
				public void execute()
					{
					}

			}


		public Song(Project parentProject, int number)
			{
			this.id = number;
			this.parentProject = parentProject;
			this.rootDir = parentProject.getRootDir() + "/" + Const.SONG_DIR_PREFIX + "_" + id;
			this.audioSourcesDir = rootDir + "/audiosources";
			this.captureDir = rootDir + "/capture";
			this.schemaFilename = rootDir + "/" + Const.SONG_SCHEMA_FILE;
			this.proxyEditor = Protux.editor;
			this.proxyRack = Protux.rack;
			this.softMapper = new SoftBusMapper(this);
			this.shuttleTimer = new ShuttleTimer();
			// updateProgressTimer = new SimpleTimer(this, 50);
			// BusMonitorRecreateTimer = new SimpleTimer(this, 50);
			// audioPluginSelector = new AudioPluginSelector(pvp);
			// busSelector = new BusSelector(mta,mixer);
			// mhe = new MustuxHistoryEngine();
			}


		protected void shuttleCallback()
			{
			// TODO Auto-generated method stub

			}


		@Override
		public void finalize()
			{
			// MIG Jmb.disconnect(this);
			}


		public boolean isAnyTrackArmed()
			{
			for (Track tr : trackList)
				{
				if (tr.armed())
					return true;
				}
			return false;
			}


		/**
		 * return the X coordinate in the editor for the block given, considering the zoom_step and the current first visible block
		 * @param block
		 * @return
		 */
		public final int blockToEditorX(long block)
			{
			int x = (int) ((block - firstBlock) / Const.ZOOM_STEP[hzoom]);
			return x;
			}

		/**
		 * return the correponding block position for the given X coordinate in the editor , considering the zoom_step and the current first visible block
		 * @param block
		 * @return
		 */
		public final long editorXToBlock(int x)
			{
			long b = firstBlock + (x * Const.ZOOM_STEP[hzoom]);
			return b;
			}


		

		public int getClipsCountForAudio(AudioMedia a)
			{
			int count = 0;
			for (Track tr : trackList)
				{
				LinkedList<AudioClip> clipList = tr.clipList;
				for (AudioClip c : clipList)
					{
					if (c.getAudioSource() == a)
						{
						count++;
						}
					}
				}
			return count;
			}


		public int getFloorY()
			{
			int nt = getNumTracks();
			if (nt > 0)
				{
				Track t = trackList.get(nt - 1);
				int by = t.getBaseY();
				int h = t.getHeight();
				int fy = by + h;
				return fy;
				}
			return 0;
			}


		public long getPlayingPos()
			{
			return softMapper.getGoingPos();
			}


		public final int getPlayingXpos()
			{
			return blockToEditorX(softMapper.getGoingPos());
			}


		public void importAudio(File srcFile) throws IOException
			{
			Track cTrack = getCurrentTrack();
			PMESG(2, "Importing audio to a new clip onto track " + cTrack.get_id());
			Praf audio = Factory.importTo(srcFile, this);
			cTrack.createClip(AudioClip.lastId++,audio, workingBlock, false);
			updateLastBlock();
			recreateEditor();
			}


		public void load() throws IOException
			{
			Properties props = new Properties(rootDir + "/" + Const.SONG_PROPERTIES_FILE);
			title = props.getProperty("title");
			artists = props.getProperty("artists");
			rate = props.getInt("rate");
			bitDepth = props.getInt("bitDepth");
			hzoom = props.getInt("hzoom");
			editingMode = props.getInt("editingMode");
			firstBlock = props.getLong("firstBlock");
			workingBlock = props.getLong("workingBlock");
			int numTracksToCreate = props.getInt("numTracks");
			PMESG(2, "Reading Song Properties : title=" + title + " rate=" + rate + "  bitDepth=" + bitDepth);
			// MIG propertiesDialog.update();
			ArrayList<String> topLevelSectionList = SchemaParser.parse(schemaFilename);
			String sectionTracks = SchemaParser.getSection(topLevelSectionList, "tracks");
			String sectionRegions = SchemaParser.getSection(topLevelSectionList, "regions");
			ArrayList<String> tracksSchemaList = SchemaParser.getSections(sectionTracks, "track");
			ArrayList<String> regionsSchemaList = SchemaParser.getSections(sectionRegions, "region");

			// MIG if (!mixer.valid_mode(rate,bitDepth)) // checks if sample rate and bitdepth are OK. If not, ask user to correct it... this verification might go to recreate();...
			// MIG {
			// MIG propertiesDialog.raise();
			// MIG rate = propertiesDialog.get_rate();
			// MIG bitDepth = propertiesDialog.get_bit_depth();
			// MIG }
			// MIG mixer.validate_buses(rate, AudioDeviceMapper.STEREO, bitDepth);//We cannot add Tracks before the mixer has validated it's buses. ..

			// MIG add_regions(regionsSchemaList);
			PMESG(2, "Reading Songs Tracks Schema : " + tracksSchemaList.size() + " track(s).");
			addTracks(tracksSchemaList);

			int activeTrackNumber = props.getInt("activeTrackNumber");
			if ((activeTrackNumber > numTracksToCreate) || (activeTrackNumber < 0))
				{
				PWARN("Invalid active track number in song.properties for song " + title + ". Assuming current as track ID #0");
				activeTrackNumber = 0;
				}
			currentTrack = trackList.get(activeTrackNumber);
			editingMode = Const.EDITING_MODE_NORMAL;
			}


		public boolean togglePlay() throws Exception
			{
			log("Toggling...");
			if (softMapper.isGoing())
				{
				softMapper.forceStop();
				proxyEditor.stopFollow();
				Protux.info("STOP");
				updateLastBlock();
				return true;
				}
			if (ProjectManager.currentProjectIsPlaying())
				{
				Protux.info("ProjectManager is still playing a file!");
				return false;
				}
			if (parentProject.anyOtherSongGoing())
				{
				PWARN("Another Mixer engine is already going !");
				}
			updateLastBlock();
			boolean b = softMapper.startGo(workingBlock);
			if (b)
				{
				proxyEditor.startFollow();
				}
			return b;
			}


		public int removeAllClipsForAudio(AudioMedia a)
			{
			int counter = 0;
			for (Track tr : trackList)
				{
				LinkedList<AudioClip> clipList = tr.clipList;
				java.util.Iterator<AudioClip> it = clipList.iterator();
				AudioClip c = clipList.getFirst();
				while (it.hasNext())
					{
					AudioClip c2 = c;
					c = it.next();
					if (c2.getAudioSource() == a)
						{
						tr.removeClip(c2);
						counter++;
						}
					}
				}
			return counter;
			}


		public int render(int mode, boolean toFile, String filename)
			{
			if (mode == 0) // 0 = Normal Mode
				{
				if (toFile)
					{
						// MIG if (!filename)
						// MIG filename = QFileDialog.getSaveFileName( title + ".praf", "Praf/Wav Files (*.praf *.wav)", pvp ) ;
						// MIG if ( !filename.isEmpty() )
						{
						// MIG int r = mixer.render_all(filename);
						// MIG if (r<0)
						// PERROR("Rendering all tracks failed");
						// PEXIT;
						return 1;
						}
					}
				// TODO render to standard location ~/protuxprojects/Project/rendered/song_x.praf
				}
			// else if (mode==...) TODO
			return -1;
			}


		public void save() throws IOException
			{
			PMESG(2, "Saving Song schema " + schemaFilename);
			SchemaParser.saveXml(new File(schemaFilename), getSchema());
			PMESG(2, "Saving Song Properties : title=" + title + " rate=" + rate + "  bitDepth=" + bitDepth);
			saveProperties();
			}


		public void saveProperties() throws IOException
			{
			java.util.Properties props = new java.util.Properties();
			props.setProperty("title", "" + title);
			props.setProperty("artists", artists);
			props.setProperty("rate", "" + rate);
			props.setProperty("bitDepth", "" + bitDepth);
			if (currentTrack != null)
				{
				props.setProperty("activeTrackNumber", "" + currentTrack.get_id());
				}
			else
				{
				props.setProperty("activeTrackNumber", "0");
				}
			props.setProperty("hzoom", "" + hzoom);
			props.setProperty("editingMode", "" + editingMode);
			props.setProperty("firstBlock", "0");
			props.setProperty("workingBlock", "0");
			props.setProperty("numTracks", "" + getNumTracks());
			props.store(new FileOutputStream(new File(rootDir + "/" + Const.SONG_PROPERTIES_FILE)), null);
			}


		public String getSchema()
			{
			// regions
			String output = "";
			if (regionList != null)
				{
				output += "<regions>\n";
				for (Region m : regionList)
					{
					String sb = "" + m.beginBlock;
					String se = "" + m.endBlock;
					output += "     <region begin=" + sb + " end=" + se + ">\n</region>\n";
					}
				output += "</regions>\n\n";
				}

			// tracks
			if (getNumTracks() > 0)
				{
				output += "<tracks>\n";
				for (Track tr : trackList)
					{
					output += tr.getSchema();
					}
				output += "</tracks>\n";
				}
			return output;
			}


		public int snappedX(int x)
			{
			int nx = x;
			if (isSnapOn)
				{
				int wx = blockToEditorX(workingBlock);
				// TODO check if it is close to upper or lower nearest clip's edges
				if (nx < 10) // check if it is close to track begin
					nx = 0;
					// MIG if (abs(nx-wx)<10) // first try to snap to working block
					{
					// PMESG("snapping !");
					nx = wx;
					}
				}
			return nx;
			}


		/** add an empty Track */
		public Track addTrack()
			{
			int nextTrackId = getNumTracks();
			Track t = new Track(this, nextTrackId, "T "+nextTrackId);
			trackList.add(t);
			return t;
			}


		/** add a Track using a trackSchema */
		public Track addTrack(String trackSchema)
			{
			Track t = addTrack();
			t.setSchema(trackSchema);
			return t;
			}


		/** add tracks using a trackSchema list */
		public void addTracks(ArrayList<String> tracksSchemaList)
			{
			int t = tracksSchemaList.size();
			for (int i = 0; i < t; i++)
				{
				addTrack(tracksSchemaList.get(i));
				}
			}

		/** remove current active track*/
		public void removeActiveTrack()
			{
			trackList.remove(currentTrack);
			}
		
		public ArrayList<Track> getTracks()
			{
			return trackList;
			}


		public Track getTrack(int trackNumber)
			{
			if ((trackNumber < 0) || (trackNumber > getNumTracks()))
				return null;
			return trackList.get(trackNumber);
			}


		public Track getCurrentTrack()
			{
			return currentTrack;
			}




		public void changeSettings(int rate, int bitDepht)
			{
			this.rate = rate;
			this.bitDepth = bitDepht;
			try
				{
				save();
				}
			catch (IOException ex)
				{
				System.out.println("Error : " + ex);
				}
			// MIG mixer.validate_buses(rate,AudioDeviceMapper.STEREO,bitDepth);
			// mta.recreate_tracks();
			}


		public void detectAndSetContext(int mousex, int mousey)
			{
				// MIG if (!Jmb.get_current_jogmode())
				{
				int ccmi;
					// MIG ccmi = cursorManager.get_current_cursor_map_index();
					// MIG if (get_clip_under_xy(mousex, mousey))
					{
						// MIG if (ccmi!=org.protux.core.gui.CURSOR_FLOAT_OVER_CLIP)
						{
						// MIG cursorManager.set_cursor(org.protux.core.gui.CURSOR_FLOAT_OVER_CLIP);
						// MIG Jmb.set_current_context(ContextUponClip);
						}
					}
					// MIG else if (splitter_under_mouse(mousex))
					{
						// MIG if (ccmi!=org.protux.core.gui.CURSOR_FLOAT_OVER_SPLITTER)
						{
						// MIG cursorManager.set_cursor(org.protux.core.gui.CURSOR_FLOAT_OVER_SPLITTER);
						// MIG Jmb.set_current_context(ContextUponSplitter);
						}
					}
					// MIG else if (get_track_under_y(mousey))
					{
						// MIG if (ccmi!=org.protux.core.gui.CURSOR_FLOAT_OVER_TRACK)
						{
						// MIG cursorManager.set_cursor(org.protux.core.gui.CURSOR_FLOAT_OVER_TRACK);
						// MIG Jmb.set_current_context(ContextUponTrack);
						}
					}
					// MIG else
					{
						// MIG if (ccmi!=org.protux.core.gui.CURSOR_FLOAT)
						{
						// MIG cursorManager.set_cursor(org.protux.core.gui.CURSOR_FLOAT);
						// MIG Jmb.set_current_context(NoContext);
						}
					}
				}
			}


		/*
		 * MIG void mouse_move_event(QMouseEvent* e) { //PENTER4; Jmb.jog(); int x = e.x(); int y = e.y(); cursorManager.update(x,y); if (Jmb.get_current_mode() == CurveMode) track.get(activeTrackNumber].audioPluginChain.followMouse(x,y); detect_and_set_context(x, y); //PEXIT4; }
		 */
		public void jog_gain_pan(int track, int mousex, int mousey)
			{
			// PENTER;
			if ((track < 0) || (track > getNumTracks()))
				{
				// MIG parentProject.parentInterface.info("INVALID TRACK");
				// PEXIT;
				return;
				}
				// MIG if (jogMode == JOG_NONE)
				{
				// PMESG("Starting jog gain/pan for track %d",track);
				// MIG jogMode = JOG_GAIN_AND_PAN;
				gainingTrack = track;
				origX = mousex;
				origY = mousey;
				}
				// MIG else
				{
				// PMESG("Finishing jog gain/pan for track %d",track);
				// MIG jogMode = JOG_NONE;
				gainingTrack = -1; // NOT NEEDED .... THIS LINE CAN BE REMOVED
				}
			// PEXIT;
			}


		public void process_drag_edge()// //MIG JmbMessage* m)
			{
				// PENTER;
				// MIG if (m.beginHold)
				{
				// MIG mta.recreate(); //FIXME, should't be neccecary, but if I don',t the cursor is screwing up the Interface
				AudioClip cli = null;// MIG get_clip_under_xy(m.mousex(),m.mousey());
				if (cli != null)
					{
					workingClip = cli;
					int x = 0;// MIG clip_area_mouse_x(m.mousex());
					if (draggingEdge == -1) // [E]
						{
						int cxm = blockToEditorX(cli.trackFirstBlock + (cli.getLenght() / 2));
						if (x < cxm)
							{
							// MIG draggingEdge = LEFT;
							}
						else
							{
							// MIG draggingEdge = RIGHT;
							}
						}
						// MIG if (draggingEdge==RIGHT)
						{
						// MIG cursorManager.set_x_offset(block_to_xpos( cli.trackLastBlock ) - x);
						}
						// MIG else
						{
						// MIG cursorManager.set_x_offset(block_to_xpos( cli.trackFirstBlock ) - x);
						}
					}
				}
				// MIG else
				{
				// track.get(activeTrackNumber].recreate();
				// MIG mta.recreate(); //FIXME, should't be neccecary, but if I don',t the cursor is screwing up the Interface
				}
			updateLastBlock();
			// PEXIT;
			}


		public void process_fade_clip() // MIG (JmbMessage* m)
			{
				// PENTER;
				// MIG if ( m.beginHold )
				{
				AudioClip cli = null;// MIG get_clip_under_xy(m.mousex(),m.mousey());
				if (cli != null)
					{
					workingClip = cli;
					// MIG origX = clip_area_mouse_x(m.mousex());
					// MIG origY = m.mousey();
					// MIG origGain = workingClip.gain;
					origBlockL = workingClip.fadeInBlocks;
					origBlockR = workingClip.fadeOuBlocks;
					}
				}
				// MIG else
				{
				currentTrack.recreateClipArea();
				}
			updateLastBlock();
			// PEXIT;
			}


		public void recreateEditor()
			{
			proxyEditor.recreate();
			proxyRack.recreate();
			}


		public void setArtists(String pArtists)
			{
			this.artists = pArtists;
			}


		public void setMasterGain(float pMasterGain)
			{
			this.masterGain = pMasterGain;
			}


		public void setTitle(String sTitle)
			{
			this.title = sTitle;
			}


		public void set_active_track(int trackNumber)
			{
			if ((trackNumber > getNumTracks()) || (trackNumber < 0))
				{
				return;
				}
			currentTrack.deactivate();
			currentTrack = trackList.get(trackNumber);
			currentTrack.activate();
			}


		private void setHzoomLevel(int level) // FIXME Not used anymore?
			{
			if ((level > 0) && (level < Const.MAX_ZOOM_LEVELS - 1))
				{
				hzoom = level;
				PMESG(3, "HZOOM Now is " + hzoom);
				// MIG parentProject.parentInterface.infoBox.update_zoom();
				// recreate();
				}
			}


		public void touch(long block, Track touchedTrack)
			{
			currentTrack.deactivate();
			workingBlock = block;
			currentTrack = touchedTrack;
			currentTrack.activate();
			recreateEditor();
			}

		public void touch(long block)
			{
			touch (block,currentTrack);
			}

		public void update_gain(int y)
			{
			float h = 0.0f;// MIG (float) pvp.height() * 0.6;
			float ofy = (float) origY - y;
			float g = 2.0f * (ofy / h) * 12;
			// MIG float fg = g + origGain;
			// MIG track.get(gainingTrack].set_gain( fg );
			}


		public void update_pan(int x)
			{
			// MIG float w = (float) (pvp.width() * 0.6);
			float ofx = (float) origX - x;
			// MIG float p = -2.0f * (ofx) / w ;
			// MIG float fp = p + origPan;
			// MIG track.get(panningTrack].set_pan( fp );
			}


		public void update_shuttle()
			{
			float maxAmount = 0.0f;// MIG ((float) SHUTTLE_SENSIBILITY / 100.0f) * 30.0f;
			scrollAmount = (int) ((float) (shuttleFactor / 100.f) * maxAmount);
			if (scrollAmount != 0)
				{
				if (shuttleFactor < 0)
					{
					scrollAmount *= -1;
					scrollLeft();
					}
				else
					{
					scrollRight();
					}
				proxyEditor.recreateClipArea();
				}
			}


		public void set_first_block(long pos)
			{
			firstBlock = (pos >= 0) ? pos : 0;
			}


		public void set_work_at(long pos)
			{
			workingBlock = pos;
			}


		public void set_progress_timer(boolean start)
			{
			if (start)
				{
				// MIG updateProgressTimer.start(200);
				}
			else
				{
					// MIG if (updateProgressTimer.isActive())
					{
					// MIG updateProgressTimer.stop();
					// MIG mta.recreate_multi_clip_area();
					}
				}
			}


		public void toggle_snap()
			{
			isSnapOn = !isSnapOn;
			}


		public void update_busmonitor()
			{
			boolean ab;
				// MIG for (int b=0; b<AudioDeviceMapper.get_total_buses_in(); b++)
				{
					// MIG if (AudioDeviceMapper.is_bus_allocated_for_playback(b))
					{
					ab = true;
					// MIG break;
					}
				}
			if (!ab)
				{
				// MIG parentProject.parentInterface.busMonitor.recreate();
				// MIG BusMonitorRecreateTimer.stop();
				}
			}


		public void update_properties()
			{
				// MIG update_last_block();
				// MIG touch_track(activeTrackNumber, block_to_xpos(workingBlock));
				// MIG Jmb.set_current_object(this);
				// MIG Jmb.set_current_mode(currentMode);
				// MIG if (mixer.get_status()==Mixer.GOING)
				{
				// MIG cursorManager.set_mode(org.protux.core.gui.MODE_PLAY);
				// MIG cursorManager.enable();
				}
				// MIG else
				{
				/** Set the cursor to something totally different then is checked in detect_and_set_context to force proper detection of the context */
				// MIG cursorManager.set_cursor(org.protux.core.gui.CURSOR_FLOAT);
				}
			// MIG detect_and_set_context(pvp.get_mouse_x(), pvp.get_mouse_y());
			}


		void set_current_jogmode()// MIG JmbJogMode* mode)
			{
			// MIG currentJogMode = mode;
			}


		int get_current_jogmode()// MIG JmbJogMode get_current_jogmode()
			{
			return 0; // //MIG currentJogMode;
			}


		public void updateLastBlock()
			{
			lastBlock = 0; // find the last blockpos in MTA
			int t = getNumTracks();
			for (int i = 0; i < t; i++)
				{
				LinkedList<AudioClip> clipList = trackList.get(i).clipList;
				for (AudioClip c : clipList)
					{
					if (c.trackLastBlock >= lastBlock)
						lastBlock = c.trackLastBlock;
					}
				}
			}


		public int getNumTracks()
			{
			return trackList.size();
			}


		public boolean changed()
			{
			return isChanged;
			}


		private boolean splitter_under_mouse(int mousex)
			{
				// MIG if ( (mousex > mta.trackPanelAreaWidth) && (mousex < mta.clipAreaBaseX) )
				{
				// MIG return true;
				}
			return false;
			}


		public long getLastVisibleBlock()
			{
			return firstBlock + proxyEditor.getClipAreaWidth() * Const.ZOOM_STEP[hzoom] + 1;
			}


		public void recheckPeakAvailability()
			{
			for (Track tr : trackList)
				{
				if (tr.visible())
					{
					if (tr.someClipsWaitingForPeaks())
						{
						log("Still waiting peaks for some clips in track " + tr.get_id());
						}
					else if (tr.someClipPeaksBecameAvailable())
						{
						log("Clips are ready in track " + tr.get_id() + ". Recreating...");
						tr.recreateClipArea(proxyEditor.getGBuffer());
						tr.repaintClipArea();
						}
					else if (tr.someClipsBlocked())
						{
						log("Some clips blocked in track " + tr.get_id());
						}
					}
				}
			}


		private void log(String string)
			{
			PMESG(3, "[ Song : " + title + " ] " + string);
			}


		public void zoomOut()
			{
			if (hzoom < Const.MAX_ZOOM_LEVELS - 1)
				{
				hzoom++;
				proxyEditor.recreate();
				proxyEditor.repaint();
				}
			}


		public void zoomIn()
			{
			if (hzoom > 0)
				{
				hzoom--;
				proxyEditor.recreate();
				proxyEditor.repaint();
				}
			}


		public void scrollLeft()
			{
			firstBlock += scrollAmount * Const.ZOOM_STEP[hzoom];
			}


		public void scrollRight()
			{
			firstBlock -= scrollAmount * Const.ZOOM_STEP[hzoom];
			if (firstBlock < 0)
				firstBlock = 0;
			}


		public int rename_dir(String dirname) // MAYBE IT IS NOT SONGS SCOPE...
			{
			// MIG QDir d (rootDir);
			// MIG String s = parentProject.rootDir + dirname;
			// MIG if (!d.rename( rootDir, s))
			// PERROR("Could not rename song dir");
			// MIG rootDir = s;
			audioSourcesDir = rootDir + "/audiosources";
			captureDir = rootDir + "/capture";
			schemaFilename = rootDir + "/mta.schema";
				// MIG for (int i=0; i<MAX_AUDIO_SOURCES ; i++)
				{
				// MIG Audio a = audioSourcesList.get_source_for_index(i);
				// MIG if (a)
				// MIG a.set_root_dir(dirname);
				}
			return 1;
			}

	}

// old stuff .. maybe usefull ------------------------

// JMB ACTIONID SLOT CONNECTIONS
// MIG Jmb.connect_actionid_to_slot(this, 400, NoMode, NoContext, JMB_SLOT(Song,go));
// MIG Jmb.connect_actionid_to_slot(this, 574, NoMode, NoContext, JMB_SLOT(Song,go_regions));
// MIG Jmb.connect_actionid_to_slot(this, 575, NoMode, NoContext, JMB_SLOT(Song,go_loop_regions));
// MIG Jmb.connect_actionid_to_slot(this, 408, NoMode, ContextUponTrack, JMB_SLOT(Song,mute_track));
// MIG Jmb.connect_actionid_to_slot(this, 408, NoMode, ContextUponClip, JMB_SLOT(Song,mute_clip));
// MIG Jmb.connect_actionid_to_slot(this, 7, NoMode, NoContext, JMB_SLOT(Song,set_editing_mode));
// MIG Jmb.connect_actionid_to_slot(this, 8, NoMode, NoContext, JMB_SLOT(Song,set_curve_mode));
// MIG Jmb.connect_actionid_to_slot(this, 1, NoMode, NoContext, JMB_SLOT(Song,cancel));
// MIG Jmb.connect_actionid_to_slot(this, 402, NoMode, ContextUponTrackorCLip, JMB_SLOT(Song,capture_from_channel_both));
// MIG Jmb.connect_actionid_to_slot(this, 403, NoMode, ContextUponTrackorCLip, JMB_SLOT(Song,capture_from_channel_left));
// MIG Jmb.connect_actionid_to_slot(this, 404, NoMode, ContextUponTrackorCLip, JMB_SLOT(Song,capture_from_channel_right));
// MIG Jmb.connect_actionid_to_slot(this, 414, NoMode, NoContext, JMB_SLOT(Song,select_bus_in));
// MIG Jmb.connect_actionid_to_slot(this, 415, NoMode, NoContext, JMB_SLOT(Song,select_bus_out));
// MIG Jmb.connect_actionid_to_slot(this, 10, NoMode, NoContext, JMB_SLOT(Song,show_song_properties));
// MIG Jmb.connect_actionid_to_slot(this, 212, NoMode, NoContext, JMB_SLOT(Song,add_track));
// MIG Jmb.connect_actionid_to_slot(this, 213, NoMode, NoContext, JMB_SLOT(Song,delete_track));
// MIG Jmb.connect_actionid_to_slot(this, 106, NoMode, NoContext, JMB_SLOT(Song,touch), Jmb.UseCollectedNumber);
// MIG Jmb.connect_actionid_to_slot(this, 106, NoMode, NoContext, JMB_SLOT(Song,touch));
// MIG Jmb.connect_actionid_to_slot(this, 107, NoMode, NoContext, JMB_SLOT(Song,touch_and_center));
// MIG Jmb.connect_actionid_to_slot(this, 401, NoMode, NoContext, JMB_SLOT(Song,toggle_armed));
// MIG Jmb.connect_actionid_to_slot(this, 501, NoMode, ContextUponClip, JMB_SLOT(Song,reset_clip_gain));
// MIG Jmb.connect_actionid_to_slot(this, 502, NoMode, ContextUponClip, JMB_SLOT(Song,reset_clip_fade_in));
// MIG Jmb.connect_actionid_to_slot(this, 503, NoMode, ContextUponClip, JMB_SLOT(Song,reset_clip_fade_out));
// MIG Jmb.connect_actionid_to_slot(this, 504, NoMode, ContextUponClip, JMB_SLOT(Song,reset_clip_fade_both));
// MIG Jmb.connect_actionid_to_slot(this, 500, NoMode, ContextUponClip, JMB_SLOT(Song,clip_gain));
// MIG Jmb.connect_actionid_to_slot(this, 500, NoMode, ContextUponTrack, JMB_SLOT(Song,track_gain));
// MIG Jmb.connect_actionid_to_slot(this, 500, NoMode, NoContext, JMB_SLOT(Song,track_gain), Jmb.UseCollectedNumber);
// MIG Jmb.connect_actionid_to_slot(this, 211, NoMode, ContextUponTrackorCLip, JMB_SLOT(Song,track_pan));
// MIG Jmb.connect_actionid_to_slot(this, 211, NoMode, NoContext, JMB_SLOT(Song,track_pan), Jmb.UseCollectedNumber);
// MIG Jmb.connect_actionid_to_slot(this, 111, NoMode, NoContext, JMB_SLOT(Song,work_next_edge));
// MIG Jmb.connect_actionid_to_slot(this, 112, NoMode, NoContext, JMB_SLOT(Song,work_previous_edge));
// MIG Jmb.connect_actionid_to_slot(this, 590, NoMode, NoContext, JMB_SLOT(Song,in_crop));
// MIG Jmb.connect_actionid_to_slot(this, 601, CurveMode, NoContext, JMB_SLOT(Song,select_audio_plugin_controller));
// MIG Jmb.connect_actionid_to_slot(this, 602, CurveMode, NoContext, JMB_SLOT(Song,remove_current_audio_plugin_controller));
// MIG Jmb.connect_actionid_to_slot(this, 603, CurveMode, NoContext, JMB_SLOT(Song,remove_audio_plugin_controller));
// MIG Jmb.connect_actionid_to_slot(this, 505, CurveMode, NoContext, JMB_SLOT(Song,add_audio_plugin_controller));
// MIG Jmb.connect_actionid_to_slot(this, 610, CurveMode, NoContext, JMB_SLOT(Song,add_node));
// MIG Jmb.connect_actionid_to_slot(this, 612, CurveMode, NoContext, JMB_SLOT(Song,drag_and_drop_node));
// MIG Jmb.connect_actionid_to_slot(this, 600, CurveMode, NoContext, JMB_SLOT(Song,audio_plugin_setup));
// MIG Jmb.connect_actionid_to_slot(this, 611, CurveMode, NoContext, JMB_SLOT(Song,node_setup));
// MIG Jmb.connect_actionid_to_slot(this, 214, NoMode, ContextUponClip, JMB_SLOT(Song,select_clip));
// MIG Jmb.connect_actionid_to_slot(this, 215, NoMode, ContextUponClip, JMB_SLOT(Song,add_clip_to_selection));
// MIG Jmb.connect_actionid_to_slot(this, 216, NoMode, ContextUponClip, JMB_SLOT(Song,remove_clip_from_selection));
// MIG Jmb.connect_actionid_to_slot(this, 217, NoMode, NoContext, JMB_SLOT(Song,invert_clip_selection));
// MIG Jmb.connect_actionid_to_slot(this, 218, NoMode, NoContext, JMB_SLOT(Song,select_all_clips));
// MIG Jmb.connect_actionid_to_slot(this, 219, NoMode, NoContext, JMB_SLOT(Song,deselect_all_clips));
// MIG Jmb.connect_actionid_to_slot(this, 100, NoMode, NoContext, JMB_SLOT(Song,zoom));
// MIG Jmb.connect_actionid_to_slot(this, 101, NoMode, NoContext, JMB_SLOT(Song,shuttle));
// MIG Jmb.connect_actionid_to_slot(this, 102, NoMode, NoContext, JMB_SLOT(Song,scroll_left));
// MIG Jmb.connect_actionid_to_slot(this, 103, NoMode, NoContext, JMB_SLOT(Song,scroll_right));
// MIG Jmb.connect_actionid_to_slot(this, 104, NoMode, NoContext, JMB_SLOT(Song,scroll_up));
// MIG Jmb.connect_actionid_to_slot(this, 105, NoMode, NoContext, JMB_SLOT(Song,scroll_down));
// MIG Jmb.connect_actionid_to_slot(this, 108, NoMode, NoContext, JMB_SLOT(Song,vertical_scroll));
// MIG Jmb.connect_actionid_to_slot(this, 109, NoMode, NoContext, JMB_SLOT(Song,goto_begin));
// MIG Jmb.connect_actionid_to_slot(this, 110, NoMode, NoContext, JMB_SLOT(Song,goto_begin));
// MIG Jmb.connect_actionid_to_slot(this, 107, NoMode, NoContext, JMB_SLOT(Song,center));
// MIG Jmb.connect_actionid_to_slot(this, 553, NoMode, ContextUponClip, JMB_SLOT(Song,split_clip));
// MIG Jmb.connect_actionid_to_slot(this, 407, NoMode, ContextUponTrackorCLip, JMB_SLOT(Song,solo_track));
// MIG Jmb.connect_actionid_to_slot(this, 571, NoMode, NoContext, JMB_SLOT(Song,create_region_start));
// MIG Jmb.connect_actionid_to_slot(this, 572, NoMode, NoContext, JMB_SLOT(Song,create_region_end));
// MIG Jmb.connect_actionid_to_slot(this, 570, NoMode, NoContext, JMB_SLOT(Song,create_region));
// MIG Jmb.connect_actionid_to_slot(this, 573, NoMode, NoContext, JMB_SLOT(Song,delete_region_under_x));
// MIG Jmb.connect_actionid_to_slot(this, 505, EditingMode, ContextUponClip, JMB_SLOT(Song,jog_clip_fade_in));
// MIG Jmb.connect_actionid_to_slot(this, 506, EditingMode, ContextUponClip, JMB_SLOT(Song,jog_clip_fade_out));
// MIG Jmb.connect_actionid_to_slot(this, 507, EditingMode, ContextUponClip, JMB_SLOT(Song,jog_clip_fade_both));
// MIG Jmb.connect_actionid_to_slot(this, 300, NoMode, ContextUponClip, JMB_SLOT(Song,drag_edge));
// MIG Jmb.connect_actionid_to_slot(this, 303, NoMode, ContextUponClip, JMB_SLOT(Song,drag_edge_both));
// MIG Jmb.connect_actionid_to_slot(this, 301, NoMode, ContextUponClip, JMB_SLOT(Song,drag_edge_left));
// MIG Jmb.connect_actionid_to_slot(this, 302, NoMode, ContextUponClip, JMB_SLOT(Song,drag_edge_right));
// MIG Jmb.connect_actionid_to_slot(this, 550, NoMode, NoContext, JMB_SLOT(Song,process_delete));
// MIG Jmb.connect_actionid_to_slot(this, 551, NoMode, ContextUponClip, JMB_SLOT(Song,copy_clip));
// MIG Jmb.connect_actionid_to_slot(this, 552, NoMode, ContextUponClip, JMB_SLOT(Song,drag_clip));
// MIG Jmb.connect_actionid_to_slot(this, 552, NoMode, ContextUponSplitter, JMB_SLOT(Song,drag_splitter));
// MIG Jmb.connect_actionid_to_slot(this, 113, NoMode, NoContext, JMB_SLOT(Song,hzoom_in));
// MIG Jmb.connect_actionid_to_slot(this, 114, NoMode, NoContext, JMB_SLOT(Song,hzoom_out));
// MIG Jmb.connect_actionid_to_slot(this, 115, NoMode, NoContext, JMB_SLOT(Song,vzoom_out));
// MIG Jmb.connect_actionid_to_slot(this, 116, NoMode, NoContext, JMB_SLOT(Song,vzoom_in));
// MIG Jmb.connect_actionid_to_slot(this, 15, NoMode, NoContext, JMB_SLOT(Song,undo));
// MIG Jmb.connect_actionid_to_slot(this, 16, NoMode, NoContext, JMB_SLOT(Song,redo));
// //MIG Jmb.connect_actionid_to_slot(this, 400, NoMode, NoContext, JMB_SLOT(Song,goto_end));
// //MIG Jmb.connect_actionid_to_slot(this, 400, NoMode, NoContext, JMB_SLOT(Song,track_gain_pan));
// JMB JOGMODE SLOT CONNECTIONS
// MIG Jmb.connect_jogmode_to_slot(this, JogCreateRegion, JMB_SLOT(Song,jog_create_region));
// MIG Jmb.connect_jogmode_to_slot(this, JogDragEdge, JMB_SLOT(Song,jog_drag_edge));
// MIG Jmb.connect_jogmode_to_slot(this, JogDragSplitter, JMB_SLOT(Song,jog_drag_splitter));
// MIG Jmb.connect_jogmode_to_slot(this, JogClipFadeBoth, JMB_SLOT(Song,jog_fade_clip));
// MIG Jmb.connect_jogmode_to_slot(this, JogClipFadeIn, JMB_SLOT(Song,jog_fade_clip));
// MIG Jmb.connect_jogmode_to_slot(this, JogClipFadeOut, JMB_SLOT(Song,jog_fade_clip));
// MIG Jmb.connect_jogmode_to_slot(this, JogClipGain, JMB_SLOT(Song,jog_fade_clip));
// MIG Jmb.connect_jogmode_to_slot(this, JogShuttle, JMB_SLOT(Song, jog_shuttle));
// MIG Jmb.connect_jogmode_to_slot(this, JogTrackPan, JMB_SLOT(Song,jog_track_pan));
// MIG Jmb.connect_jogmode_to_slot(this, JogVerticalScroll, JMB_SLOT(Song,jog_vertical_scroll));
// MIG Jmb.connect_jogmode_to_slot(this, JogZoom, JMB_SLOT(Song,jog_zoom));
/*
 * JmbSlot Song::add_clip_to_selection(JmbMessage* m) { AudioClip* cli = get_clip_under_xy(m->mousex(),m->mousey()); cli->select(); mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::add_audio_plugin_controller(JmbMessage* m) { if (m->beginHold) { Jmb::set_current_jogmode(JogAddFilterController); audioPluginSelector->start(); } else {
 * audioPluginSelector->stop(); QString filterType = audioPluginSelector->get_selected_filter_type(); track.get(activeTrackNumber]->audioPluginChain->add_audio_plugin_controller(filterType); mta->recreate(); } return (MustuxAction*) 0; } JmbSlot Song::add_node(JmbMessage* m) { track.get(activeTrackNumber]->audioPluginChain->add_node(); return
 * (MustuxAction*) 0; } JmbSlot Song::add_track(JmbMessage* m) { int newTrackNumber = numTracks; track.get(newTrackNumber]= new Track(this,newTrackNumber+1,"Unnamed", numTracks * Track::INITIAL_HEIGHT, Track::INITIAL_HEIGHT); numTracks++; mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::cancel(JmbMessage* m) // REVISE-ME { / * if (recorder &&
 * (recorder->get_status()!=Recorder::REC_NONE)) { recorder->abort(); info("Recorder aborted!"); } return (MustuxAction*) 0; / return (MustuxAction*) 0; } JmbSlot Song::capture_from_channel_both(JmbMessage* m) { volatileTrack = get_track_under_y(m->mousey()); volatileTrack->set_which_channels(MustuxAudioDeviceMapper::STEREO); //FIXME !!!!!!!
 * MustuxAudioDeviceMapper::STEREO, MustuxAudioDeviceMapper::ONLY_LEFT_CHANNEL or MustuxAudioDeviceMapper::ONLY_RIGHT_CHANNEL volatileTrack->recreate(); return (MustuxAction*) 0; } JmbSlot Song::capture_from_channel_left(JmbMessage* m) { volatileTrack = get_track_under_y(m->mousey());
 * volatileTrack->set_which_channels(MustuxAudioDeviceMapper::ONLY_LEFT_CHANNEL); volatileTrack->recreate(); return (MustuxAction*) 0; } JmbSlot Song::capture_from_channel_right(JmbMessage* m) { volatileTrack = get_track_under_y(m->mousey()); volatileTrack->set_which_channels(MustuxAudioDeviceMapper::ONLY_RIGHT_CHANNEL); volatileTrack->recreate();
 * return (MustuxAction*) 0; } JmbSlot Song::center(JmbMessage* m) { PENTER2; int w = mta->clipAreaWidth; int half = w/2; long long minimumBlockToCenter = w * Peak::zoomStep[hzoom] / 2; if ( workingBlock >= minimumBlockToCenter ) { int x = block_to_xpos(workingBlock); if (x<half) { scrollAmount = half - x; scroll_left(); } else { scrollAmount =
 * x-half; scroll_right(); } } else { mta->recreate(); } PEXIT2; return (MustuxAction*) 0; } JmbSlot Song::clip_gain(JmbMessage* m) { if (m->beginHold) { parentProject->parentInterface->info("CLIP GAIN"); Jmb::set_current_jogmode(JogClipGain); } process_fade_clip(m); return (MustuxAction*) 0; } JmbSlot Song::copy_clip(JmbMessage* m) { PENTER;
 * MustuxAction* a = new CopyClipAction(m, this); return a; PEXIT; } JmbSlot Song::create_region(JmbMessage* m) { if (m->beginHold) { origBlockL = xpos_to_block(clip_area_mouse_x(m->mousex())); origBlockR = -1; Jmb::set_current_jogmode(JogCreateRegion); } else { origBlockR = xpos_to_block(clip_area_mouse_x(m->mousex())); MtaRegion* m = new
 * MtaRegion(origBlockL,origBlockR); regionList->add_region(m); mta->recreate(); } update_last_block(); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::create_region_start(JmbMessage* m) { // int xpos ... //TODO return (MustuxAction*) 0; } JmbSlot Song::create_region_end(JmbMessage* m) { // int xpos ... //TODO return (MustuxAction*) 0; } JmbSlot
 * Song::deselect_all_clips(JmbMessage* m) { for (int i=0; i < numTracks; i++) { AudioClip* c = track.get(i]->clipList; while (c) { c->deselect(); c=c->next; } } mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::delete_region_under_x(JmbMessage* m) { // int xpos... return (MustuxAction*) 0; } JmbSlot Song::delete_track(JmbMessage* m) { int
 * trackNumber; trackNumber = activeTrackNumber; if ((trackNumber<0) || (trackNumber>=numTracks)) return (MustuxAction*) -1; if (track.get(trackNumber]->clipList) { QCString s1; s1.setNum(track.get(trackNumber]->get_total_clips()); QString mesg = "There are "+s1+" clips on this tracks\nAre you sure you want to delete it ?"; if ( QMessageBox::warning( pvp,
 * "Delete Track", mesg,"&YES", "&CANCEL", 0 , 0, 1 ) != 0) return (MustuxAction*) -1; } Track* tT = track.get(trackNumber]; if (trackNumber==numTracks-1) { PDEBUG("apagando a ultima... Whatever it means :-P" ); track.get(numTracks]=(Track*) 0; activeTrackNumber=numTracks-2; numTracks--; } else { for (int i=trackNumber+1; i<numTracks; i++) { int b =
 * track.get(i-1]->get_baseY(); int h =track.get(i-1]->get_height(); track.get(i-1]=track.get(i]; track.get(i-1]->set_baseY(b); track.get(i-1]->set_height(h); } track.get(numTracks-1]=(Track*)0; activeTrackNumber=trackNumber - 1 >=0 ? trackNumber - 1 : 0; numTracks--; } delete tT; mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::drag_and_drop_node(JmbMessage* m) {
 * track.get(activeTrackNumber]->audioPluginChain->drag_node(m->beginHold); return (MustuxAction*) 0; } JmbSlot Song::drag_clip(JmbMessage* m) { PENTER; // Drag_clip is a hold action. This MustuxAction is completely handled by Jmb engine. MustuxAction* a = new MoveClipAction(m, this); PEXIT; return a; } JmbSlot Song::drag_edge(JmbMessage* m) { PENTER;
 * if (m->beginHold) { Jmb::set_current_jogmode(JogDragEdge); draggingEdge=-1; } process_drag_edge(m); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::drag_edge_both(JmbMessage* m) { PENTER; if (m->beginHold) { Jmb::set_current_jogmode(JogDragEdge); draggingEdge= LEFT_AND_RIGHT; } process_drag_edge(m); PEXIT; return (MustuxAction*) 0; } JmbSlot
 * Song::drag_edge_left(JmbMessage* m) { PENTER; if (m->beginHold) { Jmb::set_current_jogmode(JogDragEdge); draggingEdge=LEFT; } process_drag_edge(m); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::drag_edge_right(JmbMessage* m) { PENTER; if (m->beginHold) { Jmb::set_current_jogmode(JogDragEdge); draggingEdge=RIGHT; } process_drag_edge(m); PEXIT;
 * return (MustuxAction*) 0; } JmbSlot Song::drag_splitter(JmbMessage* m) { if (m->beginHold) Jmb::set_current_jogmode(JogDragSplitter); return (MustuxAction*) 0; } JmbSlot Song::audio_plugin_setup(JmbMessage* m) { track.get(activeTrackNumber]->audioPluginChain->audio_plugin_setup(); return (MustuxAction*) 0; } JmbSlot Song::goto_begin(JmbMessage* m) {
 * firstBlock = 0; workingBlock = 0; mta->recreate_multi_clip_area(); return (MustuxAction*) 0; } JmbSlot Song::goto_end(JmbMessage* m) { set_work_at(lastBlock); center(); return (MustuxAction*) 0; } JmbSlot Song::go_loop_regions(JmbMessage* m) { int r=0; if (this->process_go(10)) { //TODO return (MustuxAction*) 0; } r = process_go(r); // return r;
 * return (MustuxAction*) 0; } JmbSlot Song::go_regions(JmbMessage* m) { int r=0; if (process_go(10)) { if (regionList) { MtaRegion* firstRegion = regionList->head(); r = mixer->go(firstRegion); parentProject->parentInterface->info("GOING THRU REGIONS ..."); PMESG("GOING (REGIONS) FROM %d ...", (int) firstRegion->beginBlock ); } else {
 * parentProject->parentInterface->info("No region to play !"); } } r = process_go(r); // return r; return (MustuxAction*) 0; } JmbSlot Song::hzoom_out(JmbMessage* m) { PENTER; if ( hzoom < Peak::MAX_ZOOM_LEVELS - 1 ) { hzoom++; if (Peak::zoomStep[hzoom] == 48) hzoom++; PMESG("HZOOM Now is %d",hzoom);
 * parentProject->parentInterface->infoBox->update_zoom(); } center(); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::hzoom_in(JmbMessage* m) { PENTER; if (hzoom > 0) { hzoom--; if(Peak::zoomStep[hzoom] == 48) hzoom--; PMESG("HZOOM Now is %d",hzoom); parentProject->parentInterface->infoBox->update_zoom(); } center(); PEXIT; return (MustuxAction*)
 * 0; } JmbSlot Song::in_crop(JmbMessage* m) { MustuxAction* a = new CropAction(m, this); return a; } JmbSlot Song::invert_clip_selection(JmbMessage* m) { for (int i=0; i < numTracks; i++) { AudioClip* c = track.get(i]->clipList; while (c) { if (c->selected()) c->deselect(); else c->select(); c=c->next; } } mta->recreate_multi_clip_area(); return
 * (MustuxAction*) 0; } JmbSlot Song::jog_clip_fade_in(JmbMessage* m) { if (m->beginHold) Jmb::set_current_jogmode(JogClipFadeIn); process_fade_clip(m); return (MustuxAction*) 0; } JmbSlot Song::jog_clip_fade_out(JmbMessage* m) { if (m->beginHold) Jmb::set_current_jogmode(JogClipFadeOut); process_fade_clip(m); return (MustuxAction*) 0; } JmbSlot
 * Song::jog_clip_fade_both(JmbMessage* m) { if (m->beginHold) Jmb::set_current_jogmode(JogClipFadeBoth); process_fade_clip(m); return (MustuxAction*) 0; } JmbSlot Song::jog_create_region(JmbMessage* m) { mta->recreate_multi_clip_area(); return (MustuxAction*) 0; } JmbSlot Song::jog_drag_edge(JmbMessage* m) { PENTER4; int x =
 * clip_area_mouse_x(m->mousex()); long long b = xpos_to_block(x + cursorManager->get_x_offset()); long long tfb = workingClip->trackFirstBlock; long long tlb = workingClip->trackLastBlock; long long sfb = workingClip->sourceFirstBlock; long long slb = workingClip->sourceLastBlock; long long totb = workingClip->audioSource->file->totalBlocks; bool
 * recr = false; if (draggingEdge==LEFT) { if (sfb + (b-tfb)>=0) { workingClip->set_left_edge(b); recr = true; } } else if (draggingEdge==RIGHT) { if (slb + (b-tlb) <= totb ) { workingClip->set_right_edge(b); recr = true; } } else if (draggingEdge==LEFT_AND_RIGHT) { long long w = tlb-tfb; if ((sfb + (b-tfb)>=0) && (slb + (b+w-tlb) <= totb )) {
 * workingClip->set_left_edge(b); workingClip->set_right_edge(b+w); recr = true; } } if (recr) workingClip->parentTrack->recreate_clip_area();// same as track.get(pointedtrack]->recreate(); maybe could use this second since is more clear PEXIT4; return (MustuxAction*) 0; } JmbSlot Song::jog_drag_splitter(JmbMessage* m) {
 * mta->set_track_panel_area_width(m->mousex()); return (MustuxAction*) 0; } JmbSlot Song::jog_fade_clip(JmbMessage* m) { PENTER4; int x = clip_area_mouse_x(m->mousex()); int y = m->mousey(); if (Jmb::get_current_jogmode() == JogClipFadeIn) { long long b = (x - origX) * Peak::zoomStep[hzoom]; long long dbi = origBlockL + b; dbi = dbi < 0? 0:dbi;
 * workingClip->set_fade_in(dbi); } else if (Jmb::get_current_jogmode() == JogClipFadeOut) { long long b = (origX - x) * Peak::zoomStep[hzoom]; long long dbo = origBlockR + b; dbo = dbo < 0? 0:dbo; workingClip->set_fade_out(dbo); } else if (Jmb::get_current_jogmode() == JogClipFadeBoth) { long long b = (x - origX) * Peak::zoomStep[hzoom]; long long
 * dbi = origBlockL + b; dbi = dbi < 0? 0:dbi; long long dbo = origBlockR + b; dbo = dbo < 0? 0:dbo; workingClip->set_fade_in(dbi); workingClip->set_fade_out(dbo); } else if (Jmb::get_current_jogmode() == JogClipGain) { int dy = (origY - y); float g = origGain + ( (float) dy / 200 ); workingClip->set_gain(g); }
 * workingClip->parentTrack->recreate_clip_area(); PEXIT4; return (MustuxAction*) 0; } JmbSlot Song::jog_track_pan(JmbMessage* m) { update_pan(clip_area_mouse_x(m->mousex())); return (MustuxAction*) 0; } JmbSlot Song::jog_vertical_scroll(JmbMessage* m) { float f; int amount; int y = m->mousey(); int vsy = origY - y; float limiter = 1.0f; if (vsy <
 * -12 || vsy > 12) { float dy = (float)MOUSE_SENSIBILITY*(((float)vsy)/pvp->height()); if (dy < 0.0f) dy *= -1.0f; if (dy > 0.6f) dy = 0.6f; limiter -= dy; } if (vsy < -MOUSE_SENSIBILITY) { f = (float) vsy / (pvp->height()); amount = (int) (f * verticalScrollTotalHeight * VERTICAL_SCROLL_SENSIBILITY / 100.0f * limiter); mtaBaseY+=amount; origY
 * +=MOUSE_SENSIBILITY; if (mtaBaseY < -(verticalScrollTotalHeight-pvp->height()) + 1) mtaBaseY=-(verticalScrollTotalHeight-pvp->height() + 1); if (verticalScrollTotalHeight < pvp->height()) mtaBaseY = 0; recreate(); } if (vsy > MOUSE_SENSIBILITY) { f = (float) vsy / (pvp->height()); amount = (int) (f * verticalScrollTotalHeight *
 * VERTICAL_SCROLL_SENSIBILITY / 100.0f * limiter); mtaBaseY+=amount; origY -=MOUSE_SENSIBILITY; if (mtaBaseY > 0) mtaBaseY = 0; recreate(); } return (MustuxAction*) 0; } JmbSlot Song::jog_zoom(JmbMessage* m) { int x = clip_area_mouse_x(m->mousex()); int y = m->mousey(); int jzxfactor = (int) ( (float) (jogZoomTotalX - x) / jogZoomTotalX * 50 ) +
 * 1; if (jzxfactor==0) jzxfactor=1; if (jzxfactor != lastJogZoomXFactor) { lastJogZoomXFactor = jzxfactor; hzoom = jzxfactor + baseJogZoomXFactor; if ( hzoom < 0 ) hzoom = 0; if ( hzoom > Peak::MAX_ZOOM_LEVELS -1 ) hzoom = Peak::MAX_ZOOM_LEVELS - 1; center(); } int vzy = y - verticalJogZoomLastY; if (vzy>10) { vzoom_in(); verticalJogZoomLastY =
 * verticalJogZoomLastY + 10; } else if (vzy<-10) { vzoom_out(); verticalJogZoomLastY = verticalJogZoomLastY - 10; } parentProject->parentInterface->infoBox->update_zoom(); return (MustuxAction*) 0; } JmbSlot Song::mute_clip(JmbMessage* m) { PENTER2; MustuxAction* a = new MuteClipAction(m, this); PEXIT2; return a; } JmbSlot
 * Song::mute_track(JmbMessage* m) { PENTER2; Track* track = get_track_under_y(m->mousey()); if (track) track->toggle_mute(); PEXIT2; return (MustuxAction*) 0; } JmbSlot Song::node_setup(JmbMessage* m) { track.get(activeTrackNumber]->audioPluginChain->node_setup(); return (MustuxAction*) 0; } JmbSlot Song::process_delete(JmbMessage* m) { PENTER;
 * AudioClip* cli = 0 ; for (int i=0; i < numTracks; i++) { bool someClipDeleted=false; cli=track.get(i]->clipList; while (cli) if (cli->selected()) { AudioClip* c = cli->next; track.get(i]->remove_clip(cli); cli=c; someClipDeleted=true; } else cli=cli->next; if (someClipDeleted) mta->recreate_multi_clip_area(); } update_last_block(); PEXIT; return
 * (MustuxAction*) 0; } JmbSlot Song::remove_clip_from_selection(JmbMessage* m) { AudioClip* cli = get_clip_under_xy(m->mousex(),m->mousey()); cli->deselect(); mta->recreate_multi_clip_area(); return (MustuxAction*) 0; } JmbSlot Song::reset_clip_gain(JmbMessage* m) { AudioClip* c = get_clip_under_xy(m->mousex(), m->mousey()); if (c)
 * c->set_gain(1.0); mta->recreate_multi_clip_area(); return (MustuxAction*) 0; } JmbSlot Song::remove_current_audio_plugin_controller(JmbMessage* m) { track.get(activeTrackNumber]->audioPluginChain->remove_controller(); return (MustuxAction*) 0; } JmbSlot Song::remove_audio_plugin_controller(JmbMessage* m) {
 * track.get(activeTrackNumber]->audioPluginChain->select_controller(m->beginHold); if (!m->beginHold) //FIXME Do you really want to have this behaviour?????? beginHold first selects the controller, second beginHold removes it!!!!! track.get(activeTrackNumber]->audioPluginChain->remove_controller(); return (MustuxAction*) 0; } JmbSlot
 * Song::reset_clip_fade_in(JmbMessage* m) { AudioClip* c = get_clip_under_xy(m->mousex(),m->mousey()); c->set_fade_in(0); mta->recreate_multi_clip_area(); return (MustuxAction*) 0; } JmbSlot Song::reset_clip_fade_out(JmbMessage* m) { AudioClip* c = get_clip_under_xy(m->mousex(),m->mousey()); c->set_fade_out(0); mta->recreate_multi_clip_area();
 * return (MustuxAction*) 0; } JmbSlot Song::reset_clip_fade_both(JmbMessage* m) { AudioClip* c = get_clip_under_xy(m->mousex(), m->mousey()); c->set_fade_in(0); c->set_fade_out(0); mta->recreate_multi_clip_area(); return (MustuxAction*) 0; } JmbSlot Song::select_all_clips(JmbMessage* m) { for (int i=0; i < numTracks; i++) { AudioClip* c =
 * track.get(i]->clipList; while (c) { c->select(); c=c->next; } } mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::select_bus_in(JmbMessage* m) { PENTER; if (m->beginHold) { Jmb::set_current_jogmode(JogSelectBusIn); busSelector->start(BusSelector::IN); volatileTrack = get_track_under_y(m->mousey()); if (!volatileTrack) volatileTrack =
 * track.get(activeTrackNumber]; } else { busSelector->stop(); int b = busSelector->get_selected_bus(); if ((b>=0) && (mixer->valid_capture_bus(b))) { volatileTrack->set_bus_in(b); } else parentProject->parentInterface->info("INVALID CAPTURE BUS FOR THIS SONG",2); mta->recreate(); } PEXIT; return (MustuxAction*) 0; } JmbSlot
 * Song::select_bus_out(JmbMessage* m) { if (m->beginHold) { Jmb::set_current_jogmode(JogSelectBusOut); busSelector->start(BusSelector::OUT); volatileTrack = get_track_under_y(m->mousey()); if (!volatileTrack) volatileTrack = track.get(activeTrackNumber]; } else { busSelector->stop(); int b = busSelector->get_selected_bus(); if ((b>=0) &&
 * (mixer->valid_playback_bus(b))) { volatileTrack->set_bus_out(b); } else parentProject->parentInterface->info("INVALID PLAYBACK BUS FOR THIS SONG",2); mta->recreate(); } return (MustuxAction*) 0; } JmbSlot Song::select_clip(JmbMessage* m) { PENTER; AudioClip* cli = get_clip_under_xy(m->mousex(), m->mousey()); for (int i=0; i < numTracks; i++) {
 * AudioClip* c = track.get(i]->clipList; while (c) { c->deselect(); c=c->next; } } cli->select(); mta->recreate(); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::select_audio_plugin_controller(JmbMessage* m) { if (m->beginHold) Jmb::set_current_jogmode(JogSelectFilterController);
 * track.get(activeTrackNumber]->audioPluginChain->select_controller(m->beginHold); return (MustuxAction*) 0; } JmbSlot Song::scroll_down(JmbMessage* m) { mtaBaseY-=verticalScrollAmount; recreate(); return (MustuxAction*) 0; } JmbSlot Song::scroll_up(JmbMessage* m) { if (mtaBaseY+verticalScrollAmount>0) mtaBaseY=0; else mtaBaseY+=verticalScrollAmount;
 * recreate(); return (MustuxAction*) 0; } JmbSlot Song::set_editing_mode(JmbMessage* m) { cursorManager->disable(); parentProject->parentInterface->info("CURRENT MODE : EDITING"); currentMode=EditingMode; Jmb::set_current_mode(currentMode); for (int i=0; i<numTracks; i++) track.get(i]->set_blur(false);
 * cursorManager->set_mode(CursorManager::MODE_FLOAT); cursorManager->set_work_cursor_color(250,50,50); for (int i=0; i<numTracks; i++) track.get(i]->audioPluginChain->deactivate(); cursorManager->enable(); mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::set_curve_mode(JmbMessage* m) { cursorManager->disable();
 * parentProject->parentInterface->info("CURRENT MODE : TRACK CURVES"); currentMode=CurveMode; Jmb::set_current_mode(currentMode); for (int i=0; i<numTracks; i++) track.get(i]->set_blur(true); cursorManager->set_mode(CursorManager::MODE_NONE); cursorManager->set_work_cursor_color(255,255,255); track.get(activeTrackNumber]->audioPluginChain->activate();
 * cursorManager->enable(); mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::show_song_properties(JmbMessage* m) { propertiesDialog->show(); return (MustuxAction*) 0; } JmbSlot Song::shuttle(JmbMessage* m) { PENTER; if (m->beginHold) { PMESG("Starting shuttle"); Jmb::set_current_jogmode(JogShuttle); jog_shuttle(m);
 * shuttleTimer->start(25); } else { PMESG("Stopping shuttle"); shuttleTimer->stop(); } PEXIT; return (MustuxAction*) 0; } JmbSlot Song::solo_track(JmbMessage* m) { Track* track = get_track_under_y(m->mousey()); if (track) track->toggle_solo(); return (MustuxAction*) 0; } JmbSlot Song::split_clip(JmbMessage* m) { PENTER; MustuxAction* a = new
 * SplitClipAction(m, this); PEXIT; return a; } JmbSlot Song::touch(JmbMessage* m) { int x = clip_area_mouse_x(m->mousex()); int trackNumber = m->collectedNumber - 1; if (trackNumber>=0) // there is a track number collected !! { if ((trackNumber<=numTracks) && (trackNumber>=0)) touch_track(trackNumber, x); } else { for (int i=0; i<numTracks; i++) {
 * if (track.get(i]->is_pointed(m->mousey())) { touch_track(i); break; } } } mta->recreate(); return (MustuxAction*) 0; } JmbSlot Song::touch_and_center(JmbMessage* m) { touch(m); center(); return (MustuxAction*) 0; } JmbSlot Song::toggle_armed(JmbMessage* m) { Track* t = get_track_under_y(m->mousey()); if (t) { if (t->armed()) t->disarm(); else
 * t->arm(); } return (MustuxAction*) 0; } JmbSlot Song::track_gain(JmbMessage* m) { MustuxAction* a = new GainTrackAction(m, this); return a; } JmbSlot Song::track_gain_pan(JmbMessage* m) { int trackNumber; if (m->collectedNumber>=0) trackNumber = m->collectedNumber*10; //FIXME else trackNumber = 2; //FIXME jog_gain_pan(trackNumber,
 * clip_area_mouse_x(m->mousex()), m->mousey()); return (MustuxAction*) 0; } JmbSlot Song::track_pan(JmbMessage* m) { if (m->beginHold) { parentProject->parentInterface->info("TRACK PAN"); int trackNumber; if (m->collectedNumber>=0) trackNumber = m->collectedNumber - 1; else trackNumber = activeTrackNumber; if ((trackNumber < 0) || (trackNumber >
 * numTracks)) { parentProject->parentInterface->info("INVALID TRACK"); } else { panningTrack = trackNumber; origX = clip_area_mouse_x(m->mousex()); origPan = track.get(trackNumber]->get_pan(); Jmb::set_current_jogmode(JogTrackPan); } } return (MustuxAction*) 0; } JmbSlot Song::jog_shuttle(JmbMessage* m) { PENTER; float f = (float)
 * clip_area_mouse_x(m->mousex()) / mta->clipAreaWidth; shuttleFactor = (int) (( f * 200.0f ) - 100.0f); //printf("f=%2.2f shuttleFactor=%d\n",f,shuttleFactor); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::vertical_scroll(JmbMessage* m) { if (m->beginHold) { verticalScrollTotalHeight = get_floorY(); origY = m->mousey();
 * Jmb::set_current_jogmode(JogVerticalScroll); } return (MustuxAction*) 0; } JmbSlot Song::vzoom_out(JmbMessage* m) { PENTER; int lasty=MTA::LOCATOR_HEIGHT; for (int i=0; i<numTracks; i++) { track.get(i]->vzoom_out(lasty); lasty+=track.get(i]->get_height(); } mta->recreate(); // mtp->recreate(); PEXIT; return (MustuxAction*) 0; } JmbSlot
 * Song::vzoom_in(JmbMessage* m) { PENTER; int lasty=MTA::LOCATOR_HEIGHT; for (int i=0; i<numTracks; i++) { track.get(i]->vzoom_in(lasty); lasty+=track.get(i]->get_height(); } mta->recreate(); // mtp->recreate(); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::work_next_edge(JmbMessage* m) { long long w = lastBlock; for (int i=0; i<numTracks; i++) {
 * AudioClip* c=track.get(i]->get_clip_after(workingBlock); if ((c) && (c->trackFirstBlock<w)) w=c->trackFirstBlock; } if (w != lastBlock) set_work_at(w); return (MustuxAction*) 0; } JmbSlot Song::work_previous_edge(JmbMessage* m) { long long w = 0; for (int i=0; i<numTracks; i++) { AudioClip* c=track.get(i]->get_clip_before(workingBlock); if ((c) &&
 * (c->trackFirstBlock >= w)) w=c->trackFirstBlock; } set_work_at(w); return (MustuxAction*) 0; } JmbSlot Song::zoom(JmbMessage* m) { PENTER; if (m->beginHold) { cursorManager->set_cursor(CursorManager::CURSOR_MAGIC_ZOOM); jogZoomTotalX = pvp->width(); jogZoomTotalY = pvp->height(); verticalJogZoomLastY = m->mousey(); baseJogZoomXFactor = hzoom -
 * ((int) ( (float) (jogZoomTotalX - clip_area_mouse_x(m->mousex())) / jogZoomTotalX * 50 ) + 1 ); lastJogZoomXFactor = -1; Jmb::set_current_jogmode(JogZoom); } else cursorManager->set_cursor(CursorManager::CURSOR_FLOAT); PEXIT; return (MustuxAction*) 0; } JmbSlot Song::undo(JmbMessage* m) { MustuxAction* a = mhe->undo(); if (a) { a->undo_action(); }
 * return (MustuxAction*) 0; } JmbSlot Song::redo(JmbMessage* m) { MustuxAction* a = mhe->redo(); if (a) { a->do_action(); } return (MustuxAction*) 0; }
 */
