package org.protux.core.editor;

import static org.protux.core.gui.ColorManager.CM_TRACK_PANEL_BG;
import static org.protux.core.gui.ColorManager.getCmColor;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.text.NumberFormat;
import java.util.ArrayList;

import org.protux.core.Const;
import org.protux.core.gui.Rectangle;

public class TrackPanel
	{
		Track assocTrack;
		Led ledMute;
		Led ledSolo;
		Led ledArmed;

		static final Font FONT_LEDS = new Font("Helvetica", 0, 10);
		static final Font FONT_PANEL = new Font("Helvetica", 0, 10);
		static final Color COLOR_LED_BRIGHT_GREEN = new Color(80, 250, 80);
		static final Color COLOR_LED_BRIGHT_RED = new Color(250, 80, 80);
		static final int LED_WIDTH = 20;
		static final int LED_HEIGHT = 14;
		static final int LED_SPACE = 5;
		static final int LED_BASEY_OFFSET = 3;
		static final int LED_MUTE_X = 90;
		static final int LED_SOLO_X = LED_MUTE_X + LED_WIDTH + LED_SPACE;
		static final int LED_REC_X = LED_SOLO_X + LED_WIDTH + LED_SPACE;
		static final int SLIDER_HEIGHT = 12;

		private ArrayList<Rectangle> clickables = new ArrayList<Rectangle>();

		Rack rack;
		Slider sliderGain;
		Slider sliderPan;
		static final Color CM_TRACK_PANEL_HEAD_INACTIVE = new Color(110, 118, 119);
		static final Color CM_TRACK_PANEL_HEAD_ACTIVE = new Color(140, 114, 115);


		TrackPanel(Track track)
			{
			this.assocTrack = track;
			this.rack = track.proxyRack;
			int baseX = assocTrack.getBaseX();
			int by = assocTrack.getRealBaseY() + LED_BASEY_OFFSET;
			ledMute = new Led("MU", baseX + LED_MUTE_X, by, COLOR_LED_BRIGHT_GREEN);
			ledSolo = new Led("RE", baseX + LED_SOLO_X, by, COLOR_LED_BRIGHT_RED);
			ledArmed = new Led("SO", baseX + LED_REC_X, by, COLOR_LED_BRIGHT_GREEN);
			clickables.add(ledMute);
			clickables.add(ledSolo);
			clickables.add(ledArmed);
			final int sliderby = assocTrack.getRealBaseY() + Rack.PANEL_HEAD_HEIGHT + Rack.PANEL_FIRST_ROW_OFFSET;
			sliderGain = new Slider(Const.MIN_GAIN, Const.MAX_GAIN, Const.DEFAULT_GAIN, false, sliderby);
			sliderPan = new Slider(Const.MIN_PAN, Const.MAX_PAN, 0f, true, sliderby + Rack.PANEL_HEAD_HEIGHT);
			clickables.add(sliderGain);
			clickables.add(sliderPan);
			}


		public void recreate()
			{
			recreate(rack.getGBuffer());
			}


		public void recreate(Graphics2D painter)
			{
			clear(painter);
			draw(painter);
			}


		private void clear(Graphics2D painter)
			{
			final int by = assocTrack.getRealBaseY();
			final int w = rack.getWidth();
			final int h = assocTrack.height;
			painter.setColor(getCmColor(CM_TRACK_PANEL_BG));
			painter.fillRect(0, by, w, h);
			painter.setColor(assocTrack.isActive ? CM_TRACK_PANEL_HEAD_ACTIVE : CM_TRACK_PANEL_HEAD_INACTIVE);
			painter.fillRect(0, assocTrack.getRealBaseY(), rack.getWidth(), Rack.PANEL_HEAD_HEIGHT);
			painter.setColor(Color.DARK_GRAY);
			painter.drawLine(0, by + h - 2, w, by + h - 2);
			painter.setColor(Color.WHITE);
			painter.drawLine(0, by + h - 1, w, by + h - 1);
			}


		/**
		 * @param painter
		 * @param row
		 * @param color
		 * @param normValue -1.0f to +1.0f
		 * @param valueLabel
		 */

		private void drawPanelBusInOut(Graphics2D painter)
			{
			int whichChannels = assocTrack.whichChannels;
			final int by = assocTrack.getRealBaseY() + Rack.BUSES_LEDS_Y;
			final int bx = assocTrack.getBaseX() + 3;
			String sir = "" + assocTrack.busIn.getId();
			if (whichChannels == Const.MONO)
				sir += "[MO]";
			else if (whichChannels == Const.STEREO)
				sir += "[LR]";
			else if (whichChannels == Const.ONLY_LEFT_CHANNEL)
				sir += "[L]";
			else
				sir += "[R]";
			String sor = "" + assocTrack.busOut.getId();
			painter.setFont(FONT_PANEL);
			painter.setColor(Color.RED);
			painter.fillRect(bx, by, Rack.BUS_LED_WIDTH, Rack.BUS_LED_HEIGHT);
			painter.setColor(Color.GREEN);
			painter.fillRect(bx, by + Rack.PANEL_ROW_HEIGHT, Rack.BUS_LED_WIDTH, Rack.BUS_LED_HEIGHT);
			painter.setColor(Color.BLACK);
			painter.drawString(sir, bx + Rack.BUS_LED_WIDTH + 3, by + Rack.FONT_PANEL_Y_OFFSET);
			painter.drawString(sor, bx + Rack.BUS_LED_WIDTH + 3, by + Rack.PANEL_ROW_HEIGHT + Rack.FONT_PANEL_Y_OFFSET);
			}


		private void drawPanelTrackName(Graphics2D canvas)
			{
			String stn = "" + assocTrack.trackNumber;
			canvas.setFont(Rack.FONT_TRACK_NAME);
			canvas.setColor(Color.DARK_GRAY);
			canvas.drawString(stn, assocTrack.getBaseX() + 4, assocTrack.getRealBaseY() + 14);
			canvas.drawString(assocTrack.name, assocTrack.getBaseX() + 15, assocTrack.getRealBaseY() + 15);
			canvas.setColor(Color.WHITE);
			canvas.drawString(stn, assocTrack.getBaseX() + 5, assocTrack.getRealBaseY() + 15);
			canvas.drawString(assocTrack.name, assocTrack.getBaseX() + 15, assocTrack.getRealBaseY() + 15);
			}


		void drawPanelLeds()
			{
			ledMute.draw();
			ledSolo.draw();
			ledArmed.draw();
			}


		public void draw(Graphics2D painter)
			{
			drawPanelHead(painter);
			sliderGain.setLevel(assocTrack.gain);
			sliderGain.draw();
			sliderPan.draw();
			sliderPan.setLevel(assocTrack.pan);
			drawPanelBusInOut(painter);
			}


		private void drawPanelHead(Graphics2D canvas)
			{
			drawPanelLeds();
			drawPanelTrackName(canvas);
			}


		public Rectangle getClickableUnder(int x, int y)
			{
			for (Rectangle clickable : clickables)
				if (clickable.contains(x, y))
					return clickable;
			return null;
			}


		public void clicked(int x, int y)
			{
			Rectangle clickable = getClickableUnder(x, y);
			if (clickable != null)
				clickable.clicked(x, y);
			}


		public void doubleClicked(int x, int y)
			{
			Rectangle clickable = getClickableUnder(x, y);
			if (clickable != null)
				{
				if (clickable instanceof Slider)
					((Slider) clickable).reset();
				}
			}

		class Led extends Rectangle
			{
				Led(String label, int baseX, int baseY, Color color)
					{
					super(rack, baseX, baseY, LED_WIDTH, LED_HEIGHT);
					this.colorOn = color;
					this.colorOff = colorOn.darker().darker().darker();
					this.label = label;
					}
				String label;
				boolean state;
				Color colorOn;
				Color colorOff;


				public final void toggle()
					{
					this.state = !this.state;
					draw();
					repaint();
					}


				@Override
				public void draw()
					{
					Graphics2D painter = canvas.getGBuffer();
					painter.setColor(state ? colorOn : colorOff);
					painter.fill3DRect(baseX, baseY, LED_WIDTH, LED_HEIGHT, true);
					painter.setColor(state ? colorOff : Color.BLACK);
					painter.setFont(FONT_LEDS);
					painter.drawString(label, baseX + 1, baseY + 11);
					}


				@Override
				public void clicked(int x, int y)
					{
					toggle();
					}

			}

		class Slider extends Rectangle
			{
				Slider(float min, float max, float defaultValue, boolean fromHalf, int baseY)
					{
					super(rack, 0, baseY, 0, SLIDER_HEIGHT);
					this.fromHalf = fromHalf;
					this.baseY = baseY;
					this.min = min;
					this.max = max;
					this.defaultValue = defaultValue;
					}

				float min;
				float max;
				float level;
				float defaultValue;
				boolean fromHalf;


				void reset()
					{
					setLevel(defaultValue);
					draw();
					repaint();
					}


				void setLevel(float level)
					{
					this.level = level;
					}


				private final void recalcGeometry()
					{
					int rw = rack.getWidth();
					width = (int) (rw * Rack.SLIDER_WIDTH_FACTOR);
					baseX = rw - width - 10;
					}


				@Override
				public void draw()
					{
					Graphics2D painter = canvas.getGBuffer();
					recalcGeometry();
					painter.setColor(getCmColor(CM_TRACK_PANEL_BG));
					painter.fillRect(baseX, baseY, width, Rack.PANEL_ROW_HEIGHT);
					painter.setColor(Color.LIGHT_GRAY);
					painter.fillRect(baseX, baseY, width, height);
					painter.setColor(Color.BLACK);
					painter.drawLine(baseX, baseY, baseX + width, baseY);
					painter.drawLine(baseX, baseY, baseX, baseY + height);
					painter.setColor(Color.WHITE);
					painter.drawLine(baseX, baseY + height, baseX + width, baseY + height);
					painter.drawLine(baseX + width, baseY, baseX + width, baseY + height);
					Color color = Color.BLACK;
					NumberFormat nf = NumberFormat.getInstance();
					nf.setMaximumFractionDigits(1);
					String valueLabel = nf.format(level);
					float delta = min;
					float norm = (level - delta) / (max - min);
					if (fromHalf)
						color = new Color(0, 55 + (int) (100 * norm), 100 - (int) (100 * norm));
					else
						color = new Color((int) (255 * norm), 0, 0);
					painter.setColor(color);
					if (fromHalf)
						{
						int halfSlide = width / 2;
						int dw = (int) (Math.abs(level) * halfSlide);
						if (level > 0)
							painter.fillRect(baseX + 1 + halfSlide, baseY + 1, dw, height - 1);
						else
							painter.fillRect(baseX + 1 + halfSlide - dw, baseY + 1, dw, height - 1);
						}
					else
						{
						int w = (int) (norm * (width - 2));
						painter.fillRect(baseX + 1, baseY + 1, w, height - 1);
						}
					painter.setFont(FONT_PANEL);
					final int sbx = baseX + (width / 2) - 10;
					painter.setColor(Color.BLACK);
					painter.drawString(valueLabel, sbx, baseY + 10);
					painter.setColor(Color.WHITE);
					painter.drawString(valueLabel, sbx + 1, baseY + 11);
					}


				@Override
				public void clicked(int x, int y)
					{
					int delta = x - baseX;
					float factor = (float) delta / width;
					level = (max - min) * factor + min;
					draw();
					repaint();
					}

			}

	}
