package org.protux.core.engines.audio;

import static org.protux.core.Debugger.PMESG;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;

import javax.swing.JComponent;
import javax.swing.JPanel;

import org.protux.core.Const;
import org.protux.core.Protux;
import org.protux.core.gui.Button;

public abstract class Bus extends JPanel implements ActionListener
	{

		public static enum Direction
			{
				IN, OUT
			}

		protected static final Dimension MIN_SIZE = new Dimension(30, 100);
		protected static final Dimension PREF_SIZE = new Dimension(100, 100);
		protected static final Dimension MAX_SIZE = new Dimension(120, Const.TOP_PANEL_HEIGHT);

		public BusVU busVU;
		protected boolean stereo = true;
		protected boolean isNullBus = false;
		protected int id;
		protected boolean stereoBus;
		protected String busName;
		protected Button butBus;
		protected Button butEFX;
		protected boolean muted;
		Direction direction;


		public Bus(int id, boolean nullBus, String busName, Direction direction)
			{
			super();

			setLayout(new BorderLayout());
			setBackground(Color.YELLOW);
			setMinimumSize(MIN_SIZE);
			setPreferredSize(PREF_SIZE);

			this.direction = direction;
			this.id = id;
			this.muted = false;
			this.isNullBus = nullBus;
			this.stereoBus = false;
			this.busName = busName;

			busVU = new BusVU(this);
			butBus = new Button(this, Protux.getIcon("ledOn"));
			butEFX = new Button(this, null);

			butBus.setText(isNullBus ? "NULL:" + id : "" + id);
			butBus.setToolTipText(busName);
			butEFX.setText("FX");

			JPanel butPanel = new JPanel(new BorderLayout());
			butPanel.setMinimumSize(new Dimension(10,Button.BUT_BUS_SIZE.height));
			butPanel.setPreferredSize(new Dimension(120,Button.BUT_BUS_SIZE.height));
			
			butPanel.add(butBus,BorderLayout.WEST);
			butPanel.add(butEFX,BorderLayout.EAST);
			ArrayList<JComponent> extraTopComponents = getExtraButtons();
			for (JComponent comp : extraTopComponents)
				{
				butPanel.add(comp,BorderLayout.CENTER);
				}
			
			this.add(butPanel,BorderLayout.NORTH);
			this.add(busVU,BorderLayout.CENTER);
			}


		protected abstract ArrayList<JComponent> getExtraButtons();


		public String getFullBusName()
			{
			return busName;
			}


		public void setNull(boolean b)
			{
			isNullBus = b;
			}


		public String getId()
			{
			return "" + id;
			}


		public void recreate()
			{
			log("Recreating ...");
			busVU.recreate();
			}


		private void log(String string)
			{
			PMESG(2, "[ Bus " + id + " ] - " + string);
			}


		public final boolean isNullBus()
			{
			return isNullBus;
			}


		public void actionPerformed(ActionEvent e)
			{
			if (e.getSource() == butBus)
				{
				muted = !muted;
				if (muted)
					butBus.setIcon(Protux.getIcon("ledOff"));
				else
					butBus.setIcon(Protux.getIcon("ledOn"));
				}
			}


		public String getBusName()
			{
			return busName;
			}


		public void setBusName(String busName)
			{
			this.busName = busName;
			}


		public Direction getDirection()
			{
			return direction;
			}


		public void setDirection(Direction direction)
			{
			this.direction = direction;
			}


		public void setId(int id)
			{
			this.id = id;
			}

	}
