package org.protux.core.engines.audio;

import static org.protux.core.Debugger.PMESG;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;

import org.protux.core.Util;
import org.protux.core.gui.FastPanel;
import org.protux.core.gui.ProjectManager;

public class BusVU extends FastPanel
	{

		protected Bus assocBus;
		protected float[][] tailDeltaYIn;
		protected float[][] tailDeltaYOut;
		protected float[][] levelOut;
		protected float[][] levelIn;
		protected boolean activeTail;

		protected static final Color BG_COLOR = new Color(218, 218, 220);
		protected static final Color BUS_LEVEL_CLEAR_COLOR = new Color(97, 93, 125);
		protected static final Color VU_BG_COLOR = new Color(203, 203, 205);
		protected static final Font DEFAULT_BUS_FONT = new Font("Helvetica", 0, 8);
		protected static final Font DEFAULT_BUS_BIG_FONT = new Font("Helvetica", Font.BOLD, 12);
		protected static final int VU_BASE_Y = 20;
		protected static final int CHANNEL_0 = 6;


		BusVU(Bus assocBus)
			{
			super("Canvas for " + assocBus.getBusName(),new Dimension(10, 10),BG_COLOR);
			this.assocBus = assocBus;
			}


		@Override
		public void recreate(Graphics2D painter)
			{
			log("Recreating...");
			if (ProjectManager.getCurrentProject() != null)
				{
				clear(painter);
				drawBase(painter);
				drawVus(painter);
				drawPeaks(painter, new float[] { 0f, 0f }, new float[] { 0f, 0f });
				}
			else
				drawDumbX(painter);
			}


		void drawBase(Graphics2D painter)
			{
			final int baseWidth = getBaseWidth();
			painter.setColor(BG_COLOR);
			painter.fill3DRect(0, 0, baseWidth, getHeight(), true);
			}


		protected void drawVus(Graphics2D painter)
			{
			final int vuWidth = getVuChannelWidth();
			final int vuHeight = getVuHeight();
			final int channels = assocBus.stereo ? 2 : 1;
			final int vuby =  VU_BASE_Y;
			for (int i = 0; i < channels; i++)
				{
				painter.setColor(VU_BG_COLOR);
				int chanbx = getVuChannelBaseX(i);
				painter.fillRect(chanbx, vuby, vuWidth, vuHeight);
				painter.setColor(Color.GRAY);
				painter.drawLine(chanbx - 1, vuby - 1, chanbx - 1, vuby + vuHeight - 1);
				painter.drawLine(chanbx - 1, vuby - 1, chanbx + vuWidth, vuby - 1);
				painter.setColor(Color.WHITE);
				painter.drawLine(chanbx, vuby, chanbx, vuby + vuHeight);
				painter.drawLine(chanbx, vuby, chanbx + vuWidth, vuby);
				painter.setColor(Color.DARK_GRAY);
				painter.drawLine(chanbx + vuWidth, vuby, chanbx + vuWidth, vuby + vuHeight);
				painter.drawLine(chanbx - 1, vuby + vuHeight, chanbx + vuWidth, vuby + vuHeight);
				painter.setColor(Color.WHITE);
				painter.drawLine(chanbx + vuWidth + 1, vuby, chanbx + vuWidth + 1, vuby + vuHeight + 1);
				painter.drawLine(chanbx - 1, vuby + vuHeight + 1, chanbx + vuWidth, vuby + vuHeight + 1);
				}
			}


		protected int getVuHeight()
			{
			return (int) (getBusHeight() * 0.9) - VU_BASE_Y;
			}


		protected int getVuChannelWidth()
			{
			int ch = getchannels();
			int w = getBaseWidth() - (CHANNEL_0 * 2);
			int vw = (w / ch) / 2;
			return vw;
			}


		protected int getVuChannelBaseX(int i)
			{
			int w = getBaseWidth() - (CHANNEL_0 * 2);
			int ch = getchannels();
			int vch = getVuChannelWidth();
			int p = w / (2 * ch) - vch / 2;
			return 0 + CHANNEL_0 + (i * (w / ch)) + p;
			}


		private int getBaseWidth()
			{
			return getWidth() - (0 * 2);
			}


		protected final int getchannels()
			{
			return (assocBus.stereo ? 2 : 1);
			}


		protected void drawPeaks(Graphics2D painter, float[] currChannelNormPeak, float[] tailDeltaY)
			{
			final int levelRange = getVuHeight() - 2;
			final int channels = assocBus.stereo ? 2 : 1;
			final int vuBaseY =  VU_BASE_Y;
			final int vuWidth = getVuChannelWidth();

			// check overload
			if (currChannelNormPeak != null) // for tail draws , levelBuffer will be null on purpose, that's why there is this test
				{
				activeTail = true;
				for (int k = 0; k < channels; k++)
					{
					float normPeak = currChannelNormPeak[k];
					final int vuBaseX = getVuChannelBaseX(k);
					if (normPeak > 0.99f)
						{
						float saturationDB = Util.coefficient_to_dB(normPeak);
						painter.setColor(Color.RED);
						painter.fillRect(vuBaseX, vuBaseY - 12, vuWidth, 6);
						painter.setColor(Color.BLACK);
						String ssat = "" + saturationDB;
						painter.setFont(DEFAULT_BUS_FONT);
						painter.drawString(ssat, vuBaseX, vuBaseY - 5);
						}
					}
				}

			// draw levels
			for (int k = 0; k < channels; k++)
				{
				final int vuBaseX = getVuChannelBaseX(k);
				float normPeak = currChannelNormPeak[k];
				int thisLevelDeltaY;
				if (currChannelNormPeak != null)
					{
					thisLevelDeltaY = (int) (normPeak * levelRange);
					}
				else
					{
					thisLevelDeltaY = 0;
					}

				// decrease current tail level
				float sf = tailDeltaY[k];
				sf *= 0.60f;
				sf = (sf < 1.0f) ? 0.0f : sf;
				tailDeltaY[k] = sf;
				if (thisLevelDeltaY < sf)
					thisLevelDeltaY = (int) sf;
				else
					tailDeltaY[k] = thisLevelDeltaY;

				// First clear where is not used
				painter.setColor(BUS_LEVEL_CLEAR_COLOR);
				painter.fillRect(vuBaseX + 1, vuBaseY + 1, vuWidth - 2, levelRange - thisLevelDeltaY);

				// now the new level
				final int hlr = levelRange / 2;
				for (int i = 0; i < thisLevelDeltaY; i++)
					{
					int y = vuBaseY + levelRange - i + 1;
					int cR = (int) ((i < hlr) ? ((float) i / hlr) * 255 : 255);
					int cG = (int) ((i < hlr) ? 255 : 255 * ((float) (levelRange - i) / hlr));
					cG = cG > 255 ? 255 : cG;
					cR = cG < 0 ? 0 : cR;
					painter.setColor(new Color(cR, cG, 0));
					painter.drawLine(vuBaseX + 1, y, vuBaseX + vuWidth - 2, y);
					}
				}
			}


		protected int getBusHeight()
			{
			return getHeight();
			}


		protected void log(String string)
			{
			PMESG(2, "[ BusCanvas for Bus " + assocBus.getId() + " ] - " + string);
			}


		public void mouseClicked(MouseEvent e)
			{
			}


		public void mousePressed(MouseEvent e)
			{
			}


		public void mouseReleased(MouseEvent e)
			{
			}


		public void mouseEntered(MouseEvent e)
			{
			}


		public void mouseExited(MouseEvent e)
			{
			}

		/*
		 * void draw_active_tails() { // draw out "Tail" peaks (MAYBE it should be done also for input...) activeTail = false; int pos = 0; for (int busOut = 0; busOut < AudioDeviceMapper.get_total_buses_out(); busOut++) { if (!app.getCurrentProject().get_current_song().mixer.valid_playback_bus(busOut)) { continue; } int channels = assocBus.stereo ?
		 * 2 : 1; boolean thereIsTail = false; for (int k = 0; k < channels; k++) { if (tailDeltaYOut[busOut][k] > 1.0f) thereIsTail = true; } if (thereIsTail) { activeTail = true; draw_bus_vu_peaks(busOutArea, pos * (BUS_WIDTH + SPACE_BETWEEN_BUSES), 0, channels, AudioDeviceMapper.PLAYBACK, null, tailDeltaYOut[busOut]); } pos++; } if (activeTail ==
		 * null) { recreate(); } }
		 */

	}
