package org.protux.core.engines.audio;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PWARN;

import java.awt.Component;
import java.util.ArrayList;

import javax.sound.sampled.Mixer;
import javax.swing.JComponent;

import org.protux.core.Const;
import org.protux.core.exceptions.HardBusException;

public class HardBus extends Bus
	{

		private static final int CLOSED = 0;
		private static final int PLAYING_TO_MONO = 1;
		private static final int PLAYING_TO_STEREO = 2;
		private static final int PLAYING_TO_LEFT_CHANNEL = 3;
		private static final int PLAYING_TO_RIGHT_CHANNEL = 4;
		private static final int CAPTURING_FROM_MONO = 5;
		private static final int CAPTURING_FROM_STEREO = 6;
		private static final int CAPTURING_FROM_LEFT_CHANNEL = 7;
		private static final int CAPTURING_FROM_RIGHT_CHANNEL = 8;

		private Mixer mixer;
		private int transferError;
		private int bytesToTransfer;
		private int stream; // !< CAPTURE or PLAYBACK
		private int status = CLOSED;
		private int blockSize;
		private int transferSize;
		private int partialPeriodSize;
		private int periodSize;
		private int format;
		private float fragmentsPerSecond;
		private float[] transferBuffer = new float[4096]; // FIXME
		private float[] levels = null;
		protected static int currentHardId = 0;

		AudioBusTransferResult transferResult = new AudioBusTransferResult();
		String descriptor; // !< those "hw:x,x" ALSA descriptors


		public HardBus(Bus.Direction direction)
			{
			this((Mixer)null,direction);
			}


		public HardBus(Mixer mixer,Bus.Direction direction)
			{
			super(++currentHardId,(mixer==null),(mixer!=null)?mixer.getMixerInfo().getName():"NULL:"+currentHardId,direction);
			this.mixer = mixer;
			}


		int open(int Rate, int pBitDepht, int whichChannels, boolean block, float pSuggestedFragmentsPerSecond)
			{
			fragmentsPerSecond = pSuggestedFragmentsPerSecond;
			int channels; // number of channels effectively open ( not the access code MONO,STEREO,ONLY_LEFT,ONLY_RIGHT, AND SO ON...)
			int pRate = Rate;
			int r = 0;
			// String busFullReference = get_short_name() + " (" + get_real_descriptor() + ")";
			final short[] cBusFullReference = null;// MIG busFullReference;
			// PMESG(2,"Opening %s bus : %s ...",stream==AudioDeviceMapper.PLAYBACK?"PLAYBACK":"CAPTURE",cBusFullReference);
			if (stream == Const.CAPTURE)
				{
				if (isAllocatedForCapture())
					{
					// PMESG(2,"Bus seems to be allocated for capture. No need to re-open it.");

					return 1;
					}
				if (stereoBus)
					{
					if ((whichChannels == Const.ONLY_LEFT_CHANNEL) && (status == CAPTURING_FROM_LEFT_CHANNEL))
						{
						PERROR("One or more channels of bus %s is already in use." + cBusFullReference);

						return Const.ERROR_BUSY_BUS;
						}
					if ((whichChannels == Const.ONLY_RIGHT_CHANNEL) && (status == CAPTURING_FROM_RIGHT_CHANNEL))
						{
						PERROR("One or more channels of bus %s is already in use." + cBusFullReference);

						return Const.ERROR_BUSY_BUS;
						}
					if ((whichChannels == Const.STEREO) && (status == CAPTURING_FROM_LEFT_CHANNEL || status == CAPTURING_FROM_RIGHT_CHANNEL))
						{
						PERROR("One or more channels of bus %s is already in use." + cBusFullReference);

						return Const.ERROR_BUSY_BUS;
						}
					channels = 2;
					}
				else
					{
					if (whichChannels == Const.STEREO)
						{
						PERROR("Bus %s does not support STEREO stuff" + cBusFullReference);

						return Const.ERROR_MONO_BUS;
						}
					channels = 1;
					}
				}
			else
				{
				if (isAllocatedForPlayback())
					{
					if ((whichChannels != Const.MONO) && (status != PLAYING_TO_MONO))
						{
						// PMESG(2,"Bus seems to be allocated for playback. No need to re-open it.");

						return 0;
						}
					PERROR("Bus %s is busy. " + cBusFullReference);

					return -1;
					}
				else
					{
					if (whichChannels == Const.MONO)
						channels = 1;
					else
						// STEREO, ONLY_LEFT or ONLY_RIGHT
						channels = 2;
					}
				}

			// not open yet ? so let's try and open it ! :-)
			if (false) // MIG !valid_mode(pRate, pBitDepht, channels))
				{
				PERROR("Invalid mode for bus %s : %d Hz, %d bits, %s"+cBusFullReference+pRate+pBitDepht+ (channels==1?"MONO":"STEREO"));
				return Const.ERROR_INVALID_BUS;
				}
			if (stream == Const.PLAYBACK)
				{
				if (false) // MIG snd_pcm_open(&pcmHandle, (get_real_descriptor()), SND_PCM_STREAM_PLAYBACK, block ? 0 : SND_PCM_NONBLOCK ) < 0 )
					{
					if (false) // MIG (errno == EBUSY)
						{
						PERROR("The playback device %s is allready in use by another application, please stop this application and restart the program" + (get_real_descriptor()));
						return Const.ERROR_CANNOT_OPEN_PLAYBACK_BUS;
						}
					// MIG PERROR("Cannot open bus %s for PLAYBACK . ALSA ERROR : %s",cBusFullReference,snd_strerror(r));

					return Const.ERROR_CANNOT_OPEN_PLAYBACK_BUS;
					}
				}
			else if (stream == Const.CAPTURE)
				{
				if (false) // MIG snd_pcm_open(&pcmHandle, (get_real_descriptor()), SND_PCM_STREAM_CAPTURE, 0) < 0)
					{
					if (false) // MIG (errno == EBUSY)
						{
						PERROR("The capture device %s is allready in use by another application, please stop this application and restart the program " + (get_real_descriptor()));
						return Const.ERROR_CANNOT_OPEN_PLAYBACK_BUS;
						}
					// MIG PERROR("Cannot open %s for probing AudioDeviceMapper.CAPTURE capabilities",(get_real_descriptor()));

					return Const.ERROR_CANNOT_OPEN_BUS;
					}
				}
			// some parameters declaration/initialization
			int dir;
			long triedPeriodSize;
			int numPeriods = 2; // IMPROVE ME (?)
			blockSize = channels * (pBitDepht / 8);
			int chunckSize = pRate * blockSize;
			int fragmentSize = (int) (chunckSize / fragmentsPerSecond);
			// MIG snd_pcm_uframes_t minPeriodSize, maxPeriodSize;

			// starting the ALSA settings
			// MIG snd_pcm_hw_params_alloca(&hwparams);

			if (false) // MIG snd_pcm_hw_params_any(pcmHandle, hwparams) < 0)
				{
				// MIG PERROR("Cannot configure bus %s. ALSA ERROR : %s ",cBusFullReference,snd_strerror(r));
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			// set the interleaved read/write format
			if (false) // MIG snd_pcm_hw_params_set_access(pcmHandle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
				{
				// MIG PERROR("Invalid access format to bus %s: %s",cBusFullReference,snd_strerror(r));
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			// set the sample format
			format = HardBusMapper.get_format(pBitDepht);
			if (false)// MIG format == SND_PCM_FORMAT_UNKNOWN)
				{
				// MIG PERROR("Cannot define pcm format to %d bits in bus %s",pBitDepht, cBusFullReference);
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			if (false) // MIG snd_pcm_hw_params_set_format(pcmHandle, hwparams, format) < 0)
				{
				// MIG PERROR("Invalid sample format to bus %s : %s",cBusFullReference,snd_strerror(r));
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			int channelsMax = 0;
			if (false) // MIG snd_pcm_hw_params_get_channels_max (hwparams, &channelsMax) < 0)
				{
				PERROR("Cannot detect max number of channels");
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			// If using the plughw device, channelsMax can be rather large, if so, set it to a sane value
			if (channelsMax > 64)
				{
				PWARN("You appear to use the plughw device, setting number of channels to a sane value (2)");
				channelsMax = 2;
				}

			// set the count of channels
			if (false) // MIG snd_pcm_hw_params_set_channels(pcmHandle, hwparams, channelsMax) < 0)
				{
				// MIG PERROR("Invalid number of channels (%d) to bus %s : %s",channels, cBusFullReference, snd_strerror(r));
				// MIG //MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			// set the stream rate
			if (false) // MIG snd_pcm_hw_params_set_rate(pcmHandle, hwparams, pRate, 0) < 0)
				{
				// MIG PERROR("Invalid sample rate (%d Hz) to bus %s . ALSA ERROR : : " + pRate, cBusFullReference, snd_strerror(r));
				// MIG ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			// the period/buffer setup process
			if (false) // MIG snd_pcm_hw_params_set_periods(pcmHandle, hwparams, numPeriods, 0) < 0)
				{
				// MIG PERROR("Cannot set periods count to %d on bus %s. ALSA ERROR : : " + numPeriods, cBusFullReference, snd_strerror(r));
				// MIG //MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			if (false) // MIG snd_pcm_hw_params_get_period_size_min (hwparams, &minPeriodSize, &dir) < 0 )
				{
				PERROR("Couldn't detect minimum period size");
				}
			if (false) // MIG snd_pcm_hw_params_get_period_size_max (hwparams, &maxPeriodSize, &dir) < 0 )
				{
				PERROR("Couldn't detect maximum period size");
				}

			// PMESG(2,"Detected min/max period size : %ld - %ld", minPeriodSize, maxPeriodSize);

			periodSize = fragmentSize / blockSize;

			if (false)// MIG (periodSize < minPeriodSize) || (periodSize > maxPeriodSize) )
				{
				// MIG PERROR("This periodSize %ld is not within the max/min (%ld/%ld) allowed periodSize!", periodSize, maxPeriodSize, minPeriodSize);
				// MIG ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}
			triedPeriodSize = periodSize;

			if (stream == Const.PLAYBACK)
				{
				// PMESG(2,"Trying OPEN PLAYBACK BUS with these values: fps=%2.3f fragmentSize=%d periods=%d periodSize=%ld", fragmentsPerSecond, fragmentSize, numPeriods, periodSize);
				}
			else
				{
				// PMESG(2,"Trying OPEN CAPTURE BUS with these values: fps=%2.3f fragmentSize=%d periods=%d periodSize=%ld",fragmentsPerSecond, fragmentSize, numPeriods, periodSize);
				}

			if (false) // MIG snd_pcm_hw_params_set_period_size (pcmHandle, hwparams, periodSize, 0) < 0)
				{
				// MIG PERROR("setting period size to %ld on bus %s. ALSA ERROR : : " + periodSize,cBusFullReference, snd_strerror(r));
				// MIG //MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			if (false) // MIG snd_pcm_hw_params_set_buffer_size(pcmHandle, hwparams, numPeriods * periodSize) < 0 )
				{
				// MIG PERROR("Cannot set buffer size to %ld on bus %s. ALSA ERROR : : " + numPeriods * periodSize, cBusFullReference, snd_strerror(r));
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			if (false) // MIG snd_pcm_hw_params_get_period_size (hwparams, &periodSize, &dir) < 0)
				{
				PERROR("Can't get period size");
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			if (triedPeriodSize != periodSize)
				{
				// PMESG(2,"Could not set periodSize to %ld . ALSA let me use only : %ld . Continuing anyway...",triedPeriodSize, periodSize);
				}

			// write the parameters to descriptor
			if (false) // MIG snd_pcm_hw_params(pcmHandle, hwparams) < 0)
				{
				// MIG PERROR("Unable to set hw params : %s",snd_strerror(r));
				// MIG goto ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION;
				}

			transferSize = periodSize * blockSize;
			transferBuffer = new float[transferSize];
			// MIG bzero(transferBuffer, transferSize);

			// set the actual values for fragmentSize and fps
			if (fragmentSize != transferSize)
				{
				fragmentSize = transferSize;
				fragmentsPerSecond = (float) chunckSize / fragmentSize;
				// PMESG(2,"Fixing fragmentSize and fps to match transferSize : fragmentSize=%d fps=%2.3f",fragmentSize,fragmentsPerSecond);
				}

			if (stream == Const.PLAYBACK)
				{
				if (channels == 1)
					status = PLAYING_TO_MONO;
				else
					status = PLAYING_TO_STEREO; // shouldnt it have status for "playing to left only/ playing to right only" situatiosn ???
				}
			else
				{
				if (whichChannels == Const.ONLY_LEFT_CHANNEL)
					status = CAPTURING_FROM_LEFT_CHANNEL;
				else if (whichChannels == Const.ONLY_RIGHT_CHANNEL)
					status = CAPTURING_FROM_RIGHT_CHANNEL;
				else if (whichChannels == Const.STEREO)
					status = CAPTURING_FROM_STEREO;
				else if (whichChannels == Const.MONO)
					status = CAPTURING_FROM_MONO;
				}

			if (status != CLOSED)
				{
				boolean inOutFlag = (stream == Const.CAPTURE) ? false : false;// MIG FIX "IN":"OUT";
				// MIG PMESG(2,"BUS %s %s OPEN : %d Hz, %s, %d bits periodSize=%d transferSize=%d",inOutFlag,cBusFullReference,pRate ,channels==1?"MONO":"STEREO",pBitDepht,(int) periodSize,(int) transferSize);
				}
			else
				{
				PERROR("Status is still closed, at this point ??!?!?!");
				}

			levels = new float[channels];

			ALSA_ERROR_AFTER_OPEN_AND_HWPARAMS_ALLOCATION:

			if (false) // MIG snd_pcm_close(pcmHandle)<0)
				{
				PERROR("Could not close pcmHandle after previous ALSA error");
				}
			// SHOULDNT I CALL snd_pcm_hw_params_free(hwparams) now ?
			// If I try I get a segfault.. but why.. ?
			// No, you shouldn't. you only have to do this when you
			// allocate memory with malloc, but we use alloca. When snd_pcm_close is called, the
			return r; // hwparams is automatically freed By Remon
			}


		AudioBusTransferResult transfer() throws HardBusException
			{
			return transfer(-1);
			}


		AudioBusTransferResult transfer(int bytesToTransfer)  throws HardBusException
			{
			transferResult.clean();
			int periodsToTransfer;
			// FIXME ALSA uses an long for its periodSize value, adopt this all classes using this value, or
			// comparing to it, to make comparisons robuster!
			if ((bytesToTransfer == -1) || (bytesToTransfer > (int) periodSize)) // -1 means ALL bytes
				{
				periodsToTransfer = periodSize;

				}
			else
				{
				periodsToTransfer = bytesToTransfer / blockSize;
				}
			int periodsTransferred = 0;
			float[] ptr = transferBuffer;
			while (periodsToTransfer > 0)
				{
				if (stream == Const.CAPTURE)
					{
					// MIG periodsTransferred = snd_pcm_readi(pcmHandle, ptr, periodsToTransfer);
					}
				else
					{
					// MIG periodsTransferred = snd_pcm_writei(pcmHandle, ptr, periodsToTransfer);
					}
				if (periodsTransferred >= 0)
					{
					// MIG ptr += periodsTransferred * blockSize;
					transferResult.bytesTransferred += periodsTransferred * blockSize;
					periodsToTransfer -= periodsTransferred;
					}
				else
					{
					int err = 0;
					if (false)// MIG periodsTransferred == -EAGAIN) // overrun
						{
						PWARN("Buffer OVERRUN, retrying ...");
						// MIG snd_pcm_wait(pcmHandle, 1000);
						continue;
						}
					else if (false)// MIG periodsTransferred == -EPIPE) // underrun
						{
						PWARN("Buffer UNDERRUN, recovering ...");
						// MIG snd_pcm_status_t *bus_status;
						// MIG snd_pcm_status_alloca(&bus_status);
						// MIG snd_pcm_status(pcmHandle, bus_status);
						if (false) // MIG snd_pcm_status_get_state(bus_status) == SND_PCM_STATE_XRUN)
							{
							/*
							 * MIG struct timeval now, diff, tstamp; gettimeofday(&now, 0); snd_pcm_status_get_trigger_tstamp(bus_status, &tstamp); timersub(&now, &tstamp, &diff); PWARN("Buffer UNDERRUN of at least %.3f msecs, recovering ...", diff.tv_sec * 1000 + diff.tv_usec / 1000.0);
							 */
							}
						// MIG err = snd_pcm_prepare(pcmHandle);
						if (err < 0)
							{
							// MIG PERROR("Can't recover : %s",snd_strerror(periodsTransferred));
							transferResult.error = Const.ERROR_UNDERRUN;
							transferResult.xrun = -1;
							}
						}
					else
						// unknown error. try to recover anyway
						{
						PWARN("Unknown problem in transfer. Trying to recover anyway...");
						// MIG err = snd_pcm_prepare(pcmHandle);
						if (err < 0)
							{
							// MIG PERROR("Can't recover : %s\n", snd_strerror(err));
							}
						}
					break;
					}
				}
			transferResult.message = HardBusMapper.get_error_mesg(transferResult.error);

			return transferResult;
			}


		void clean()
			{

			// MIG bzero(transferBuffer, transferSize);

			}


		int close()
			{

			if (status == CLOSED)
				{
				// PMESG(2,"Bus %s Seems to be already closed. Ignoring close call...",get_real_descriptor());

				return 1;
				}
			if (transferBuffer != null)
				{
				// delete transferBuffer;
				}
			transferBuffer = null;
			if (false) // MIG snd_pcm_close(pcmHandle) <0 )
				{
				PERROR("Cannot close Bus ");
				return -1;
				}
			status = CLOSED;
			// PMESG(2," BUS %s CLOSED", get_full_name() );
			return 1;
			}


		int getTransferSize()
			{
			return transferSize;
			}


		float[] getTransferBuffer(int whichChannels)
			{
			return transferBuffer; // IMPROVE ME.. Mustux consider whichChannels some way...
			}


		boolean is_stereo()
			{
			return stereoBus;
			}


		@Deprecated 
		String get_short_name()
			{
			return ""+super.id;//AudioDeviceMapper.get_card_name(parentCardId) + " - " + br;
			}


		float getFps()
			{
			return fragmentsPerSecond;
			}


		boolean isAllocatedForPlayback()
			{
			boolean b = ((status == PLAYING_TO_STEREO) || (status == PLAYING_TO_LEFT_CHANNEL) || (status == PLAYING_TO_RIGHT_CHANNEL));
			return b;
			}


		boolean isAllocatedForCapture()
			{
			return ((status == CAPTURING_FROM_RIGHT_CHANNEL) || (status == CAPTURING_FROM_LEFT_CHANNEL) || (status == CAPTURING_FROM_STEREO));
			}


		@Deprecated 
		String get_real_descriptor()
			{
			if (HardBusMapper.get_plughw_use())
				{
				return "plug" + descriptor;
				}
			return descriptor;
			}


		float[] get_levels() // REVIEW
			{
			int channels = is_stereo() ? 2 : 1;

			if ((transferSize < 0) || (transferBuffer == null)) // This can happen at the time the buffers are cleared! of course, it's a change of 1:100000 or something but it happens!!
				return levels; // LG: but the buffers are only bzero'd, not destroyed. They are always valid (as long the bus is open, ofcourse). They are invalid only when bus is closed, but nobody is supposed to call get_levels from a closed bus... (RFC)
			float[] p = transferBuffer;
			float[] max = new float[channels];
			for (int k = 0; k < channels; k++)
				max[k] = 0.0f;
			short sho = 0;
			int SIZE_OF_SHORT = 2;
			int totalsamples = transferSize / (channels * SIZE_OF_SHORT);
			for (int i = 0; i < totalsamples; i++)
				{
				for (int k = 0; k < channels; k++)
					{
					// MIG sho = (short) abs(*(short*)p);
					if (sho > max[k])
						{
						max[k] = sho;
						}
					// MIG p+=SIZE_OF_SHORT;
					}
				}
			for (int k = 0; k < channels; k++)
				levels[k] = max[k];
			// delete max;
			return levels;
			}


		@Override
		protected ArrayList<JComponent> getExtraButtons()
			{
			return new ArrayList<JComponent>();
			}

	}
