package org.protux.core.engines.audio;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;
import static org.protux.core.Debugger.PWARN;

import java.util.ArrayList;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Line;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;

import org.protux.core.Const;
import org.protux.core.engines.audio.Bus.Direction;
import org.protux.core.exceptions.DeviceInitializationException;
import org.protux.core.exceptions.HardBusException;
import org.protux.core.exceptions.ProbingException;

public class HardBusMapper
	{

		static ArrayList<HardBus> hardBusInList = new ArrayList<HardBus>();
		static ArrayList<HardBus> hardBusOutList = new ArrayList<HardBus>();
		static boolean initialized = false;
		static boolean tryPlugHw = false;
		static Mixer.Info[] info;

		private static float suggestedFps = 20.f;
		private static ArrayList<AudioCard> cardList = new ArrayList<AudioCard>();


		public static void init()
			{
			if (initialized)
				return;
			info = AudioSystem.getMixerInfo();
			if (true) // UNCO , this is just to force null hard buses (to test them) info.length == 0)
				{
				warn("No mixers available. Assuming In/Out null hard buses");
				HardBus nbi = new HardBus(Bus.Direction.IN);
				HardBus nbo = new HardBus(Bus.Direction.OUT);
				nbi.setNull(true);
				nbo.setNull(true);
				hardBusInList.add(nbi);
				hardBusOutList.add(nbo);
				}
			else
				{
				for (int i = 0; i < info.length; i++)
					{
					Mixer mixer = AudioSystem.getMixer(info[i]);
					Line.Info lineInfoPlayback = new Line.Info(SourceDataLine.class);
					Line.Info lineInfoCapture = new Line.Info(TargetDataLine.class);
					if (mixer.isLineSupported(lineInfoPlayback))
						{
						log("FOUND PLAYBACK DEVICE :" + info[i].getName());
						hardBusOutList.add(new HardBus(mixer,Bus.Direction.OUT));
						}
					if (mixer.isLineSupported(lineInfoCapture))
						{
						log("FOUND CAPTURE DEVICE :" + info[i].getName());
						hardBusInList.add(new HardBus(mixer,Bus.Direction.IN));
						}
					}
				}
			initialized = true;
			}


		private static void log(String m)
			{
			PMESG(2, "[ AudioDeviceMapper ] " + m);
			}


		private static void error(String m)
			{
			PERROR("[ AudioDeviceMapper ] " + m);
			}


		private static void warn(String m)
			{
			PWARN("[ AudioDeviceMapper ] " + m);
			}


		public static int probeValidModes()
			{
			int r = 0;
			/*
			 * if (hardBusInList.size() != 0) { PMESG(2, "Probing CAPTURE Buses ------------------------------------------------"); for (int j = 0; j < hardBusInList.size(); j++) { HardBus b = hardBusInList.get(j); if (b.probe_valid_modes() < 0) r = -1; } } else { PMESG(2, " No CAPTURE buses found :-("); } if (hardBusOutList.size() != 0) { PMESG(2,
			 * "Probing PLAYBACK Buses ------------------------------------------------"); for (int j = 0; j < hardBusOutList.size(); j++) { HardBus b = hardBusOutList.get(j); if (b.probe_valid_modes() < 0) r = -1; } } else { PMESG(2, " No PLAYBACK buses found :-("); }
			 */
			return r;
			}


		public static void list_buses()
			{
			System.out.println("");
			System.out.println("PLAYBACK BUSES :\n");
			System.out.println("Mustux Bus ID\tDescriptor\tBus Name\n");
			System.out.println("======================================================================\n");
			for (int b = 0; b < getTotalHardOuts(); b++)
				{
				System.out.println("  " + b + " \t\t" + get_bus_out_descriptor(b) + " \t\t" + getHardOutFullName(b));
				}
			System.out.println("");
			System.out.println("CAPTURE BUSES :\n");
			System.out.println("Mustux Bus ID\tDescriptor\tBus Name\n");
			System.out.println("======================================================================\n");
			for (int b = 0; b < getTotalHardIns(); b++)
				{
				System.out.println("  " + b + " \t\t" + get_bus_in_descriptor(b) + " \t\t" + getHardInFullName(b));
				}
			}


		public static void openHardIn(HardBus bus, int pRate, int pBitDepht, int whichChannels)
			{
			openHardIn(bus, pRate, pBitDepht, whichChannels, false, Const.SUGGESTED_FPS);
			}


		public static void openHardIn(HardBus bus, int pRate, int pBitDepht, int whichChannels, boolean block, float suggestedFps)
			{
			bus.open(pRate, pBitDepht, whichChannels, block, suggestedFps);
			}


		public static void openHardOut(HardBus bus, int pRate, int pBitDepht, int whichChannels, boolean block)
			{
			if (!bus.isNullBus)
				bus.open(pRate, pBitDepht, whichChannels, block, suggestedFps);
			}


		public static AudioBusTransferResult transferIn(HardBus bus)  throws HardBusException
			{
			return bus.transfer();
			}


		public static int clean_bus_in(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return -1;
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				{
				return Const.ERROR_INVALID_BUS;
				}
			bus.clean();
			return 1;
			}


		public static int close_bus_in(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				{
				PERROR("Invalid Bus " + busId);
				return -1;
				}
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				{
				return Const.ERROR_INVALID_BUS;
				}
			int r = bus.close();
			return r;
			}


		public static int close_bus_out(int busId)
			{

			if ((busId < 0) || (busId > hardBusOutList.size()))
				{
				PERROR("Invalid Bus " + busId);
				return -1;
				}
			HardBus bus = hardBusOutList.get(busId);
			if (bus == null)
				{

				return Const.ERROR_INVALID_BUS;
				}
			int r = bus.close();

			return r;
			}


		public static boolean is_bus_allocated_for_capture(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return false;
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				{
				return false;
				}
			return bus.isAllocatedForCapture();
			}


		public static boolean is_any_bus_allocated()
			{
			boolean any = false;
			for (int i = 0; i < hardBusOutList.size(); i++)
				{
				if (hardBusOutList.get(i).isAllocatedForPlayback())
					any = true;
				}
			for (int i = 0; i < hardBusInList.size(); i++)
				{
				if (hardBusInList.get(i).isAllocatedForCapture())
					any = true;
				}
			return any;
			}


		public static int close_all_buses()
			{

			close_all_buses_out();
			close_all_buses_in();

			return 1;
			}


		public static int close_all_buses_out()
			{

			for (int i = 0; i < hardBusOutList.size(); i++)
				hardBusOutList.get(i).close();

			return 1;
			}


		public static int close_all_buses_in()
			{

			for (int i = 0; i < hardBusInList.size(); i++)
				hardBusInList.get(i).close();

			return 1;
			}


		public static int getTransferSize()
			{
			return 4096;// TODO
			}


		public static int get_bus_in_transfer_size(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return -1;
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				{
				return Const.ERROR_INVALID_BUS;
				}
			int r = bus.getTransferSize();
			return r;
			}


		public static float[] getTransferBuffer(HardBus bus, int whichChannels)
			{
			return bus.getTransferBuffer(whichChannels);
			}


		public static String get_bus_in_name(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return "INVALID BUS";
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				return "INVALID BUS";
			return bus.get_short_name();
			}


		public static String getHardInFullName(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return "INVALID BUS";
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				return "INVALID BUS";
			return bus.getName();
			}


		public static String get_bus_out_name(int busId)
			{
			if ((busId < 0) || (busId > hardBusOutList.size()))
				return "INVALID BUS";
			HardBus bus = hardBusOutList.get(busId);
			if (bus == null)
				return "INVALID BUS";
			return bus.get_short_name();
			}


		public static String getHardOutFullName(int busId)
			{
			if ((busId < 0) || (busId > hardBusOutList.size()))
				return "INVALID BUS";
			HardBus bus = hardBusOutList.get(busId);
			if (bus == null)
				return "INVALID BUS";
			return bus.getName();
			}


		public static String get_bus_out_descriptor(int busId)
			{
			if ((busId < 0) || (busId > hardBusOutList.size()))
				return "INVALID BUS";
			HardBus bus = hardBusOutList.get(busId);
			if (bus == null)
				return "INVALID BUS";
			return bus.getName();
			}


		public static String get_bus_in_descriptor(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return "INVALID BUS";
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				return "INVALID BUS";
			return bus.getName();
			}


		public static int getTotalCards()
			{
			return cardList.size();
			}


		public static String getCardName(int cardId)
			{
			if ((cardId >= cardList.size()) || (cardId < 0))
				return "INVALID";
			return cardList.get(cardId).getShortName();
			}


		public static String getCardFullName(int cardId)
			{
			if ((cardId >= cardList.size()) || (cardId < 0))
				return "INVALID";
			return cardList.get(cardId).getFullName();
			}


		public static int getTotalHardOuts()
			{
			return hardBusOutList.size();
			}


		public static int getTotalHardIns()
			{
			return hardBusInList.size();
			}


		public static boolean validateHardOut(int busId, int pRate, int pChannels, int pBitDepht)
			{/*
				 * if ((busId < 0) || (busId > hardBusOutList.size())) return false; HardBus bus = hardBusOutList.get(busId); if (bus == null) { return false; } boolean b = bus.valid_mode(pRate, pBitDepht, pChannels); return b;
				 */
			return false;
			}


		public static boolean validateHardIn(int busId, int pRate, int pChannels, int pBitDepht)
			{
			/*
			 * if ((busId < 0) || (busId > hardBusInList.size())) return false; HardBus bus = hardBusInList.get(busId); if (bus == null) { return false; } return bus.valid_mode(pRate, pBitDepht, pChannels);
			 */
			return false;
			}


		public static boolean is_stereo_bus_out(int busId)
			{
			if ((busId < 0) || (busId > hardBusOutList.size()))
				{
				return false;
				}
			HardBus bus = hardBusOutList.get(busId);
			if (bus == null)
				{
				return false;
				}
			return bus.is_stereo();
			}


		public static boolean is_stereo_bus_in(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				{
				return false;
				}
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				{
				return false;
				}
			return bus.is_stereo();
			}


		public static int get_bus_out_index_for_descriptor(String desc)
			{
			for (int i = 0; i < hardBusOutList.size(); i++)
				{
				if (hardBusOutList.get(i).getName().equals(desc))
					{
					return i;
					}
				}
			return -1;
			}


		public static int get_bus_in_index_for_descriptor(String desc)
			{
			for (int i = 0; i < hardBusInList.size(); i++)
				if (hardBusInList.get(i).getName().equals(desc))
					return i;
			return -1;
			}


		public static float get_bus_in_fps(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return -1.0f;
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				return -1.0f;
			return bus.getFps();
			}


		public static float get_bus_out_fps(int busId)
			{
			if ((busId < 0) || (busId > hardBusOutList.size()))
				return -1.0f;
			HardBus bus = hardBusOutList.get(busId);
			if (bus == null)
				return -1.0f;
			return bus.getFps();
			}


		public static int get_format(int bitDepht)// MIG snd_pcm_format_t get_format(int bitDepht)
			{
			switch (bitDepht)
				{
				// MIG case 16 : return SND_PCM_FORMAT_S16_LE;
				// MIG case 20 : return SND_PCM_FORMAT_S20_3LE;
				// MIG case 24 : return SND_PCM_FORMAT_S24_LE;
				// MIG case 32 : return SND_PCM_FORMAT_S32_LE;
				// MIG case 64 : return SND_PCM_FORMAT_FLOAT64_LE;
				}
			return 0;// MIG SND_PCM_FORMAT_UNKNOWN;
			}


		public static boolean get_plughw_use()
			{
			return tryPlugHw;
			}


		public static void set_plughw_use(boolean flag)
			{
			tryPlugHw = flag;
			}


		public static float[] get_bus_out_levels(int busId)
			{
			if ((busId < 0) || (busId > hardBusOutList.size()))
				return null;
			HardBus bus = hardBusOutList.get(busId);
			if (bus == null)
				return null;
			return bus.get_levels();
			}


		public static float[] get_bus_in_levels(int busId)
			{
			if ((busId < 0) || (busId > hardBusInList.size()))
				return null;
			HardBus bus = hardBusInList.get(busId);
			if (bus == null)
				return null;
			return bus.get_levels();
			}


		public static int find_bus_out(int rate, int bitDepth, int channels)
			{
			// TODO
			return -1;
			}


		public static int find_bus_in(int rate, int bitDepth, int channels)
			{
			// TODO
			return -1;
			}


		public static String get_error_mesg(int err)
			{
			switch (err)
				{
				case Const.ERROR_INVALID_CARD:
					return "ERROR_INVALID_CARD";
				case Const.ERROR_INVALID_BUS:
					return "ERROR_INVALID_BUS";
				case Const.ERROR_INVALID_RATE_CHANNELS_OR_BITDEPTH:
					return "ERROR_INVALID_RATE_CHANNELS_OR_BITDEPTH";
				case Const.ERROR_BUSY_BUS:
					return "ERROR_BUSY_BUS";
				case Const.ERROR_CANNOT_INIT_MAPPER:
					return "ERROR_CANNOT_INIT_MAPPER";
				case Const.ERROR_CANNOT_OPEN_PLAYBACK_BUS:
					return "ERROR_CANNOT_OPEN_PLAYBACK_BUS";
				case Const.ERROR_CANNOT_OPEN_CAPTURE_BUS:
					return "ERROR_CANNOT_OPEN_CAPTURE_BUS";
				case Const.ERROR_CANNOT_OPEN_BUS:
					return "ERROR_CANNOT_OPEN_BUS";
				case Const.ERROR_MONO_BUS:
					return "ERROR_MONO_BUS";
				case Const.ERROR_BUS_DOES_NOT_SUPPORT_THIS_BIT_DEPTH:
					return "ERROR_BUS_DOES_NOT_SUPPORT_THIS_BIT_DEPTH";
				case Const.ERROR_BUS_DOES_NOT_SUPPORT_THIS_CHANNELS:
					return "ERROR_BUS_DOES_NOT_SUPPORT_THIS_CHANNELS";
				case Const.ERROR_BUS_DOES_NOT_SUPPORT_THIS_SAMPLE_RATE:
					return "ERROR_BUS_DOES_NOT_SUPPORT_THIS_SAMPLE_RATE";
				case Const.ERROR_8_BITS_BANNED_IN_MUSTUX_APPLICATIONS:
					return "ERROR_8_BITS_BANNED_IN_MUSTUX_APPLICATIONS";
				case Const.ERROR_CANNOT_DEFINE_FORMAT:
					return "ERROR_CANNOT_DEFINE_FORMAT";
				case Const.ERROR_CANNOT_SET_PARAMS_ANY:
					return "ERROR_CANNOT_SET_PARAMS_ANY";
				default:
					return "UNKNOWN ERROR";
				}

			}


		private static void probe_buses(int direction) throws ProbingException, DeviceInitializationException // MIG snd_pcm_stream_t stream)
			{
			throw new ProbingException("Not a GNU/Linux Operating System");
			/*
			 * MIG //MIG snd_ctl_t *handle; int card, err, dev, idx; //MIG snd_ctl_card_info_t *info; //MIG snd_pcm_info_t *pcminfo; //MIG snd_ctl_card_info_alloca(&info); //MIG snd_pcm_info_alloca(&pcminfo); boolean toNextCard=false; String desc; card = -1; audioDeviceList.size() = 0; //MIG if (snd_card_next(&card) < 0 || card < 0) //MIG { //MIG
			 * //PERROR("no soundcards found..."); //MIG return -1; //MIG } while ((card >= 0) && (audioDeviceList.size() < MAX_CARDS)) { toNextCard=false; dev = -1; //MIG char name[32]; //MIG sSystem.out.println(name, "hw:%d", card); audioDeviceList.size()++; if ((err = snd_ctl_open(&handle, name, 0)) < 0) { //PERROR("Control open (%i): %s", card,
			 * snd_strerror(err)); toNextCard=true; } if ((err = snd_ctl_card_info(handle, info)) < 0) { //PERROR("Control hardware info (%i): %s", card, snd_strerror(err)); snd_ctl_close(handle); toNextCard=true; } while (!toNextCard) { unsigned int count; if (snd_ctl_pcm_next_device(handle, &dev)<0) //PERROR("No buses available on soundcard
			 * :-("); if (dev < 0) break; snd_pcm_info_set_device(pcminfo, dev); snd_pcm_info_set_subdevice(pcminfo, 0); snd_pcm_info_set_stream(pcminfo, stream); if ((err = snd_ctl_pcm_info(handle, pcminfo)) < 0) { if (err != -ENOENT) //PERROR("Control digital audio info (%i): %s", card, snd_strerror(err)); continue; } cardShortName[card] =
			 * snd_ctl_card_info_get_id(info); audioDeviceList[card] = snd_ctl_card_info_get_name(info); count = snd_pcm_info_get_subdevices_avail(pcminfo); if (count > 128) { //PERROR("Detected more then 32 subdevices, impossible! Aborting routine"); return -1; } for (idx = 0; idx < (int)count; idx++) { snd_pcm_info_set_subdevice(pcminfo, idx); if
			 * ((err = snd_ctl_pcm_info(handle, pcminfo)) < 0) { //PERROR("control digital audio playback info (%i): %s", card, snd_strerror(err)); } else { if (stream == SND_PCM_STREAM_CAPTURE) { if (busIn.size() < .size()) { desc = "hw:"+QString::number(card); AudioBus b=new Bus(card, idx, CAPTURE, desc); busIn[busIn.size()]=b; busIn.size()++; } }
			 * else { if (busOut.size() < .size()) { desc = "hw:"+QString::number(card); AudioBus b=new Bus(card, idx, PLAYBACK, desc); busOut[busOut.size()]=b; busOut.size()++; } } } } } snd_ctl_close(handle); if (snd_card_next(&card) < 0) { //PERROR("No valid soundcards :-("); break; } }
			 */
			}


		public static ArrayList<HardBus> getHardIns()
			{
			return hardBusInList;
			}


		public static ArrayList<HardBus> getHardOuts()
			{
			return hardBusOutList;
			}


		/** Transfer all buses out which are allocated for playback 
		 * All transfers are non-blocked, the blocking is caused by sleeping the thread for some time
		 * then we get the proper framing.
		 * */
		public static void transferOut()  throws HardBusException, InterruptedException
			{
			boolean firstBus = true;
			for (HardBus bus : hardBusOutList)
				{
				if (bus.isAllocatedForPlayback())
					{
					bus.transfer();
					bus.clean();
					}
				if (firstBus)
					{
					Thread.sleep(Const.FRAME_SLEEP_TIME);
					firstBus = false;
					}
				}
			}


		public static HardBus getHardBusByName(String s)
			{
			for (HardBus b : hardBusInList)
				{
				if (b.getName().equals(s))
				return b;
				}
			for (HardBus b : hardBusOutList)
				{
				if (b.getName().equals(s))
				return b;
				}
			return null;
			}


		public static ArrayList<HardBus> getAllOut()
			{
			return hardBusInList;
			}
		public static ArrayList<HardBus> getAllIn()
			{
			return hardBusOutList;
			}
	}
