package org.protux.core.engines.audio;

import static org.protux.core.Debugger.PERROR;
import static org.protux.core.Debugger.PMESG;
import static org.protux.core.Debugger.PWARN;

import java.io.IOException;
import java.util.ArrayList;

import org.protux.core.BaseThread;
import org.protux.core.Const;
import org.protux.core.editor.AudioClip;
import org.protux.core.editor.Editor;
import org.protux.core.editor.Region;
import org.protux.core.editor.Song;
import org.protux.core.editor.Track;
import org.protux.core.editor.Rack;
import org.protux.core.exceptions.HardBusException;
import org.protux.core.media.audio.AudioMedia;

/**
 * Mixer is the SOFT MAPPER , while AudioDeviceMapper is the Hard Mapper<br>
 * SIGNAL ROUTING (PUT ME ON DOCS...) <br>
 * AUDIO (FROM CLIPS) TRACK GAIN CLIP GAIN CLIP FADES GAIN TRACK PAN FILTER CHAIN SOFTBUS OUT HARDBUS OUT
 */

public class SoftBusMapper
	{

		private static final boolean maxPerformance = false;

		class PlayThread extends BaseThread
			{

				@Override
				public void run()
					{
					PMESG(2, "-- Play Thread starting ...");
					if (maxPerformance)
						setMaxPriority();
					if ((goStartPos >= 0) || (goEndPos >= 0))
						{
						playNormal();
						}
					else
						{
						playRegions();
						}
					PMESG(2, "-- Play Thread stopped.");
					}


				private void playNormal()
					{
					try
						{
						while (true)
							{
							goNext();
							}
						}
					catch (InterruptedException e)
						{
						}
					catch (HardBusException ex)
						{
						ex.printStackTrace();
						}
					catch (IOException e)
						{
						PWARN("Thread will stop due I/O problems.");
						}
					}


				private void playRegions()
					{
					for (Region m : getRegionList())
						{
						PMESG(2, "Going thru region [" + m.beginBlock + "," + m.endBlock + "]");
						goStartPos = m.beginBlock;
						goEndPos = m.endBlock;
						goingPos = m.beginBlock;
						while (true)
							{
							System.out.println("Going (regions) ....");
							// / TODO
							}
						}
					}
			}

		class ListenThread extends BaseThread
			{

				@Override
				public void run()
					{
					setMaxPriority();
					while (true)
						{
						System.out.print("Listenning ...");
						// TODO
						}
					}

			}

		private PlayThread goThread = null;
		private ListenThread listenerThread = null;
		Song assocSong;
		public static final int DYNAMIC_SLICE_FACTOR = 10; // improve-me get from Global Properties
		public static final int SAMPLE_SIZE = 2;
		public static final int RENDER_BUFFER_SIZE = 16384; // in bytes
		public static final int MIN_SAMPLE = -32768;
		public static final int MAX_SAMPLE = 32767;
		public static final int NOT_DEFINED = -1;
		private static final float DEFAULT_SUGGESTED_FPS = 20.0f; // WILL BE USER CONFIGURABLE SOON

		public static final int ERROR_EMPTY_MTA = -10;
		public static final int REACHED_END_OF_MTA = -111;

		public static final int STOPPED = 0;
		public static final int ARMED = 1;
		public static final int GOING = 2;
		public static final int PAUSED = 3;
		public static final int ABORTED = 4;

		public long goStartPos;
		public long goEndPos;
		public long goingPos;
		public long listenPos;
		public int goingStep;
		public int listenStep;
		public float commonFpsOut, commonFpsIn;
		public int status = STOPPED;
		private ArrayList<Region> regionList;
		private Editor proxyEditor;
		private Rack proxyRack;


		/**
		 * A Mixer is a Song's aggregated engine object (1 to 1) which can playback several audio chunks from several Tracks, at the same time it can recorded on pre-specified tracks (the armed) tracks.
		 */
		public SoftBusMapper(Song pSong)
			{
			this.assocSong = pSong;
			this.proxyEditor = assocSong.proxyEditor;
			this.proxyRack = assocSong.proxyRack;
			this.status = STOPPED;
			this.goStartPos = 0;
			this.goEndPos = -1;
			}


		public boolean startGo(ArrayList<Region> regionList)
			{
			this.regionList = regionList;
			return startGo(-1, -1); // this forces then region list mode
			}


		public boolean startGo(long pStartPos)
			{
			return startGo(pStartPos, 0);// REVIEW ME
			}


		public boolean startGo(long pStartPos, long pEndPos)
			{
			if (goThread != null)
				{
				PWARN("PLAYBACK ALREADY GOING !");
				return true;
				}
			goThread = new PlayThread();
			boolean emptyEditor = proxyEditor.empty();
			boolean armed = assocSong.isAnyTrackArmed();
			if (!armed && emptyEditor)
				{
				PMESG(3, "No Content to play and no track armed to record.. thus.. nothing to do !");
				return false;
				}
			if (emptyEditor)
				{
				PWARN("Editor Seems to be empty. Performing only recording (No Playback Bus will be open)");
				emptyEditor = true;
				}
			goStartPos = pStartPos;
			goEndPos = pEndPos;
			goingPos = goStartPos;
			listenPos = goStartPos;
			int nt = assocSong.getNumTracks();
			for (int t = 0; t < nt; t++)
				{
				Track track = assocSong.getTrack(t);
				SoftBus softbusOut = track.getSoftOut();
				HardBus busOut = softbusOut.getHardBus();
				if (!busOut.isNullBus())
					HardBusMapper.openHardOut(busOut, assocSong.rate, assocSong.bitDepth, track.get_which_channels(), true);
				}
			goingStep = HardBusMapper.getTransferSize() / (2 * (assocSong.bitDepth / 8)); // goingStep is in blocks // WHAT IS THIS "2" ?
			status = GOING;
			if (!emptyEditor)
				{
				goThread.start();
				}
			return true;
			}


		public synchronized void forceStop()
			{
			// If status == ARMED or STOPPED the goingThread will finish exec.
			// When status == ARMED the listenerThread will continue exec.
			if (goThread != null)
				{
				try
					{
					goThread.interrupt();
					}
				catch (Exception ex)
					{
					ex.printStackTrace();
					}
				// FOR LATER listenerThread.wait();
				if (assocSong.isAnyTrackArmed())
					status = ARMED;
				else
					status = STOPPED;
				}
			goThread = null;
			afterStop();
			}


		public void afterStop()
			{
			int r = HardBusMapper.close_all_buses_out();
			if (r < 0)
				{
				PWARN("Cannot close one or more playback buses. Continuing anyway ...");
				}
			int nt = assocSong.getNumTracks();
			for (int t = 0; t < nt; t++)
				{
				Track track = assocSong.getTrack(t);
				if (track.cleanup_buffers() < 0)
					{
					PWARN("Track " + (t + 1) + " cleanup_buffers returned an error!");
					// FORLATER if (status==ABORTED) track.discard_recorded_clip();
					}
				}
			}


		public long getGoingPos()
			{
			return goingPos;
			}


		public synchronized boolean isGoing()
			{
			return goThread != null;
			}


		public synchronized void goNext() throws HardBusException, InterruptedException, IOException
			{
			PMESG(2, "Going on goingPos=" + goingPos);
			boolean thereAreSolos = proxyRack.anySolo();
			int busBufferSize = HardBusMapper.getTransferSize();
			for (Track track : assocSong.getTracks())
				{
				if ((thereAreSolos && (!track.isSolo())) || (track.isMuted()))
					continue;
				HardBus busOut = track.getSoftOut().getHardBus();
				int busBufferBlockSize = 4; // REVIEW ME
				float[] mixBuffer = HardBusMapper.getTransferBuffer(busOut, track.get_which_channels());
				float[] sourceBuffer = new float[busBufferSize];
				mixTrackPortion(track, sourceBuffer, mixBuffer, busBufferBlockSize);
				}
			HardBusMapper.transferOut();
			goingPos += goingStep;
			}


		public boolean listening() // should be called "listen"
			{
			// Capture process ( in all armed tracks)
			if (status == STOPPED)
				{
				System.out.println("Listener Thread : status seems to be STOPPED. Finishing...");
				return false;
				}
			int nt = assocSong.getNumTracks();
			for (int t = 0; t < nt; t++)

				{
				Track track = assocSong.getTrack(t);
				if (!track.armed())
					continue;
					// MIG HardBus busIn = track.getSoftIn().getHardBus();
					// MIG if (AudioDeviceMapper.is_bus_allocated_for_capture(busIn))
					{
					AudioBusTransferResult tr = null;// //MIG AudioDeviceMapper.bus_in_transfer(busIn);
					if (tr.error == 0)
						{
						short[] captureBuffer = null;// //MIG AudioDeviceMapper.get_bus_in_transfer_buffer(busIn, track.get_which_channels());
						int captureBufferSize = 0;// MIG AudioDeviceMapper.get_bus_in_transfer_size(busIn);
						if (status == GOING)
							{
							// MIG track.feed_recording_clip(captureBuffer, captureBufferSize);
							listenPos += captureBufferSize / (2 /* FIXME */* (assocSong.bitDepth / 8)); // listenStep is in blocks
							}
						}
					else
						{
						PERROR("Cannot read from bus ");
						}
					}
				}
			return true;
			}


		public int abort()
			{
			status = ABORTED;
			try
				{
				goThread.wait();
				}
			catch (InterruptedException ex)
				{
				System.out.println("Error : " + ex);
				}

			return 1;
			}


		public int getStatus()
			{
			return status;
			}


		private final int mixTrackPortion(Track track, float[] sourceBuffer, float[] mixBuffer, int mixBufferBlockSize) throws IOException
			{
			long p;
			boolean stereoSource = false;
			int overcounter = 0;
			final float trackGain = track.getGain();
			final float trackGainFactor = (trackGain >= 0.0f) ? 1.0f + trackGain / 6.0f : 1.0f / (float) Math.pow(2.0f, -1.0f * trackGain / 6.0f);
			final float panFactor = track.getPan();
			final float panL = panFactor >= 0 ? 1.0f - panFactor : 1.0f;
			final float panR = panFactor >= 0 ? 1.0f : 1.0f + panFactor;
			final long lastPos = goingPos + goingStep;
			AudioClip clip = track.getClipUnderBlock(goingPos);
			if (clip == null)
				clip = track.getClipAfter(goingPos);
			while ((clip != null) && (!clip.isMuted) && (clip.trackFirstBlock < lastPos))
				{
				AudioMedia audioSource = clip.audioSource;
				p = ((clip.trackFirstBlock < goingPos) ? goingPos : clip.trackFirstBlock);
				System.out.println("Mixing clip "+clip.name + " at "+p);
				float clipGainFactor = clip.gain;
				long sourceFirstBlockToMix = clip.sourceFirstBlock + (goingPos - clip.trackFirstBlock);
				int mixBufferRelativeMixStartBlock = (int) (p - goingPos);

				int numBlocksToReadFromSource = 0;
				if ((clip.trackFirstBlock < goingPos) && (clip.trackLastBlock > goingPos + goingStep))
					numBlocksToReadFromSource = goingStep;
				else if ((clip.trackFirstBlock < goingPos) && (clip.trackLastBlock <= goingPos + goingStep))
					numBlocksToReadFromSource = (int) (clip.trackLastBlock - goingPos);
				else if ((clip.trackFirstBlock >= goingPos) && (clip.trackLastBlock > goingPos + goingStep))
					numBlocksToReadFromSource = (int) (goingPos + goingStep - clip.trackFirstBlock);
				else
					numBlocksToReadFromSource = clip.getLenght();
				int numSamplesToReadFromSource = numBlocksToReadFromSource;
				if (audioSource.channels == Const.STEREO)
					{
					stereoSource = true;
					numSamplesToReadFromSource *= 2;
					}
				audioSource.prepareForReadingSamples(numSamplesToReadFromSource);

				// read the audio from the file
				audioSource.blockSeek(sourceFirstBlockToMix);
				audioSource.readSamples(sourceBuffer);

				// Mix
				int pSourceBuffer = 0;
				int pMixerBuffer = mixBufferRelativeMixStartBlock * mixBufferBlockSize; // destination point in the mixbuff, where the data will be mixed onto.
				int srcCounter = 0;
				int channel = 0;
				while (srcCounter < numBlocksToReadFromSource) // && (playable))
					{
					// APPLY GAINS, ACCORDDING TO THE SIGNAL ROUTING
					float fadeAndGrossFadeGainFactor = clip.getFadeFactorFor(goingPos + srcCounter);
					float newSample = sourceBuffer[pSourceBuffer] * trackGainFactor * clipGainFactor * fadeAndGrossFadeGainFactor;

					// APPLY TRACK PAN (only if it is a stereo clip)
					newSample *= (channel == 1) ? panR : panL;
					if (newSample < MIN_SAMPLE || newSample > MAX_SAMPLE)
						overcounter++;

					// Mix sample onto the buffer
					mixNewSampleIntoMixBuffer(mixBuffer, pMixerBuffer, newSample);

					// make a copy of the sample in case of a mono-source
					if (!stereoSource)
						{
						pMixerBuffer += SAMPLE_SIZE;
						mixNewSampleIntoMixBuffer(mixBuffer, pMixerBuffer, newSample);
						}
					pSourceBuffer += SAMPLE_SIZE;
					pMixerBuffer += SAMPLE_SIZE;
					srcCounter++;
					if (++channel > 1)
						channel = 0;
					}
				// FOR LATER int lastSubClipBlock = relativeSliceBlock + lenght; if (lastSubClipBlock>maxPlayableBlockInSlice) maxPlayableBlockInSlice=lastSubClipBlock;
				clip = clip.next;
				}

			// FORLATER Process Filter Chain
			// FORLATER track.controllerChain.process(mixBuffer, mixBufferSize);

			// Check overloads
			if (overcounter > 0) // Saturation (clipping)
				{
				// do something ...PMESG(3, "Detected " + sat + " over(s), please lower the output volume...");
				}
			return overcounter; // improve-me .. must return the amount overloaded.
			}


		private boolean valid_mode(int rate, int bitDepth)
			{
			boolean validMode = true;
			// IMPROVE-ME
			if ((rate < 22050) || (bitDepth != 16)) // FOR FUTURE : .. && pBitDepth!=20,24,32 ... and so on
				validMode = false;
			return validMode;
			}


		private static final void mixNewSampleIntoMixBuffer(float[] mixBuffer, int mixBufferPos, float addingSample)
			{
			float newSample = mixBuffer[mixBufferPos] + addingSample;
			newSample = newSample < MIN_SAMPLE ? MIN_SAMPLE : newSample;
			newSample = newSample > MAX_SAMPLE ? MAX_SAMPLE : newSample;
			mixBuffer[mixBufferPos] = newSample;
			}


		public ArrayList<Region> getRegionList()
			{
			return null;
			}


		public void arm(SoftBus busIn, int whichChannels)
			{
			if (listenerThread == null)
				{
				HardBusMapper.openHardIn(busIn.getHardBus(), assocSong.rate, assocSong.bitDepth, whichChannels);
				status = ARMED;
				listenerThread = new ListenThread();
				listenerThread.start();
				PMESG(3, "CAPTURE  THREAD START : SUCESS");
				}
			}


		int disarm(int busIn)
			{
			int r = 0;
			if (!assocSong.isAnyTrackArmed())
				{
				status = STOPPED;
				try
					{
					listenerThread.wait();// "... it will finish soon, cause status now is STOPPED"
					}
				catch (InterruptedException ex)
					{
					PERROR("" + ex);
					}
				}
			boolean noMoreUsed = true;
			int nt = assocSong.getNumTracks();
			for (int t = 0; t < nt; t++)
				{
				Track track = assocSong.getTrack(t);
				if (false) // MIG ((track.armed()) && (track.getSoftIn().getHardBus() == busIn))
					noMoreUsed = false;
				}
			if (noMoreUsed)
				{
				r = HardBusMapper.close_bus_in(busIn);
				if (r == Const.ERROR_INVALID_BUS)
					{
					PERROR("Invalid Bus");
					}
				else if (r == Const.ERROR_BUSY_BUS)
					{
					PERROR("Busy bus");
					}
				}

			return r;
			}

	}
