package org.protux.core.gui;

import static org.protux.core.Debugger.PMESG;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;

import org.protux.core.Const;
import org.protux.core.Project;
import org.protux.core.Protux;

public class InfoBox extends FastPanel
	{

		private static final Color BACKGROUND_COLOR = new Color(160, 170, 150);
		private static final Font DEFAULT_FONT = new Font("Helvetica", 0, 10);
		private static final Font FONT_SONG_NAVIGATOR = new Font("Arial", Font.BOLD, 12);
		private static final int ZOOM_INFO_X = 5;
		private static final int ZOOM_INFO_Y = 80;
		private static final int ZOOM_INFO_W = 150;
		private static final int ZOOM_INFO_H = 20;
		private static final Dimension MIN_DIM = new Dimension(200, 10);


		public InfoBox()
			{
			super("InfoBox",MIN_DIM, BACKGROUND_COLOR);
			}


		@Override
		public void recreate(Graphics2D painter)
			{
			Project currentProject = Protux.getCurrentProject();
			log("Recreating...");
			if (currentProject != null)
				{
				clear(painter);
				drawInfo(painter, currentProject);
				drawZoomInfo(painter);
				}
			else
				drawDumbX(painter);
			}


		void drawInfo(Graphics2D painter, Project project)
			{
			painter.setFont(FONT_SONG_NAVIGATOR);
			final int BW = 20;
			final int BH = 16;
			final int ns = project.getNumSongs();
			for (int i = 0; i < ns; i++)//
				{
				final int bx = i * BW + 1;
				final int by = 2;
				painter.setColor(i % 2 == 0 ? Color.GRAY : Color.DARK_GRAY);
				painter.fill3DRect(bx, by, BW, BH, true);
				painter.setColor((project.getCurrentSongId() == i) ? Color.GREEN : Color.BLACK);
				painter.drawString("" + (i + 1), bx + 5, by + 13);
				}
			String sinfo = "" + (project.getCurrentSongId() + 1) + " - " + project.get_current_song().title;
			painter.setColor(Color.black);
			painter.setFont(DEFAULT_FONT);
			painter.drawString("Project : " + project.getTitle(), 5, 35);
			painter.drawString("Song    : " + sinfo, 5, 50);
			painter.drawString("          " + project.get_current_song().rate + " Bit Depth: " + project.get_current_song().bitDepth, 5, 65);
			}


		void clearZoomInfo(Graphics2D painter)
			{
			clear(painter, ZOOM_INFO_X, ZOOM_INFO_Y, ZOOM_INFO_W, ZOOM_INFO_H + 16);
			}


		void drawZoomInfo(Graphics2D painter)
			{
			int hzoom = Protux.getCurrentProject().get_current_song().hzoom;
			painter.setColor(Color.BLACK);
			painter.setFont(DEFAULT_FONT);
			String sHzoom = "HZ 1:" + Const.ZOOM_STEP[hzoom];
			if (hzoom <= Const.MAX_ZOOM_USING_SOURCEFILE)
				sHzoom = sHzoom + " Micro View";
			painter.drawString(sHzoom, ZOOM_INFO_X, ZOOM_INFO_Y + 15);
			}


		private void log(String string)
			{
			PMESG(2, "[ InfoBox ] " + string);
			}


		public void mouseClicked(MouseEvent e)
			{
			}


		public void mousePressed(MouseEvent e)
			{
			}


		public void mouseReleased(MouseEvent e)
			{
			}


		public void mouseEntered(MouseEvent e)
			{
			}


		public void mouseExited(MouseEvent e)
			{
			}
	}
