package org.protux.core.media.audio;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;

import org.protux.core.Byter;
import org.protux.core.Const;
import org.protux.core.Debugger;
import org.protux.core.exceptions.PeakException;

import static org.protux.core.Debugger.*;


public class Praf extends AudioMedia
	{
	private static final int FRAG_SIZE = 4096;
	private static final int COPY_BUFFER_SIZE = 65536;
	protected static final int POS_AUDIO_TYPE = 0;
	protected static final int POS_NUMBER_CHANNELS = 4;
	protected static final int POS_SAMPLE_RATE = 8;
	protected static final int POS_BIT_DEPTH = 12;
	protected static final int POS_AUDIO_SIZE = 16;
	protected static final int POS_PROJECT_LABEL = 24;
	protected static final int POS_GROUP_LABEL = 40;
	protected static final int POS_HARDWARE_LABEL = 56;
	protected static final int POS_DESCRIPTION = 72;


	public Praf (File aFile) throws IOException
		{
		super(aFile, false, Const.PRAF_HEADER_SIZE);
		this.bestFragSize = FRAG_SIZE;
		}


	public Praf (File aFile, boolean peakLess) throws IOException
		{
		super(aFile, peakLess, Const.PRAF_HEADER_SIZE);
		}


	@Override
	public void finalize()
		{
		closeFile();
		}


	@Override
	public void readHeader() throws IOException
		{
		super.readHeader();
		String sAudioType = Byter.readString(header, POS_AUDIO_TYPE, 4);
		if (!sAudioType.equalsIgnoreCase("PRAF"))
			{
			PERROR("File does not seem to be a PRAF file");
			throw new IOException();
			}
		this.channels = Byter.readInt(header, POS_NUMBER_CHANNELS);
		this.bitDepth = Byter.readInt(header, POS_BIT_DEPTH);
		this.rate = Byter.readInt(header, POS_SAMPLE_RATE);
		this.projectLabel = Byter.readString(header, POS_PROJECT_LABEL, 16);
		this.groupLabel = Byter.readString(header, POS_GROUP_LABEL, 16);
		this.hardwareLabel = Byter.readString(header, POS_HARDWARE_LABEL, 16);
		this.description = Byter.readString(header, POS_DESCRIPTION, 16);

		configure();
		PMESG(2, "Header of " + phisicalFile + " successfully read : " + getInfo());
		}


	@Override
	public void writeHeader() throws IOException
		{
		rewindAll();
		Byter.writeInto_U15Buffer(header, "PRAF", POS_AUDIO_TYPE);
		Byter.writeInto_U15Buffer(header, channels, POS_NUMBER_CHANNELS);
		Byter.writeInto_U15Buffer(header, bitDepth, POS_BIT_DEPTH);
		Byter.writeInto_U15Buffer(header, rate, POS_SAMPLE_RATE);
		Byter.writeInto_U15Buffer(header, projectLabel, POS_PROJECT_LABEL);
		Byter.writeInto_U15Buffer(header, groupLabel, POS_GROUP_LABEL);
		Byter.writeInto_U15Buffer(header, hardwareLabel, POS_HARDWARE_LABEL);
		Byter.writeInto_U15Buffer(header, description, POS_DESCRIPTION);
		dumpHeader();
		super.writeHeader();
		}


	@Override
	public void appendSamplesFromFile(AudioMedia f) throws IOException
		{
		if (peak != null)
			{
			peak.sourceMediaChanged();
			}
		byte buf[] = new byte[COPY_BUFFER_SIZE];
		RandomAccessFile src = f.getFile();
		this.goBotton();
		f.rewindAudio();
		int t = 0;
		int x = 0;
		int tot = 0;
		System.out.print("Copying : ");
		while (t >= 0)
			{
			t = src.read(buf);
			if (t > 0)
				{
				tot += t;
				x++;
				raFile.write(buf, 0, t);
				if (x > 20)
					{
					System.out.print(".");
					x = 0;
					}
				}
			}
		System.out.println(".");
		audioSize = raFile.length() - headerSize;
		totalBlocks = audioSize / blockSize;
		PMESG(1, "Appended " + tot + " bytes. New audioSize=" + audioSize);
		if (peak != null)
			{
			try
				{
				peak.createRAMBuffers();
				}
			catch (PeakException e)
				{
				// TODO Auto-generated catch block
				e.printStackTrace();
				}
			}

		}


	@Override
	public void appendSamplesFromBuffer(byte[] buffer, int bufferSize) throws IOException
		{
		//MIG size_t size ;
		//MIG if (fseek(file, 0, SEEK_END)<0)
		//MIG 	{
		//PERROR2("Cannot to go file botton to append samples");
		//MIG 				return -1;
		//MIG 			}
		//MIG size = fwrite(buffer, 1 , bufferSize, file);
		//MIG if ((int) size < bufferSize)
		//MIG 	{
		//PERROR2("Could not append the whole buffer to file");
		//MIG 	return -1;
		//MIG 	}
		}


	@Override
	public void fixAudioSize() throws IOException
		{
		PMESG(3, "Fixing PRAF file size ...");
		long fileSize = raFile.length();
		audioSize = fileSize - headerSize;
		totalBlocks = audioSize / blockSize;
		Byter.writeInto_U15Buffer(header, audioSize, POS_AUDIO_SIZE);
		super.writeHeader();
		}


	@Override
	public AudioMedia copyTo(File newFile)
		{
		Praf newPraf = null;
		try
			{
			newPraf = new Praf(newFile);
			newPraf.set(channels, rate, bitDepth, projectLabel, groupLabel, hardwareLabel, description);
			newPraf.writeHeader();
			byte buf[] = new byte[FRAG_SIZE];
			rewindAudio();
			int size;
			while (!eof())
				{
				size = readFragment(buf, FRAG_SIZE);
				if (newPraf.writeFragment(buf, size) < 0)
					{
					//error
					}
				}
			}
		catch (IOException e)
			{
			// TODO Auto-generated catch block
			e.printStackTrace();
			}
		return newPraf;
		}

	}
