package org.protux.core.media.audio;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Iterator;

import org.protux.core.Byter;
import org.protux.core.Const;

import static org.protux.core.Debugger.*;

public class Wav extends AudioMedia
	{
	private static final int FRAG_SIZE = 4096;
	final int POS_AUDIO_TYPE = 0;
	final int POS_CHUNK1_SIZE = 4;
	final int POS_RIFF_TYPE = 8;
	final int POS_CHUNK2_ID = 12;
	final int POS_CHUNK2_SIZE = 16;
	final int POS_COMPRESSION_CODE = 20;
	final int POS_NUMBER_CHANNELS = 22;
	final int POS_SAMPLE_RATE = 24;
	final int POS_BYTES_PER_SEC = 28;
	final int POS_BLOCK_SIZE = 32;
	final int POS_BIT_DEPTH = 34;

	public Wav(File aFile) throws IOException
		{
		this(aFile, false);
		this.bestFragSize = FRAG_SIZE;
		}

	public Wav(File aFile, int mode) throws IOException
		{
		this(aFile, false);
		}


	public Wav(File aFile, boolean peakLess) throws IOException
		{
		this (aFile, peakLess, Const.READ_AND_WRITE);
		}

	public Wav(File aFile, boolean peakLess, int mode) throws IOException
		{
		super(aFile, peakLess, Const.WAV_COMMON_HEADER_SIZE, mode);
		}


	@Override
	public void readHeader() throws IOException
		{
		super.readHeader();
		String sAudioType = Byter.readString(header, POS_AUDIO_TYPE, 4);
		String sRiffType = Byter.readString(header, POS_RIFF_TYPE, 4);
		if ((!sAudioType.equalsIgnoreCase("RIFF")) || (!sRiffType.equalsIgnoreCase("WAVE")))
			{
			PERROR("File does not seem to be a WAV/RIFF file");
			throw new IOException();
			}
		this.channels = header[POS_NUMBER_CHANNELS] + header[POS_NUMBER_CHANNELS + 1] * 256;
		int r1 = header[POS_SAMPLE_RATE];
		int r2 = header[POS_SAMPLE_RATE + 1];
		int r3 = header[POS_SAMPLE_RATE + 2];
		int r4 = header[POS_SAMPLE_RATE + 3];
		this.rate = r1 + r2 * 256 + r3 * 16384 + r4 * 65536;
/* FOR LATER
  		this.channels = Byter.readUShort(header,POS_NUMBER_CHANNELS);
		this.rate = Byter.readInt(header,POS_SAMPLE_RATE);
		this.bitDepth = Byter.readUShort(header, POS_BIT_DEPTH);
 */		
		int b1 = header[POS_BIT_DEPTH];
		int b2 = header[POS_BIT_DEPTH + 1];
		this.bitDepth = b1 + b2 * 256;
		log( "Audio is "+bitDepth+" bits "+ rate+" hz"+ ((channels==Const.MONO)?"MONO":"STEREO"));
		configure();
		dumpHeader();
		}



	@Override
	public void writeHeader() throws IOException
		{
		/*MIG
		 char audioType[4];
		 char riffType[4];
		 char useLess[8];
		 short compressionCode;
		 channels=pChannels;
		 rate = pRate;
		 bitDepth = pBitDepth;
		 audioSize=pAudioSize;
		 blockSize = channels * (bitDepth/8);
		 bytesPerSec = rate * channels * (bitDepth/8);
		 sampleSize = blockSize / channels;
		 totalBlocks = audioSize / blockSize;
		 rewind_all();		//WAS: rewind_audio(); Ehm, we shouldn't do that, should we :-) By Remon
		 audioType[0] = 'R';
		 audioType[1] = 'I';
		 audioType[2] = 'F';
		 audioType[3] = 'F';
		 fwrite(&(audioType), 4, 1 , file);
		 long wFileSize = audioSize + headerSize;
		 fwrite(&(wFileSize), sizeof(long), 1 , file); //WARNING: sizeof long can vary from system to system ! But WAV specs
		 //ask explicitly for a size of 4 bytes !! by Martin
		 riffType[0] = 'W';
		 riffType[1] = 'A';
		 riffType[2] = 'V';
		 riffType[3] = 'E';
		 fwrite(&(riffType), 4, 1 , file);
		 useLess[0] = 'f';
		 useLess[1] = 'm';
		 useLess[2] = 't';
		 useLess[3] = ' ';
		 useLess[4] = 16; //size of WAVE section sub-chunk, usually 16 (for linear quantizised PCM)
		 useLess[5] = 0;
		 useLess[6] = 0;
		 useLess[7] = 0;
		 fwrite(&(useLess),  8, 1 , file);
		 compressionCode=1;
		 fwrite(&(compressionCode), sizeof(short), 1, file); //WARNING: again specs ask for 2 bytes
		 short wChannels = (short) channels;
		 fwrite(&(wChannels), sizeof(short), 1 , file); //WARNING: again specs ask for 2 bytes
		 long wRate = (long) rate;
		 fwrite(&(wRate), sizeof(long), 1 , file); //WARNING: again specs ask for 4 bytes
		 long wBytesPerSec = (long) bytesPerSec;
		 fwrite(&(wBytesPerSec), sizeof(long), 1 ,file); //WARNING: again specs ask for 4 bytes
		 short wBlockSize = (short) blockSize;
		 fwrite(&(wBlockSize), sizeof(short), 1 , file); //WARNING: again specs ask for 2 bytes
		 short wBitDepth = (short) bitDepth;
		 fwrite(&(wBitDepth), sizeof(short), 1 , file); //WARNING: again specs ask for 2 bytes
		 useLess[0] = 'd';
		 useLess[1] = 'a';
		 useLess[2] = 't';
		 useLess[3] = 'a';
		 long dataSize = wFileSize - 36;
		 fwrite(&(useLess),  4, 1 , file);
		 fwrite(&(dataSize), 4, 1 , file);
		 */
		}



	@Override
	public void appendSamplesFromFile(AudioMedia f) throws IOException
		{
		// TODO
		}



	@Override
	public void appendSamplesFromBuffer(byte[] buffer, int bufferSize) throws IOException
		{
		//MIG size_t size ;
		//MIG if (fseek(file, 0, SEEK_END)<0)
		//MIG 	{
		//MIG PERROR2("Cannot to go file botton to append samples");
		//MIG return -1;
		//MIG }
		//MIG size = fwrite(buffer, 1 , bufferSize, file);
		//MIG if ((int) size < bufferSize)
		//MIG 	{
		//MIG PERROR2("Could not append the whole buffer to file");
		//MIG return -1;
		//MIG }
		}



	@Override
	public void fixAudioSize() throws IOException
		{
		PMESG(3, "Fixing WAV file size ...");
		long fileSize = raFile.length();
		audioSize = fileSize - headerSize;
		totalBlocks = audioSize / blockSize;
		// FOR LATER if (fileSize > 2Giga) PERROR (Wav supports file only up to 2 gbytes);
		long wFileSize = audioSize + headerSize - 8;
		rewindAll();
		//MIG fseek(file, 4, SEEK_SET);
		//MIG fwrite(&(wFileSize), sizeof(long), 1 , file);
		//MIG long dataSize = wFileSize - 36;
		//MIG fseek(file, 40, SEEK_SET);
		//MIG fwrite(&(dataSize), sizeof(long), 1 , file);
		PMESG(3, "Fixed audioSize to " + wFileSize);
		}



	@Override
	public AudioMedia copyTo(File newFile) throws IOException
		{
		return null;
		}


	/** Unit test */
	public static void main(String[] args)
		{
		try
			{
			Wav wav = new Wav(new File("test.wav"));
			wav.readHeader();
			}
		catch (Exception e)
			{
			PERROR("Error : " + e);
			}
		}



	}
