package org.protux.input.jmb;

import java.awt.Frame;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.StringReader;
import java.util.Timer;
import java.util.TimerTask;

import org.protux.core.Debugger;
import org.protux.core.GlobalProperties;
import org.protux.core.OneShotTimer;
import org.protux.core.Util;
import org.protux.core.editor.Editor;
import org.protux.core.exceptions.JmbException;
import org.protux.core.gui.InfoViewer;
import static org.protux.core.Debugger.*;

public class Jmb implements KeyListener
	{

	protected ReleaseTimer releaseTimer;
	protected SecondChanceTimer secondFactTimer;
	//private ReleaseTimer clearOutputTimer;
	public static final int MAX_JMB_ACTIONS = 4096;
	public static final int MAX_TIME_DIFFERENCE_FOR_DOUBLE_KEY_CONSIDERATION = 50;
	public static final int CLEAR_TIME = 2000;
	public static final int ASSUME_HOLD_TIME = 200; // it will wait a release for 100 ms. Otherwise it will assume a hold
	public static final int DOUBLE_FACT_WAIT_TIME = 200;
	public static final int S_FKEY_FKEY = 8; // >K>K   - press one key and then other key fast and sequentially
	public static final int S_FKEY_FKEY2 = 9; // >K>KK  - press one key and then two other keys fast and sequentially
	public static final int S_FKEY2_FKEY = 10; // >KK>K  - press two keys and then another one fast and sequentially
	public static final int S_FKEY2_FKEY2 = 11; // >KK>KK  - press two keys and then another one fast and sequentially
	public static final int S_FKEY_HKEY = 12; // >K[K]  - press fastly one key and then hold another key, sequentially
	public static final int S_FKEY_HKEY2 = 13; // >K[KK] - press fastly one key and then hold two ther ones, sequentially
	public static final int S_FKEY2_HKEY = 14; // >KK[K] - press two keys together fastly and then hold a third one, sequentially
	public static final int S_FKEY2_HKEY2 = 15; // >KK[KK] - press two keys together fastly and then hold a third one, sequentially
	public static final int MAX_ACTIONS = 300;
	public static final int JMB_EVENT_PRESS = 1;
	public static final int JMB_EVENT_RELEASE = 2;
	public static final int FKEY = 0; // <K>    - press one key fast
	public static final int FKEY2 = 1; // <KK>   - press two keys fast, together
	public static final int HKEY = 2; // [K]    - Hold one key
	public static final int HKEY2 = 3; // [KK]   - Hold two keys, together
	public static final int D_FKEY = 4; // <<K>>  - double press one key fast
	public static final int D_FKEY2 = 5; // <<KK>> - double press two keys fast, together
	public static final int FHKEY = 6; // <[K]>  - press fast and then hold one key
	public static final int FHKEY2 = 7; // <[KK]> - press fast and then hold two keys together
	public static final int STACK_SIZE = 4;
	public static final String actionTypeLabel[] = { "<K>", "<KK>", "[K]", "[KK]", "<<K>>", "<<KK>>", "<[K]>", "<[KK]>", ">K>K", ">K>KK", ">KK>K", ">KK>KK", ">K[K]", ">K[KK]", ">KK[K]", ">KK[KK]" };
	private Frame device;
	private Editor viewPort;
	private static JmbAction actionArray[] = new JmbAction[MAX_JMB_ACTIONS];
	private JmbAction map[] = new JmbAction[MAX_JMB_ACTIONS];
	private InfoViewer infoView;
	private String sCollectedNumber;
	private long eventTime[] = new long[STACK_SIZE];
	private int mousex;
	private int mousey;
	private int eventType[] = new int[STACK_SIZE];
	private int eventStack[] = new int[STACK_SIZE];
	private int stackIndex;
	private int collectedNumber;
	private int collectedDigit;
	private int pressEventCounter; // that avoid more than 2 press events in a same fact
	private int pairOf2;
	private int pairOf3;
	private int mapSize;
	private int fact1_k1;
	private int fact1_k2;
	private int fact2_k1;
	private int fact2_k2;
	private int fact1Type;
	private int wholeMapIndex;
	private int wholeActionType;
	private boolean isHolding;
	private boolean isPressEventLocked;
	private boolean isHoldingOutput;
	private boolean isFirstFact;
	private boolean everythingLocked;
	private boolean isDoubleKey;
	private boolean delayedSetCurrentObject;
	private boolean isCollecting;
	private boolean aborted;
	

		private String mapParsedLine;

		private static int controlCode[] = { 
			KeyEvent.VK_SPACE, 
			KeyEvent.VK_ESCAPE, 
			KeyEvent.VK_TAB, 
			0, // KeyEvent.VK_BACKTAB, 
			KeyEvent.VK_BACK_SPACE, 
			0, // KeyEvent.VK_RETURN, 
			KeyEvent.VK_ENTER, 
			KeyEvent.VK_INSERT, 
			KeyEvent.VK_DELETE, 
			KeyEvent.VK_PAUSE,
			0, // KeyEvent.VK_PRINT, 
			0, // KeyEvent.VK_SYSREQ, 
			KeyEvent.VK_CLEAR, 
			KeyEvent.VK_HOME, 
			KeyEvent.VK_END,
			KeyEvent.VK_LEFT, 
			KeyEvent.VK_UP, 
			KeyEvent.VK_RIGHT, 
			KeyEvent.VK_DOWN, 
			0, // KeyEvent.VK_PRIOR, 
			0, // KeyEvent.VK_NEXT, 
			KeyEvent.VK_SHIFT, 
			0, // KeyEvent.VK_CTRL, 
			KeyEvent.VK_META, 
			KeyEvent.VK_ALT, 
			0, // KeyEvent.VK_CAPS, 
			0, // KeyEvent.VK_NUMLOCK, 
			0 // KeyEvent.VK_SCROLL 
			};


		public Jmb(Frame device, InfoViewer infoView)
			{
			this.device = device;
			this.infoView = infoView;
			releaseTimer = new ReleaseTimer();
			secondFactTimer = new SecondChanceTimer();
			//clearOutputTimer = new Timer();
			releaseTimer.setActive(false);
			secondFactTimer.setActive(false);
			device.addKeyListener(this);
			viewPort = null;
			for (int i = 0; i < MAX_JMB_ACTIONS; i++)
				actionArray[i] = null;
			initMap();
			collectedDigit = 0;
			collectedNumber = -1;
			sCollectedNumber = "-1";
			everythingLocked = false;
			delayedSetCurrentObject = false;
			isFirstFact = true;
			}


		public void reset()
			{
			isFirstFact = true;
			isDoubleKey = false;
			fact1_k1 = 0;
			fact1_k2 = 0;
			fact2_k1 = 0;
			fact2_k2 = 0;
			isHolding = false;
			isPressEventLocked = false;
			stackIndex = 0;
			pressEventCounter = 0;
			wholeMapIndex = -1;
			aborted = false;
			for (int i = 0; i < STACK_SIZE; i++)
				{
				eventType[i] = 0;
				eventStack[i] = 0;
				eventTime[i] = 0;
				}
			}


		public void set_clear_time(int value)
			{
			}


		public void set_hold_sensitiveness(int value)
			{
			}


		public void set_double_fact_interval(int value)
			{
			}


		// Everthing starts here
		public void keyPressed(KeyEvent e)
			{
			if (!aborted)
				{
				update_mouse_variables();
				if (!isPressEventLocked)
					{
					if (pressEventCounter < 2)
						{
						pressEventCounter++;
						push_event(JMB_EVENT_PRESS, e);
						press_checker();
						if (!releaseTimer.isActive())
							{
							//UNCO releaseTimer.start(ASSUME_HOLD_TIME);//, true);
							}
						}
					else
						{
						isPressEventLocked = true;
						}
					}
				}
			}


		public void keyReleased(KeyEvent e)
			{
			if (aborted)
				{
				reset();
				return;
				}
			update_mouse_variables();
			if (!is_fake(JMB_EVENT_RELEASE, e))
				{
				push_event(JMB_EVENT_RELEASE, e);
				release_checker();
				}
			}


		// This pushed an event to the stack
		public void keyTyped(KeyEvent arg0)
			{
			}


		private void push_event(int pType, KeyEvent e)
			{
			if (stackIndex < STACK_SIZE)
				{
				int pCode = uppercase(e.getKeyCode());
				eventType[stackIndex] = pType;
				eventStack[stackIndex] = pCode;
				long ts = e.getWhen();
				eventTime[stackIndex] = ts;
				// log("Pushing EVENT " + stackIndex + " " + (pType == JMB_EVENT_PRESS ? "PRESS" : "RELEASE") + " key=" + pCode + "  at " + ts);
				stackIndex++;
				// For debuggin purposes
				for (int j = 0; j < STACK_SIZE; j++)
					{
					// log("eventStack[" + j + "]=" + (eventType[j] == JMB_EVENT_PRESS ? "P" : "R") + " " + eventStack[j] + " at timestamp=" + eventTime[j]);
					}
				}
			}


		private int uppercase(int k) //TODO Find a more elegant way...
			{
			if ((k > 90) && (k < 106))
				k -= 32;
			return k;
			}


		// this is a intermediate step to push. Only non-fake events are allowed to be pushed
		private boolean is_fake(int eventType, KeyEvent e)
			{
			boolean b = false;
			int code = uppercase(e.getKeyCode());
			// detect remaing release from last hold action.
			if ((stackIndex == 0) && (eventType == JMB_EVENT_RELEASE))
				b = true;
			// detect auto-repeat (IMPROVE ME)
			if ((stackIndex == 1) || (stackIndex == 3))
				{
				if ((this.eventType[stackIndex] == this.eventType[stackIndex - 1]) && (eventStack[stackIndex] == eventStack[stackIndex - 1]))
					{
					b = true;
					}
				}
			// Detected a invalid JMB_EVENT_PRESS  /  ?????  /
			if ((code != eventStack[0]) && (code != eventStack[1]) && (code != eventStack[2]) && (code != eventStack[3]))
				b = true;
			return b;
			}


		// this is the method that detects wether a fact is double key (KK) or single key (K)
		// by measuring the time difference between the first 2 presses (in case event[0] AND
		// event[1] are JMB_EVENT_PRESS'es
		void press_checker()
			{
			if (stackIndex == 2) // first two events happend
				{
				log("Checking if 2 first events are PRESS'es and what is the time diff between them");
				if ((eventType[0] == JMB_EVENT_PRESS) && (eventType[1] == JMB_EVENT_PRESS))
					{
					isDoubleKey = (eventTime[1] - eventTime[0] < MAX_TIME_DIFFERENCE_FOR_DOUBLE_KEY_CONSIDERATION);
					}
				if (isDoubleKey)
					{
					log("Detected Double Key (KK)");
					}
				}
			}


		// This is the one who consolidate a fact. It can detect
		// if the fact is K or KK, and also calls the push_fact handler
		private void release_checker()
			{
			if (isHolding)
				{
				finish_hold();
				return;
				}
			// If it is not fake, then lets take a look in the pairs
			// Remember that event 0 is always JMB_EVENT_PRESS
			if (eventType[1] == JMB_EVENT_RELEASE)
				{
				// event 0 and event 1 are a pair because event 1 must be the release of 0
				// since the event 1 was a release, then the first fact is finished
				// Now we must push_fact, but only if it is a release very fast (that
				// didnt cause a HOLD.
				if (!isHolding)
					push_fact(eventStack[0], 0);
				return;
				}
			// event 1 is ALSO a JMB_EVENT_PRESS. so we must check for a double key
			if (isDoubleKey) // ( very short time difference between 2 first presses)
				{
				// Now we dont know if this is the event 2 or event 3
				if (stackIndex == 3)
					{
					// So we are processing event (??), which is a release So this can be release of event 0 or 1 , so..
					if (eventStack[2] == eventStack[0])
						pairOf2 = 0;
					else
						pairOf2 = 1;
					// the event is not finished yet but it must be avoided any reentrance (any new JMB_EVENT_PRESS)
					isPressEventLocked = true;
					return;
					}
				// we are in the event 3 . the last! the event is finished , surely. But we must check the pairs
				// we already know who event 2 is pair of. So ...
				if (pairOf2 == 0)
					pairOf3 = 1;
				else
					pairOf3 = 0;
				// The event is finished, and we know the pairs. So..
				// I must push_fact, but only if it is a release very fast (that
				// didnt cause a HOLD.
				if (!isHolding)
					push_fact(eventStack[0], eventStack[1]);
				}
			else
				{
				// although the sequence was P-P-R-R (press, press, release, release)
				// the time difference between two first presses was too long
				// so it CAN'T be a double key (KK).
				// So the only possible explanation is that this is a >K>K action
				// where user typed too fast, so the release of 1st P happend AFTER the
				// 2nd press. In this case, I have to assume this is a >K>K
				// action, and dispatch TWO facts.
				// Now forcing prematurally a 2 facts by splitting and syncronizing
				// these events into 2 different facts. Probably this is a >K>K
				// of course, this  >K>K assumption mentioned above can be made ONLY if it is the first fact.
				if (isFirstFact)
					{
					log("Inital Double Key too slow. It must be a premature >K>K. Dispatching 2 individual <K> facts.");
					int f1_k1 = eventStack[0];
					int f2_k1 = eventStack[1];
					push_fact(f1_k1, 0);
					push_fact(f2_k1, 0);
					}
				}
			}


		//  JMB ENGINE : PRESS LEVEL HANDLING -------------------------
		private void push_fact(int k1, int k2)
			{
			log("Pushing FACT : K=(" + k1 + "," + k2 + ")");
			infoView.clear();
			if (isFirstFact)
				{
				releaseTimer.stop(); // quit the holding check..
				log("Cancelling releaseTimer ...");
				// this is the first fact
				log("F1 detected : K=(" + k1 + "," + k2 + ")");
				// first try to find some action like k1k2
				fact1_k1 = k1;
				fact1_k2 = k2;
				// first check if this fact is just a collected number
				check_number_collection();
				if (isCollecting)
					{
					// another digit was collected. Fine , go on jmb job..
					reset();
					return;
					}
				int mapIndex = identify_first_fact();
				if (mapIndex < 0)
					{
					log("F1 alone does not match anything in the map. Waiting for F2 ...");
					// Action not identified. Maybe is part of a double fact action. so...
					give_a_chance_for_second_fact();
					return;
					}
				log("F1 matches map in position " + mapIndex);
				// there is a single-fact action which matches this !! now must check if is not an immediate action
				if (!map[mapIndex].isInstantaneous)
					{
					log("Although this could be an SINGLE PRESS Action, it is not protected, so...");
					// action is not an immediate action, so...
					give_a_chance_for_second_fact();
					return;
					}
				// Action exists AND it is a immediate action. So
				// forces it to be a single fact action
				log("This is protected (immediate) action. It'll be treated as <K>");
				dispatch_action(mapIndex);
				conclusion();
				}
			else
				// ok . We are in the second fact.
				{
				secondFactTimer.stop();
				log("Cancelling secondChanceTimer ...");
				fact2_k1 = k1;
				fact2_k2 = k2;
				if (fact2_k1 != 0)
					{
					// this is the second press
					log("F2 detected : K=(" + k1 + "," + k2 + ")");
					}
				// try to complement the first press.
				wholeMapIndex = identify_first_and_second_facts_together();
				if (wholeMapIndex >= 0)
					{
					log("F1+F2 matches with action " + wholeMapIndex + " !! Dispatching it...");
					if (everythingLocked)
						{
						infoView.displayMainMessage("Jog Mouse Board is LOCKED !");
						return;
						}
					dispatch_action(wholeMapIndex);
					}
				else
					{
					log("No match for F1+F2 :-(");
					infoView.displayError("Unknown " + actionTypeLabel[wholeActionType] + " Action !");
					}
				conclusion();
				}
			}


		private int identify_first_fact()
			{
			fact1Type = 0;
			// First we need to know the first fact type.
			if (fact1_k2 == 0) // <K> or [K]
				{
				if (!isHolding)
					{
					log("Detected Fact <K>");
					fact1Type = FKEY;
					}
				else
					{
					log("Detected Fact [K]");
					fact1Type = HKEY;
					}
				}
			else
				// <KK> or [KK]
				{
				if (!isHolding)
					{
					log("Detected Fact <KK>");
					fact1Type = FKEY2;
					}
				else
					{
					log("Detected Fact [KK]");
					fact1Type = HKEY2;
					}
				}
			// Fact 1 Type identified .
			log("Checking for SINGLE FACT action : " + actionTypeLabel[fact1Type] + " (" + fact1_k1 + "," + fact1_k2 + ") ");
			for (int i = 0; i < mapSize; i++)
				{
				int ap1k1 = map[i].fact1_key1;
				int ap1k2 = map[i].fact1_key2;
				int ap2k1 = map[i].fact2_key1;
				int ap2k2 = map[i].fact2_key2;
				// // PMESG4("COMPARING %d,%d,%d,%d \tWITH %d,%d,%d,%d", ap1k1,ap1k2, ap2k1, ap2k2, fact1_k1, fact1_k2, fact2_k1,fact2_k2);
				if (map[i].type != fact1Type)
					continue;
				if ((((ap1k1 == fact1_k1) && (ap1k2 == fact1_k2)) || ((ap1k1 == fact1_k2) && (ap1k2 == fact1_k1))) && (ap2k1 == 0) && (ap2k2 == 0))
					{
					// 'i' is a candidate for first press
					log("Found a match in map position " + i);
					return i;
					}
				}
			return -1;
			}


		// This is called whenever a second fact happens right after the first
		private int identify_first_and_second_facts_together()
			{
			log("Composing F1 K=(" + fact1_k1 + "," + fact1_k2 + ") and F2 K=(" + fact2_k1 + "," + fact2_k2 + ") ");
			if (fact2_k1 != 0)
				{
				if (!isHolding)
					{
					if (fact1_k2 == 0) // first press is <K> (I know that its not [K] because if it was I'd never reach identify_first_and_second_facts_together())
						{
						if (fact2_k2 == 0)
							if (fact1_k1 == fact2_k1)
								{
								log("Whole action is a <<K>>");
								wholeActionType = D_FKEY; // <<K>>
								}
							else
								{
								log("Whole action is a >K>K");
								wholeActionType = S_FKEY_FKEY; // >K>K
								}
						else
							{
							log("Whole action is a >K>KK");
							wholeActionType = S_FKEY_FKEY2; // >K>KK
							}
						}
					else
						{
						if (fact2_k2 == 0)
							{
							log("Whole action is a >KK>K");
							wholeActionType = S_FKEY2_FKEY; // >KK>K
							}
						else
							{
							if (((fact1_k1 == fact2_k1) && (fact1_k2 == fact2_k2)) || ((fact1_k1 == fact2_k2) && (fact1_k2 == fact2_k1)))
								{
								log("Whole action is a <<KK>>");
								wholeActionType = D_FKEY2;
								}
							else
								{
								log("Whole action is a >KK>KK");
								wholeActionType = S_FKEY2_FKEY2;
								}
							}
						}
					}
				else
					{
					if (fact1_k2 == 0) // first press is <K> (I know that its not [K] because if it was I'd never reach identify_first_and_second_facts_together())
						{
						if (fact2_k2 == 0)
							if (fact1_k1 == fact2_k1)
								{
								log("Whole action is a <[K]>");
								wholeActionType = FHKEY;
								}
							else
								{
								log("Whole action is a >K[K]");
								wholeActionType = S_FKEY_HKEY;
								}
						else
							{
							log("Whole action is a >K[KK]");
							wholeActionType = S_FKEY_HKEY2;
							}
						}
					else
						{
						if (fact2_k2 == 0)
							{
							log("Whole action is a KK[K]");
							wholeActionType = S_FKEY2_HKEY;
							}
						else
							{
							log("Whole action is a >KK[KK]");
							wholeActionType = S_FKEY2_HKEY2;
							}
						}
					}
				}
			else
				{
				log("F2 is null ( K=(0,0) ). Assuming whole action type is " + actionTypeLabel[fact1Type]);
				wholeActionType = fact1Type;
				}
			// whole action type identified .
			log("Searching for a " + actionTypeLabel[wholeActionType] + " action that matches F1 K=(" + fact1_k1 + "," + fact1_k2 + ") and F2 K=(" + fact2_k1 + "," + fact2_k2 + ")");
			for (int i = 0; i < mapSize; i++)
				{
				if (map[i].type != wholeActionType)
					continue;
				int ap1k1 = map[i].fact1_key1;
				int ap1k2 = map[i].fact1_key2;
				int ap2k1 = map[i].fact2_key1;
				int ap2k2 = map[i].fact2_key2;
				// PMESG4("COMPARING %d,%d,%d,%d  \tWITH  %d,%d,%d,%d", ap1k1,ap1k2, ap2k1, ap2k2, fact1_k1, fact1_k2, fact2_k1,fact2_k2);
				if ((((ap1k1 == fact1_k1) && (ap1k2 == fact1_k2)) || ((ap1k1 == fact1_k2) && (ap1k2 == fact1_k1))) && (((ap2k1 == fact2_k1) && (ap2k2 == fact2_k2)) || ((ap2k1 == fact2_k2) && (ap2k2 == fact2_k1))))
					{
					// 'i' is a candidate the whole action
					log("Found a match : action #" + i);
					return i;
					}
				}
			log("No candidates found :-(");
			return -1;
			}


		private void give_a_chance_for_second_fact()
			{
			secondFactTimer.start(DOUBLE_FACT_WAIT_TIME);
			isFirstFact = false;
			isHolding = false;
			isPressEventLocked = false;
			stackIndex = 0;
			pressEventCounter = 0;
			fact2_k1 = 0;
			fact2_k2 = 0;
			wholeMapIndex = -1;
			for (int i = 0; i < STACK_SIZE; i++)
				{
				eventType[i] = 0;
				eventStack[i] = 0;
				eventTime[i] = 0;
				}
			}


		// ----------------------------- JMB ENGINE : ACTION LEVEL HANDLING -------------------------
		private void dispatch_action(int mapIndex)
			{
			int actionId = map[mapIndex].id;
			//MIG JmbMessage m = new JmbMessage(false, get_collected_number(), mousex, mousey);
			log("Processing Action id=" + actionId);
			//MIG broadcast_action(actionId,m);
			}


		private void finish_hold()
			{
			infoView.clear();
			releaseTimer.stop();
			log("Cancelling releaseTimer ...");
			log("Finishing hold action " + wholeMapIndex);
			if (false) //MIG currentJogMode)
				{
				/*
				 // If there is a currentJogMode, it means we are jogging. When jogging the mouse moves and
				 // because of this, the currentSetContext can change. So the "hold actions" slot may not be 
				 // called when we reach this point due changed contexts. But we know which slot was called 
				 // when starting the hold action, so we call it again ;-)
				 JmbMessage m = new JmbMessage(false, get_collected_number(), mousex, mousey);
				 if (holdingMustuxAction && holdingMustuxAction.handleByJmb)
				 {
				 currentJogMode=0;
				 holdingMustuxAction.finish_hold(&m);
				 if (holdingMustuxAction.prepare_actions())
				 {
				 holdingMustuxAction.set_valid(true);
				 }
				 holdingMustuxAction.do_action();
				 holdingMustuxAction = null;
				 }
				 else
				 {
				 if (targetSlotHoldAction)
				 {
				 targetSlotHoldAction(currentObject, &m);
				 }
				 currentJogMode=0;
				 }
				 */
				}
			else if (wholeMapIndex >= 0)
				{
				// WTF is this good for!?!?
				// int actionId = map[wholeMapIndex].id;
				//MIG JmbAction jmbe = map[wholeMapIndex];
				//MIG JmbMessage m = new JmbMessage(false, get_collected_number(), mousex, mousey);
				//MIG broadcast_action(actionId,m);
				}
			if (delayedSetCurrentObject)
				{
				//MIG currentObject = tempCurrentObject;
				delayedSetCurrentObject = false;
				}
			//MIG targetSlotHoldAction=0;
			conclusion();
			}


		private void conclusion()
			{
			infoView.displayStatus("Action Finished");
			reset();
			hold_output();
			}


		private void abort()
			{
			infoView.displayStatus("Action Aborted");
			aborted = true;
			hold_output();
			}


		private void hold_output()
			{
			if (!isHoldingOutput)
				{
				//MIG catcher.clearOutputTimer.start(clearTime);
				isHoldingOutput = true;
				}
			}


		protected void assume_hold() // no release so far ? so consider it a hold...
			{
			isHolding = true;
			if (false) //MIG JmbEventCatcher.clearOutputTimer != null)
				{
				//MIG catcher.clearOutputTimer.stop();
				isHoldingOutput = false;
				}
			//MIG infoView.clear_line(1);
			wholeMapIndex = -1;
			if (isFirstFact)
				{
				fact1_k1 = eventStack[0];
				fact1_k2 = eventStack[1];
				wholeMapIndex = identify_first_fact(); // I can consider first press the last because there is nothing after a []
				}
			else
				{
				fact2_k1 = eventStack[0];
				fact2_k2 = eventStack[1];
				wholeMapIndex = identify_first_and_second_facts_together();
				}
			if (wholeMapIndex >= 0)
				{
				int actionId = map[wholeMapIndex].id;
				log("Processing HOLD Action index=" + wholeMapIndex + " id=" + actionId);
				JmbAction jmbe = map[wholeMapIndex];
				infoView.displayDirections(jmbe.useX, jmbe.useX, jmbe.useY, jmbe.useY);
				//MIG JmbMessage m = new JmbMessage(true, get_collected_number(), mousex, mousey);
				//MIG broadcast_action(actionId,&m);
				}
			else
				{
				String s = "Unknown " + actionTypeLabel[fact1Type] + " Action !";
				infoView.displayError(s);
				abort();
				}
			stop_collecting();
			// note that we dont call conclusion() here :-)
			}


		void clear_output()
			{
			if ((isHoldingOutput))
				{
				// UNCOMMENT ME ASAP output.clear();
				//MIG clearOutputTimer.stop();
				isHoldingOutput = false;
				// UNCOMMENT ME ASAP output.draw_all_grids();
				for (int k = 0; k < 4; k++)
					{
					//infoView.clear_line(k);
					//infoView.draw_line_grid(k);
					}
				infoView.displayMainMessage("JMB ACTIVE");
				}

			}


		void quit_second_chance()
			{
			log("Stopping secondChanceTimer ...");
			secondFactTimer.stop();
			if (!isHolding) // if it is holding, there is no need to push a new fact
				{
				log("No second fact ... Forcing a null second fact");
				push_fact(0, 0); // no second press performed, so I am adding a pair of Zeros as second press and keep going...
				}
			}


		private void check_number_collection()
			{
			if ((fact1_k1 >= 0x30) && (fact1_k1 <= 0x39))
				{
				if (!isCollecting)
					{
					log("Starting number collection...");
					sCollectedNumber = "";
					}
				isCollecting = true;
				collectedDigit++;
				sCollectedNumber += fact1_k1; // FIXME it had a ",1" complement after fact1_k1... why?
				log("Collected " + sCollectedNumber + " so far...");
				String sn = "NUMBER " + sCollectedNumber;
				collectedNumber = Util.toInt(sCollectedNumber);
				infoView.displayInstruction(sn);
				}
			else
				stop_collecting();
			}


		public int get_collected_number()
			{
			int n = collectedNumber;
			collectedNumber = -1; // collectedNumber has a life of only one get.
			return n;
			}


		private void stop_collecting()
			{
			isCollecting = false;
			collectedDigit = 0;
			collectedNumber = Util.toInt(sCollectedNumber);
			sCollectedNumber = "-1";
			//infoView.info("Number Collection Finished");
			}


		public void lock()
			{
			everythingLocked = true;
			}


		public void unlock()
			{
			everythingLocked = false;
			}


		// NOTE : an actionId will be linked to both a context and a slot
		int connect_actionid_to_slot()//MIG Object object,int actionId,JmbMode* mode,JmbContext* context,MustuxAction* (*aslot)( void*, JmbMessage* ),int property)
			{
			/* if ((actionId>MAX_JMB_ACTIONS) || (actionId<0))
			 {
			 PERROR("Trying to connect a Jmb Slot to an invalid action id");
			 return -1;
			 }
			 // No JmbAction for this actionId? fine, add the first
			 if (!actionArray[actionId])
			 actionArray[actionId] = new JmbAction(object, mode, context, aslot, property);
			 
			 // It looks like there is allready an JmbAction for this actionId
			 else
			 {
			 // First check if the context asociated with the JmbAction is global, if so we cant' add another action to this actionId
			 if (actionArray[actionId].property == ApplicationGlobal)
			 {
			 PERROR("Cannot add another possible action for this actionId=%d (there is a Application Global action associated to it already)",actionId);
			 ;
			 return -1;
			 }
			 // The JmbAction for this actionId is not a global one, we can add another 
			 //  JmbAction to this actionId.
			 //  If the property equals UseCollectedNumber, check if there isn't a JmbAction
			 //  which is using it allready, we only are allowed to have one JmbAction with this property,
			 //  or when the objects are from the same type, then it is allowed to have multiple UseCollectedNumber
			 //  for the same actionId
			 //
			 JmbAction* action = actionArray[actionId];
			 
			 boolean allowNewActionWithUseCollectedNumber = false;
			 boolean noActionWithUseCollectedNumber = true;
			 if (property == UseCollectedNumber)
			 {
			 // check if there is allready an action which has the property UseCollectedNumber 
			 while (action)
			 {
			 if (action.property == UseCollectedNumber)
			 {
			 // Ooops, there is allready an action with property UseCollectedNumber,
			 //check if the object is from the same type, if so it is allowed to 
			 //  add another action to this actionId 
			 
			 noActionWithUseCollectedNumber = false;
			 // This is a dirty trick to get the classNames of the objects by using the class definition 
			 //   Q_OBJECT macro. Since we know that object is either a QWidget or a MustuxWidget, we can 
			 //   do this trick without problems. Also, we don't want to force a developer to use a 
			 //   MustuxWidget, but since we cannot cast a MustuxWidget to a QWidget, we do it vica versa
			 //   which can be done without a problem.
			 // 
			 MustuxWidget* widgeta;
			 MustuxWidget* widgetb;
			 widgeta = (MustuxWidget*) object;
			 widgetb = (MustuxWidget*) action.object;
			 if ( widgeta.className() == widgetb.className() )
			 {
			 allowNewActionWithUseCollectedNumber = true;
			 action.onlyActionWhichIsUsingUseCollectedNumber = false;
			 break;
			 }
			 }
			 action = action.next;
			 }
			 if ((!allowNewActionWithUseCollectedNumber) && (!noActionWithUseCollectedNumber))
			 {
			 PERROR("There is allready (another type of) object which is using UseCollectedNumber for this actionId (%d)", actionId);
			 return -1;
			 }
			 }
			 // Ok, we made it up to here, so we are allowed to make a new JmbAction and put it in the actionArray at index actionId 
			 action = new JmbAction(object, mode, context, aslot, property);
			 action.next=actionArray[actionId];
			 actionArray[actionId]=action;
			 if (allowNewActionWithUseCollectedNumber)
			 action.onlyActionWhichIsUsingUseCollectedNumber = false;
			 }
			 // PMESG2("Connecting : actionID %d will call object %p slot %p when current context is %p and current mode is %p",actionId,object,aslot,context, mode);
			 
			 ;*/
			return 1;
			}


		void disconnect(Object object)
			{
			/*	 JmbAction action;
			 JmbAction prev = null;
			 int counter=0;
			 for (int i=0; i<MAX_JMB_ACTIONS; i++)
			 {
			 //MIG action = actionArray[i]; 
			 if (action!=null)
			 {
			 counter=0;
			 prev=null;
			 while(action!=null)
			 {
			 counter +=1;
			 if (false) //MIG action.object == object)
			 {
			 if (prev==null)
			 {
			 //MIG actionArray[i] =  action.next;
			 action =null;
			 //MIG action = actionArray[i];
			 counter = 0;
			 if (action==null)
			 break;
			 }
			 else
			 {
			 //MIG prev.next = action.next;
			 action=null;
			 action = prev;
			 }
			 }
			 if (counter >= 1)
			 {
			 prev = action;
			 //MIG action = action.next;
			 }
			 }
			 }
			 }
			 */}


		int connect_jogmode_to_slot() //MIGObject object, JmbJogMode* mode, MustuxAction* (*slot)(void*, JmbMessage*))
			{
			/*	 //MIG mode.slot=slot;
			 //MIG mode.object=object;
			 }
			 

			 int broadcast_action()//MIG int actionId, JmbMessage* m)
			 {
			 // Get the (first) action linked to the actionId
			 
			 JmbAction action = actionArray[actionId];
			 if (action!=null)
			 {
			 // PMESG("ActionId=%d not connected to anything !",actionId);
			 return -1;
			 }
			 
			 //MIG MustuxAction* (*targetSlot)(Object object,JmbMessage*) = 0;
			 Object obj = null;
			 
			 // Is this action's context a ApplicationGlobal one? Then there is only one action connected 
			 //  to this actionId, we can get the action's object and slot directly.
			 
			 if (action.property == ApplicationGlobal)
			 {
			 // PMESG2("Getting context variables for Global context");
			 obj=action.object;
			 //MIG targetSlot = *(action.slot);
			 }
			 
			 // There are more actions connected to the actionId, iterate over the actions to catch the right one 
			 else
			 {
			 boolean match = false;
			 boolean checkForValidContextAndMode = false;
			 while(action)
			 {
			 // check if we're using collectedNumber, if so the  actions property has to be UseCollectedNumber. If we aren't using collectedNumber, then the actions property should be != collectedNumber 
			 if ((m.collectedNumber == -1) && (action.property != UseCollectedNumber))
			 {
			 checkForValidContextAndMode = true;
			 }
			 else if ((m.collectedNumber >= 0) && (action.property == UseCollectedNumber))
			 {
			 // There can be more actions with the same kind of object which areusing UseCollectedNumber. If action.onlyActionWhichIsUsingUseCollectedNumber equals true then there is only one such kind of object at any time, and since there can only be  	 one actionId which is using UseCollectedNumber we don't have to check for object equality
			 
			 if (action.onlyActionWhichIsUsingUseCollectedNumber)
			 {
			 match = true;
			 break;
			 }
			 else
			 {
			 // Now there can exist more than one of the same object at a time which is using UseCollectedNumber so we need to check for currentContext and/or currentMode equality, and/or object equality against currentObject
			 checkForValidContextAndMode = true;
			 }
			 }
			 // Here we check for currenObject == action.object and if action.mode and action.contextmatch. 
			 if (checkForValidContextAndMode)
			 {
			 if (action.object == currentObject)
			 {
			 // Is this action ObjectGlobal? If so we allready know that currentObject equals action.object so we have a match
			 
			 if ((!action.context) && (!action.mode))
			 {
			 match = true;
			 break;
			 }
			 // The current set Context and/or Mode do matter, so we check for these.
			 
			 else if (is_valid_context(action.context) && is_valid_mode(action.mode))
			 {
			 match = true;
			 break;
			 }
			 }
			 checkForValidContextAndMode = false;
			 }
			 // PMESG("Check next action");
			 action = action.next;
			 }
			 if (match)
			 {
			 // This action is valid, get the object and slot 
			 // PMESG2("Getting context variables for NON Global action");
			 obj=action.object;
			 targetSlot = *(action.slot);
			 }
			 }
			 
			 MustuxAction* k;
			 if (obj)
			 {
			 // PMESG("Broadcasting to object=%p : Calling JmbSlot %p",obj,targetSlot);
			 k = targetSlot(obj, m);
			 if (k && (!isHolding))
			 {
			 if (k.prepare_actions() && k.do_action())
			 {
			 k.set_valid(true);
			 ((MustuxSlotContainer*)obj).mhe.push(k);
			 }
			 }
			 if (k && isHolding)
			 {
			 if (k.begin_hold(m))
			 {
			 ((MustuxSlotContainer*)obj).mhe.push(k); 
			 holdingMustuxAction = k;
			 }
			 else
			 {
			 // OOPSSS, something went wrong when making the MustuxAction
			 // set following stuff to zero to make finish_hold do nothing
			 delete k;
			 targetSlotHoldAction = 0;
			 currentJogMode = 0;
			 wholeMapIndex = -1;
			 // 				finish_hold();
			 }
			 }
			 // If the 'fact' was a "hold" 'fact', there has to be a release at some time. Since we know
			 allready the slot which is then called, store the targetSlot to targetSlotHoldAction
			 and call it directly when the 'hold' action is finished (done in finish_hold() )
			 
			 if (isHolding)
			 targetSlotHoldAction = *(action.slot);
			 }
			 else
			 PERROR("Could not match current context %p with actionId=%d",currentContext,actionId)
			 ;*/
			return 1;
			}


		void jog()
			{
			/*if (false) //MIG currentJogMode)
			 {
			 update_mouse_variables();
			 JmbMessage m =new JmbMessage(false, collectedNumber, mousex, mousey);
			 
			 if (false) //MIG holdingMustuxAction && holdingMustuxAction.handleByJmb && holdingMustuxAction.valid())
			 {//MIG holdingMustuxAction.jog(&m);
			 }
			 else
			 {
			 //MIG MustuxAction* (*targetSlot)(Object object,JmbMessage*) = *(currentJogMode.slot);
			 Object obj = null; //MIG currentJogMode.object;
			 if (false) //MIG targetSlot)
			 {
			 //MIG targetSlot(obj, &m);
			 }
			 }
			 }*/
			}


		void set_current_context()//MIG JmbContext* context)
			{
			//MIG currentContext=context;
			}


		int get_current_context()//MIG JmbContext* get_current_context()
			{
			return 0;//MIG currentContext;
			}


		void set_current_mode()//MIG JmbMode* mode)
			{
			//MIG currentMode=mode;
			}


		public int get_current_mode()//MIG JmbMode* get_current_mode()
			{
			return 0;//MIG currentMode;
			}


		void set_current_object(Object object)
			{
			if (isHolding)
				{
				//MIG tempCurrentObject = (MustuxSlotContainer*) object;
				delayedSetCurrentObject = true;
				}
			else
				{
				//MIB currentObject = (MustuxSlotContainer*) object;
				}
			}


		boolean is_valid_context()//MIG JmbContext* context)
			{
			boolean b = false;
			/*	 
			 if (!context)
			 {
			 // We have a null context here, so we don't have to check for the validity of it against the current set one. We can return true here.
			 
			 // PMESG("Empty context, this context is valid!");
			 return true;
			 }
			 
			 if (context.child0 && context.child1)
			 {
			 // PMESG("This context has childs, check the childs");
			 if (context.boolOperator == OR)
			 {
			 b = is_valid_context(context.child0);
			 if (!b)
			 b = is_valid_context(context.child1);
			 }
			 else if (context.boolOperator == AND)
			 {
			 b = is_valid_context(context.child0);
			 if (b)
			 b = is_valid_context(context.child1);
			 }
			 }
			 else if (context.boolOperator == NOT)
			 {
			 b = is_valid_for_currentcontext(context.child0, NOT);
			 }
			 else 
			 {
			 // PMESG("This context has no childs, check if it is valid against currentContext");
			 b = is_valid_for_currentcontext(context, AND);
			 }
			 */
			return b;
			}


		boolean is_valid_mode()//MIG JmbMode* mode)
			{
			boolean b = false;
			/*
			 if (false)//MIG !mode)
			 {
			 // We have a null mode here, so we don't have to check for the validity of it	 against the current set one. We can return true here.
			 
			 // PMESG("Empty mode, this mode is valid!");
			 return true;
			 }
			 
			 if (mode.child0 && mode.child1)
			 {
			 // PMESG("This mode has childs, check the childs");
			 if (mode.boolOperator == OR)
			 {
			 b = is_valid_mode(mode.child0);
			 if (!b)
			 b = is_valid_mode(mode.child1);
			 }
			 else if (mode.boolOperator == AND)
			 {
			 b = is_valid_mode(mode.child0);
			 if (b)
			 b = is_valid_mode(mode.child1);
			 }
			 }
			 else if (mode.boolOperator == NOT)
			 {
			 b = is_valid_for_currentmode(mode.child0, NOT);
			 }
			 else 
			 {
			 // PMESG("This mode has no childs, check if it is valid against currentMode");
			 b = is_valid_for_currentmode(mode, AND);
			 }
			 */
			return b;
			}


		boolean is_valid_for_currentcontext()//MIG JmbContext* context, int boolOperator)
			{
			boolean b = false;
			/*
			 if (!currentContext)
			 {
			 PWARN("CurrentContext is not set");
			 return false;
			 }
			 
			 if (boolOperator == AND)
			 {
			 // PMESG("Checking if this currentContext contains this context");
			 b = context_contains_context(context, currentContext);
			 }
			 else if (boolOperator == NOT)
			 {
			 // PMESG("Checking if this currentContext contains NOT this context");
			 b = !context_contains_context(context, currentContext);
			 }
			 */return b;
			}


		boolean is_valid_for_currentmode()//MIG JmbMode* mode, int boolOperator)
			{
			boolean b = false;
			/*	 
			 if (!currentMode)
			 {
			 PWARN("CurrentMode is not set");
			 return false;
			 }
			 
			 if (boolOperator == AND)
			 {
			 // PMESG("Checking if this currentMode contains this mode");
			 b = mode_contains_mode(mode, currentMode);
			 }
			 else if (boolOperator == NOT)
			 {
			 // PMESG("Checking if this currentMode contains NOT this mode");
			 b = !mode_contains_mode(mode, currentMode);
			 }
			 */return b;
			}


		boolean context_contains_context()//MIG JmbContext* contextToBeContained, JmbContext* context)
			{
			boolean b = false;
			/*	 
			 if (context.child0 && context.child1)
			 {
			 // PMESG("currentContext has childs, check the childs");
			 b = context_contains_context(contextToBeContained, context.child0);
			 if (!b)
			 b = context_contains_context(contextToBeContained, context.child1);
			 }
			 else
			 {
			 // PMESG("currentContext has no child context, test if this currentContext == context");
			 if (contextToBeContained == context)
			 b = true;
			 }
			 */
			return b;
			}


		boolean mode_contains_mode()//MIG JmbMode* modeToBeContained, JmbMode* mode)
			{
			boolean b = false;
			/*
			 if (mode.child0 && mode.child1)
			 {
			 // PMESG("currentMode has childs, check the childs");
			 b = mode_contains_mode(modeToBeContained, mode.child0);
			 if (!b)
			 b = mode_contains_mode(modeToBeContained, mode.child1);
			 }
			 else
			 {
			 // PMESG("currentMode has no child modes, test if this currentMode == mode");
			 if (modeToBeContained == mode)
			 b = true;
			 }
			 */
			return b;
			}


		private void initMap()
			{
			mapSize = 0;
			int index = 0;
			int x;
			mapParsedLine = null;
			int totalTags = JmbMap.tag.length;
			for ( int z = 0; z< totalTags; z++)
				{
				mapParsedLine = JmbMap.tag[z];
				// ignore comments
				x = mapParsedLine.indexOf("#");
				if (x >= 0)
					mapParsedLine = mapParsedLine.substring(0, x);
				mapParsedLine = mapParsedLine.trim();
				if (mapParsedLine.length() == 0)
					{
					continue;
					}
				boolean useX = false, useY = false;
				String sid = null;
				String tagL = null;
				String tagR = null;
				String parsedKeyText1 = null;
				String parsedKeyText2 = null;
				String parsedKeyText3 = null;
				String parsedKeyText4 = null;
				String incognite = null;
				String axisLabel = null;
				int keyCode1 = 0, keyCode2 = 0, keyCode3 = 0, keyCode4 = 0;
				int parsedActionType = -1;
				sid = NEXT_TOKEN();
				int actionId = Util.toInt(sid);
				tagL = NEXT_TOKEN();
				if (!tagL.equals(">")) // non sequence action
					{
					boolean doubleKey = false;
					parsedKeyText1 = NEXT_TOKEN();
					if (IS_A_TAG(parsedKeyText1))
						{
						//PARSE ERROR. SHOULD NOT BE A TAG HERE!
						}
					incognite = NEXT_TOKEN();
					if (NOT_A_TAG(incognite))
						{
						parsedKeyText2 = incognite;
						tagR = NEXT_TOKEN();
						doubleKey = true;
						}
					else
						tagR = incognite;
					if (tagL.equals("<") && tagR.equals(">"))
						{
						if (doubleKey)
							parsedActionType = FKEY2;
						else
							parsedActionType = FKEY;
						}
					else if (tagL.equals("<<") && tagR.equals(">>"))
						{
						if (doubleKey)
							parsedActionType = D_FKEY2;
						else
							parsedActionType = D_FKEY;
						}
					else if (tagL.equals("[") && tagR.equals("]"))
						{
						if (doubleKey)
							parsedActionType = HKEY2;
						else
							parsedActionType = HKEY;
						}
					else if (tagL.equals("<[") && tagR.equals("]>"))
						{
						if (doubleKey)
							parsedActionType = FHKEY2;
						else
							parsedActionType = FHKEY;
						}
					else
						{
						error("initmap","tagL and tagR are not a consistent pair !");
						}
					}
				else
					// sequence action TODO
					{
					parsedKeyText1 = NEXT_TOKEN();
					NEXT_TOKEN(); // get middle tag
					parsedKeyText2 = NEXT_TOKEN();
					tagR = NEXT_TOKEN();
					}
				axisLabel = NEXT_TOKEN();
				if ((axisLabel != null) && (axisLabel.indexOf("LR") >= 0))
					{
					useX = true;
					}
				if ((axisLabel != null) && (axisLabel.indexOf("UD") >= 0))
					{
					useY = true;
					}
				// find hex codes for parsedkeytext1 and 2
				keyCode1 = getHexCode(parsedKeyText1);
				keyCode2 = getHexCode(parsedKeyText2);
				keyCode3 = getHexCode(parsedKeyText3);
				keyCode4 = getHexCode(parsedKeyText4);
				// Fix the keyCode positions
				if ((parsedActionType == D_FKEY) || (parsedActionType == FHKEY))
					{
					keyCode3 = keyCode1;
					}
				else if ((parsedActionType == D_FKEY2) || (parsedActionType == FHKEY))
					{
					keyCode3 = keyCode1;
					keyCode4 = keyCode2;
					}
				else if ((parsedActionType == S_FKEY_FKEY) || (parsedActionType == S_FKEY_HKEY))
					{
					keyCode3 = keyCode2;
					keyCode2 = 0;
					}
				log("ADDED action: id=" + actionId + " type=" + parsedActionType + " keys=" + keyCode1 + "," + keyCode2 + "," + keyCode3 + "," + keyCode4 + " useX=" + useX + " useY=" + useY);
				map[index] = new JmbAction();
				map[index].set(actionId, parsedActionType, keyCode1, keyCode2, keyCode3, keyCode4, useX, useY);
				index++;
				}
			mapSize = index;
			log("Optimizing map for best performance ...");
			int optimizedActions = 0;
			for (int i = 0; i < mapSize; i++)
				{
				int c1A = map[i].fact1_key1;
				int c2A = map[i].fact1_key2;
				if ((map[i].type == FKEY))
					{
					boolean alone = true;
					for (int j = 0; j < mapSize; j++)
						{
						if (j == i)
							{
							continue;
							}
						int tt = map[j].type;
						int t1A = map[j].fact1_key1;
						if ((t1A == c1A) && ((tt == D_FKEY) || (tt == FHKEY) || (tt == S_FKEY_FKEY) || (tt == S_FKEY_FKEY2) || (tt == S_FKEY_HKEY) || (tt == S_FKEY_HKEY2)))
							{
							alone = false;
							}
						}
					if (alone)
						{
						// log("Setting <K> action id=" + map[i].id + " as protected (Instantaneous response)");
						map[i].set_instantaneous(true);
						optimizedActions++;
						}
					else
						{
						map[i].set_instantaneous(false);
						}
					}
				else if ((map[i].type == FKEY2))
					{
					boolean alone = true;
					for (int j = 0; j < mapSize; j++)
						{
						if (j == i)
							continue;
						int tt = map[j].type;
						int t1A = map[j].fact1_key1;
						int t2A = map[j].fact1_key2;
						if (((t1A == c1A) && (t2A == c2A)) && ((tt == HKEY) || (tt == D_FKEY2) || (tt == S_FKEY2_FKEY) || (tt == S_FKEY2_FKEY2) || (tt == S_FKEY2_HKEY) || (tt == S_FKEY2_HKEY2)))
							{
							alone = false;
							}
						}
					if (alone)
						{
						log("Setting <KK> action id=" + map[i].id + " for instantaneous response");
						map[i].set_instantaneous(true);
						optimizedActions++;
						}
					else
						{
						map[i].set_instantaneous(false);
						}
					}
				}
			log("JMB MAP INITIALIZED! " + mapSize + " actions registered ( " + optimizedActions + " optimized) .");
			}


		private static int getHexCode(String text)
			{
			int hc = 0;
			int x;
			String s;
			if ((text != null) && (!text.equals("")))
				{
				s = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
				x = s.indexOf(text);
				if (x >= 0)
					{
					hc = 65 + x;
					}
				else
					{
					s = "|SPACE  | ESC    |TAB     |BACKTAB |BKSPACE |RETURN  |ENTER   |INSERT  |DELETE  |PAUSE   |PRINT   |SYSREQ  |CLEAR   |HOME    |END     |LARROW  |UARROW  |RARROW  |DARROW  |PRIOR   |NEXT    |SHIFT   |CTRL    |META    |ALT     |CAPS    |NUMLOCK |SCROLL  ";
					x = s.indexOf("|" + text);
					if (x >= 0)
						{
						hc = controlCode[x / 9];
						}
					}
				}
			return hc;
			}


		public String NEXT_TOKEN()
			{
			mapParsedLine = mapParsedLine + " ";
			int x = mapParsedLine.indexOf(" ");
			String variable = null;
			if (x > 0)
				{
				variable = mapParsedLine.substring(0, x);
				mapParsedLine = mapParsedLine.substring(x + 1);
				}
			mapParsedLine = mapParsedLine.trim();
			return variable;
			}


		public static boolean IS_A_TAG(String x)
			{
			boolean b = (x.equals("<") || x.equals("<<") || x.equals("[") || x.equals("<[") || x.equals(">") || x.equals(">>") || x.equals("]") || x.equals("]>"));
			return b;
			}


		public static boolean NOT_A_TAG(String s)
			{
			return !IS_A_TAG(s);
			}

	
		// Internal structure that holds JmbActions for a given actionId
		//MIG JmbContext* currentContext;
		//MIG JmbMode* currentMode;
		//MIG JmbJogMode* currentJogMode;
		//MIG MustuxSlotContainer* currentObject;
		//MIG MustuxSlotContainer* tempCurrentObject;
		//MIG MustuxAction* (*targetSlotHoldAction)(Object object,JmbMessage*) = 0;
		//MIG MustuxAction* holdingMustuxAction = (MustuxAction*) 0;
		class ReleaseTimer extends OneShotTimer
			{
				synchronized void start(long ms)
					{
					log("Starting ReleaseTimer ...");
					try
						{
						setActive(true);
						schedule(new ReleaseTimeoutTask(), ms);
						}
					catch (Exception e)
						{
						}
					}
			}

		class SecondChanceTimer extends OneShotTimer
			{
				synchronized void start(long ms)
					{
					log("Starting SecondChanceTimer");
					try
						{
						setActive(true);
						schedule(new SecondChanceTimeoutTask(), ms);
						}
					catch (Exception e)
						{
						}
					}
			}

		class ReleaseTimeoutTask extends TimerTask
			{
				@Override
				public void run()
					{
					try
						{
						releaseTimer.stop();
						}
					catch (Exception e)
						{
						}
					log("Release Timeout : (" + Jmb.ASSUME_HOLD_TIME + " ms). Assuming this is a hold and dispatching it");
					assume_hold();
					}
			}

		class SecondChanceTimeoutTask extends TimerTask
			{
				@Override
				public void run()
					{
					try
						{
						secondFactTimer.stop();
						}
					catch (Exception e)
						{
						}
					log("Second-Chance Timeout : (" + Jmb.DOUBLE_FACT_WAIT_TIME + " ms)");
					quit_second_chance();
					}
			}


		public Frame getDevice()
			{
			return device;
			}


		public void setDevice(Frame device)
			{
			this.device = device;
			}


		public void set_viewport_widget(Editor w)
			{
			viewPort = w;
			}


		public void update_mouse_variables()
			{
			if (viewPort != null)
				{
				mousex = viewPort.getMouseX();
				mousey = viewPort.getMouseY();
				}
			}


		public boolean is_locked()
			{
			return everythingLocked;
			}


		
		private static void error(String where, String what)
			{
			Debugger.error("Jmb",where,what);
			}


		private static void log(String m)
			{
			PMESG(4, "[ Jmb ] " + m);
			}
	}
